#include "StdAfx.h"

#include <stdio.h>
#include <wtypes.h>
#include <winbase.h>
#include <winioctl.h>

static TCHAR	_tcSoftware[256];
static DWORD	_opflags;
static HANDLE	hDebugFile = NULL;
static HANDLE	hDebugFileMutex = NULL;

#include "DebugOutput.h"
#include "mcrcommon.h"

//#pragma managed

CDebugOutput::CDebugOutput(void)
{
	hDebugFile      = NULL;
	hDebugFileMutex = NULL;
}

CDebugOutput::~CDebugOutput(void)
{
}


////////////////////////////////////////////////////////////////////////////////////////////
// Debugging support

char CDebugOutput::DebugInit(TCHAR *tcSft, TCHAR *tcLogFile, DWORD dwFlags)
{
	OSVERSIONINFO	osVer;
	int             ii;
	hDebugFile      = NULL;
	hDebugFileMutex = NULL;

	osVer.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osVer);

	_opflags = dwFlags;

	for (ii=0; ii<sizeof(_tcSoftware)/sizeof(TCHAR); ii++)
	{
		_tcSoftware[ii] = tcSft[ii];
	}

	if (_opflags & DEBUG_FILE)
	{
		if (!hDebugFile)
		{
		hDebugFile = CreateFile(tcLogFile,
								GENERIC_READ | GENERIC_WRITE,
								FILE_SHARE_READ | FILE_SHARE_WRITE,
								NULL,
								OPEN_ALWAYS,
								FILE_ATTRIBUTE_NORMAL,
								NULL);
		}

		if (!hDebugFileMutex)
		{
			hDebugFileMutex = CreateMutex(NULL, FALSE, _T("Mutex_MC_DebugFile"));
		}


		if (hDebugFile == INVALID_HANDLE_VALUE)
		{
			hDebugFile = NULL;

			return(MCR_FAIL);
		}


		if ((_opflags & DEBUG_NOHELLO) == 0)
		{
        #ifdef _DEBUG_INFO
			DebugMsg(_T("************************************************"));
			DebugMsg(_T("Debugging Services initialized for "));
			DebugMsg(_tcSoftware);
			DebugMsg(_T("************************************************"));
        #endif
		}

		return MCR_SUCCESS;

	} else
	{
		return MCR_FAIL;
	}
}


void CDebugOutput::DebugClose(void)
{
	if (hDebugFile)
		CloseHandle(hDebugFile);

	if (hDebugFileMutex)
		CloseHandle(hDebugFileMutex);
}


// Even though we receive a TCHAR, the file write is BYTE based so we need to convert the TCHAR
// string back to char
void CDebugOutput::DebugMsg(TCHAR *pMsg)
{
	int     ii, iLength;
	char	cBuf[1024];
    char    cTemp[1024];
	ULONG	cbReturned;

	if (!hDebugFile)
		return;

	if ((_opflags & DEBUG_NOLINEHEAD) == 0)
	{
		// convert _tcSoftware to a character
		// convert TCHAR to char and copy the TCHAR string
		for (ii=0; ii<sizeof(_tcSoftware)/sizeof(TCHAR); ii++)
		{
			if (_tcSoftware[ii] == 0)
				break;
		}

		iLength = ii+1;

		if (iLength >= sizeof(cTemp))
		{
			iLength = sizeof(cTemp) - 1;
		}

		for (ii=0; ii<iLength; ii++)
		{
			cTemp[ii] = (char)_tcSoftware[ii];
		}
		cTemp[ii] = 0;

		sprintf_s(cBuf,sizeof(cBuf), "%s (%lx): ",cTemp,GetCurrentThreadId());
	}
	else
	{
		cBuf[0] = '\0';
	}
    
	// convert TCHAR to char and copy the TCHAR string
	for (ii=0; ii<sizeof(cTemp); ii++)
	{
		if (pMsg[ii] == 0)
			break;
	}

	iLength = ii+1;

	if (iLength >= sizeof(cTemp))
	{
		iLength = sizeof(cTemp) - 1;
	}
	for (ii=0; ii<iLength; ii++)
	{
		cTemp[ii] = (char)*pMsg++;
	}
	cTemp[ii] = 0;

	strcat_s(cBuf,sizeof(cBuf), cTemp);
	strcat_s(cBuf,sizeof(cBuf), "\r\n");

	if (_opflags & DEBUG_FILE)
	{
		WaitForSingleObject(hDebugFileMutex, INFINITE);
		SetFilePointer(hDebugFile,0,NULL,FILE_END);
		WriteFile(hDebugFile,cBuf,(DWORD)strlen(cBuf),&cbReturned,NULL);
		FlushFileBuffers(hDebugFile);
		ReleaseMutex(hDebugFileMutex);
	}
}


char CDebugOutput::DebugReset(void)
{

	if (!hDebugFile)
		return FALSE;

	if (_opflags & DEBUG_FILE)
	{
		SetFilePointer(hDebugFile,0,NULL,FILE_BEGIN);
		SetEndOfFile(hDebugFile);

		return(MCR_SUCCESS);
	}
	else
		return(MCR_FAIL);
}