
#include <stdlib.h>
#include <malloc.h>
#include "IntDLL.h"
#include "IntApi.h"
#include "SoftwareConfig.h"
#include "McrSerialInterface.h"
#include "McrSerialProtocol.h"
#include "DebugOutput.h"

extern CDebugOutput *cDebugOutput;
extern LARGE_INTEGER g_liPerfCountValueTopOfSecond;

//
//		Name:			MCR_ApiGetDLLInitStatus()
//
//		Description:	Returns DLL initialization status
//

DWORD __stdcall MCR_ApiGetDLLInitStatus(void)
{
	DWORD InitStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetDLLInitStatus[["));
	#endif

	AcquireApiMutex();
	InitStatus = g_dllInitStatus;
	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return(InitStatus);
}

//
//		Name:			MCR_ApiGetTimeCodeStatus()
//
//		Description:	Returns current status of time code, one of DLL_TC_STATUS_xxx
//

DWORD __stdcall MCR_ApiGetTimeCodeStatus(void)
{
	LAST_LOCK_STATUS srLockStatus;
	DWORD            dwReturnStatus;

	#ifdef _DEBUG_INFO
	static UINT8 ucSavedLockStatus;
	#endif

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetTimeCodeStatus[["));
	#endif

	AcquireApiMutex();

	srLockStatus = g_LastAlarmStatusResponse.srLockStatus;

	ReleaseApiMutex();
	
	dwReturnStatus = -1;

	switch (srLockStatus.ucLockStatus)
	{
		case LOCK_INVALID:
			#ifdef _DEBUG_INFO
			{
				if (ucSavedLockStatus != srLockStatus.ucLockStatus)
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("McrGet status: lock invalid"));
					cDebugOutput->DebugMsg(tcDebugBuff);

					ucSavedLockStatus = srLockStatus.ucLockStatus;
				}

			}
			#endif

			dwReturnStatus = DLL_TC_STATUS_INVALID;
			break;

		case LOCK_LOST:
			#ifdef _DEBUG_INFO
			{
				if (ucSavedLockStatus != srLockStatus.ucLockStatus)
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("McrGet status: lock lost"));
					cDebugOutput->DebugMsg(tcDebugBuff);

					ucSavedLockStatus = srLockStatus.ucLockStatus;
				}

			}
			#endif

			dwReturnStatus = DLL_TC_STATUS_LOST;
			break;

		case LOCK_RESTORED:
			#ifdef _DEBUG_INFO
			{
				if (ucSavedLockStatus != srLockStatus.ucLockStatus)
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("McrGet status: lock restored"));
					cDebugOutput->DebugMsg(tcDebugBuff);

					ucSavedLockStatus = srLockStatus.ucLockStatus;
				}

			}
			#endif

			dwReturnStatus = DLL_TC_STATUS_NORMAL;
			break;

		case LOCK_PRESENT:
			#ifdef _DEBUG_INFO
			{
				if (ucSavedLockStatus != srLockStatus.ucLockStatus)
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("McrGet status: present "));
					cDebugOutput->DebugMsg(tcDebugBuff);

					ucSavedLockStatus = srLockStatus.ucLockStatus;
				}

			}
			#endif

			dwReturnStatus = DLL_TC_STATUS_NORMAL;
			break;

		case LOCK_NOT_PRESENT:
		default:
			#ifdef _DEBUG_INFO
			{
				if (ucSavedLockStatus != srLockStatus.ucLockStatus)
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("McrGet status: lock not present: %d"),(int)srLockStatus.ucLockStatus);
					cDebugOutput->DebugMsg(tcDebugBuff);

					ucSavedLockStatus = srLockStatus.ucLockStatus;
				}

			}
			#endif

			dwReturnStatus = DLL_TC_STATUS_CALIB;
			break;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return dwReturnStatus;

}



//
//		Name:			MCR_ApiGetLockStatus()
//
//		Description:	Returns current lock status, includes an overall lock status and the lock status of each reference
//

void __stdcall MCR_ApiGetLockStatus(API_LAST_LOCK_STATUS *pLockStatus)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetLockStatus[["));
	#endif

	AcquireApiMutex();

	pLockStatus->ucLockStatus          = g_LastAlarmStatusResponse.srLockStatus.ucLockStatus;
	pLockStatus->ucRtcStatus           = g_LastAlarmStatusResponse.srLockStatus.ucRtcStatus;
	pLockStatus->ucGpsStatus           = g_LastAlarmStatusResponse.srLockStatus.ucGpsStatus;
	pLockStatus->ucTimeCodeStatus      = g_LastAlarmStatusResponse.srLockStatus.ucTimeCodeStatus;
	pLockStatus->ucHsoStatus           = g_LastAlarmStatusResponse.srLockStatus.ucHsoStatus;
	pLockStatus->ucHsoAchievementLevel = g_LastAlarmStatusResponse.srLockStatus.ucHsoAchievementLevel;
	
	pLockStatus->ucReserved1           = g_LastAlarmStatusResponse.srLockStatus.ucReserved1;
	pLockStatus->ucReserved2           = g_LastAlarmStatusResponse.srLockStatus.ucReserved2;
	pLockStatus->ucReserved3           = g_LastAlarmStatusResponse.srLockStatus.ucReserved3;
	pLockStatus->ucReserved4           = g_LastAlarmStatusResponse.srLockStatus.ucReserved4;

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

}

//
//		Name:			MCR_ApiGetDeviceStatus()
//
//		Description:	Returns current status of MCR device, one of HW_STATUS_xxx
//

DWORD __stdcall MCR_ApiGetDeviceStatus(void)
{
	// Read the status from the card

	UINT8	ucDeviceStatus;
	DWORD   dwStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetDeviceStatus[["));
	#endif

	AcquireApiMutex();

	ucDeviceStatus = g_LastAlarmStatusResponse.ucHwStatus;

	ReleaseApiMutex();

	if (ucDeviceStatus == API_HW_STATUS_NORMAL)
	{
		dwStatus = API_HW_STATUS_NORMAL;
	} else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!McrGet device status: bad status: %d"),(int)ucDeviceStatus);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		dwStatus = API_HW_STATUS_FAILED;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return dwStatus;
}

//
//		Name:			MCR_ApiSetFrameRate()
//
//		Description:	Sets sub-second frame rate for time code type being decoded.
//						Used to derive wMilliseconds member of MCR_TIME_REV_A
//
//		Returns:		TRUE if set successfully
//						FALSE if specified rate is invalid for current McrSync-configured time code type
//
//		Note: for MCR and later, this value is now ignored. The frame rate comes directly
//            from the card. Therefore, this function is deprecated.

BOOL __stdcall MCR_ApiSetFrameRate(WORD rate)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiSetFrameRate[["));
	#endif

	AcquireApiMutex();

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return(TRUE);
}


//
//		Name:			MCR_ApiPollTimeCode()
//
//		Description:	Polls the MCR for one time code frame.  This will be accurate to +/- one count
//						of the time code's frame rate.  Not recommended for high-precision timing.
//
//		Returns:		Time code status, one of DLL_TC_STATUS_xxx from INTAPI.H
//

DWORD __stdcall MCR_ApiPollTimeCode(PMCR_TIME_REV_A mcrTime)
{
	DWORD dwReturn;
	TIME_CODE_FRAME_REV_A	timeFrame;

	// Read the data from the card

	LARGE_INTEGER liTmp;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiPollTimeCode[["));
	#endif

	//AcquireApiMutex();
	QueryPerformanceCounter(&liTmp);

	dwReturn = ReadAndProcessTimeFrame(&timeFrame, &liTmp, mcrTime);

	if (dwReturn)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR_ApiPollTimeCode: populate fail: %d"), (int)dwReturn);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		ReleaseApiMutex();
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("]]"));
		#endif
		return DLL_TC_STATUS_DRVFAIL;
	}

	timeFrame.intPerfCount.QuadPart = liTmp.QuadPart;

	//ReleaseApiMutex();

	if ((timeFrame.ucLockStatus == LOCK_RESTORED) ||
		(timeFrame.ucLockStatus == LOCK_PRESENT))
	{
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("]]"));
		#endif
		return(DLL_TC_STATUS_NORMAL);
	}

	#ifdef _DEBUG_ERROR
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR_ApiPollTimeCode: lost lock: %d"), (int)timeFrame.ucLockStatus);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
	return DLL_TC_STATUS_LOST;

}


//
//		Name:			MCR_ApiEnableCallback()
//
//		Description:	Adds or creates asynchronous notification callback capability.  If the calling
//						thread has already registered a callback function, we add the requested
//						asynchronous notification to the current capacity.  Otherwise, we create a new
//						callback function node.
//
//		Returns:		On success, the new asynchronous notification mask
//						On failure, 0
//

DWORD __stdcall MCR_ApiEnableCallback(DWORD cbToAdd, PMCR_ASYNC_CALLBACK_ROUTINE cbFunction)
{
	PCALLBACK_NODE	cbNode, cbListTail;
	DWORD			curThreadId;
	DWORD			dwHooksMask;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiEnableCallback[["));
	#endif

	AcquireApiMutex();

	// only one thread in this process may manipulate the callback function list
	// at any one time.
	dwHooksMask = 0;

	__try
	{

		// synchronize with other threads

		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("Enable callback critical section start"));
		#endif
		EnterCriticalSection(&g_srDllShared.csCBListData);

		// find (or create) a callback node to manipulate

		curThreadId = GetCurrentThreadId();
		cbNode = g_cbListHead;

		while (cbNode && cbNode->cbThreadId != curThreadId)
		{
			cbListTail = cbNode;
			cbNode = (PCALLBACK_NODE)cbNode->next;
		}

		if (!cbNode)
		{
			// add a new callback function node for this process

			if (!g_cbListHead)
			{
				// the first node in our list

				g_cbListHead = (PCALLBACK_NODE)malloc(sizeof(CALLBACK_NODE));
				g_cbListHead->next = NULL;
				g_cbListHead->prev = NULL;
				cbNode = g_cbListHead;
			}
			else
			{
				// a new node at the end of the list

				cbNode = (PCALLBACK_NODE)malloc(sizeof(CALLBACK_NODE));
				cbNode->prev = cbListTail;
				cbListTail->next = cbNode;
			}

			// initialize the new node

			cbNode->cbHooksMask = 0;
			cbNode->locked = FALSE;

			cbNode->cbThreadId = curThreadId;

			DuplicateHandle(GetCurrentProcess(),
							GetCurrentThread(),
							GetCurrentProcess(),
							&cbNode->cbThreadHandle,
							NULL,
							FALSE,
							DUPLICATE_SAME_ACCESS);
		}

		if (!cbNode->cbHooksMask)
		{
			// this is the first callback request for calling thread, set callback function address

			cbNode->cbFunction = cbFunction;
		}

		// Add the hook to this call back, if it's not already there

		if (!(cbNode->cbHooksMask & cbToAdd))
		{

			cbNode->cbHooksMask |= cbToAdd;
			dwHooksMask = cbNode->cbHooksMask;

			// Increment the CB counters to determine the interrupt status beging

			fnAddCallbackEnableCounts(cbToAdd);
		}
	}

	__finally
	{
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("Enable callback critical section end"));
		#endif

		LeaveCriticalSection(&g_srDllShared.csCBListData);

	}

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
	return dwHooksMask;
}








//
//		Name:			MCR_ApiDisableCallback()
//
//		Description:	Disables one or more hooks to a thread's callback function.  When all hooks are
//						cleared the callback function is no longer called and is removed from the callback
//						list.
//
//		Returns:		On success, new hook mask
//						On failure, ASYNC_MASK_ERROR
//

DWORD __stdcall MCR_ApiDisableCallback(DWORD cbToRemove)
{
	DWORD			curThreadId;
	PCALLBACK_NODE	cbNode;
	DWORD			dwHooksMask;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiDisableCallback[["));
	#endif

	AcquireApiMutex();

	// only one thread in this process may manipulate the callback function list
	// at any one time.
	dwHooksMask = 0;

	__try
	{

		// synchronize with other threads

		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("Disable callback critical section start"));
		#endif
		EnterCriticalSection(&g_srDllShared.csCBListData);

		curThreadId = GetCurrentThreadId();
		cbNode = g_cbListHead;

		// Look for the node that matches the current thread

		while (cbNode && (cbNode->cbThreadId != curThreadId))
		{
			cbNode = (PCALLBACK_NODE)cbNode->next;
		}

		// If no node matches the thread, or there is no matching callback, then exit.

		if (!cbNode || !(cbNode->cbHooksMask & cbToRemove))
		{
			#ifdef _DEBUG_ERROR
			cDebugOutput->DebugMsg(_T("!!!MCR_ApiDisableCallback: callback not found"));
			#endif
			dwHooksMask = ASYNC_MASK_ERROR;
		}
		else
		{
			cbNode->cbHooksMask &= ~cbToRemove;

			if (!cbNode->cbHooksMask)
			{
				// all hooks have been disabled so remove this callback node from the list

				CloseHandle(cbNode->cbThreadHandle);

				// maintain head and list pointers

				if (cbNode == g_cbListHead)
					g_cbListHead = (PCALLBACK_NODE)cbNode->next;
				else
					((PCALLBACK_NODE)cbNode->prev)->next = cbNode->next;	// route around current node in list

				free(cbNode); // remove node
				cbNode = NULL;
			}

			if (cbNode)
				dwHooksMask = cbNode->cbHooksMask;
			else
				dwHooksMask = 0;

			// Decrememnt the callback counters (which disables the interrupt if necessary)
			fnRemoveCallbackEnableCounts(cbToRemove);
		}
	}

	__finally
	{
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("Disable callback critical section end"));
		#endif

		LeaveCriticalSection(&g_srDllShared.csCBListData);
	}

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
	return dwHooksMask;
}


//
//		Name:			MCR_ApiShutdown()
//
//		Description:	Should be called before a _process_ (using an instance of the DLL)
//						terminates.  Allows DLL to gracefully terminate threads and deallocate
//						resources on behalf of the calling process.
//                      Deprecated-now does nothing

void __stdcall MCR_ApiShutdown(void)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiShutdown[["));
	#endif

	//AcquireApiMutex();

	//ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
	return;	// now this does nothing
}


//
//		Name:			MCR_ApiClearEnhanced()
//
//		Description:	Clears enhanced features (once/frame and once/second interrupts).
//
//		Note:			This routine violates the multiprocess/multithreaded premise of this DLL (which never worked).
//						A better mechanism should be implemented eventually.
//
//		Returns:		TRUE on completion
//

BOOL __stdcall MCR_ApiClearEnhanced(void)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiClearEnhanced[["));
	#endif

	AcquireApiMutex();
	MCR_ApiDisableCallback(ASYNC_ENDTC);
	MCR_ApiDisableCallback(ASYNC_STARTTC);
	MCR_ApiDisableCallback(ASYNC_PERSEC);
	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
	return TRUE;
}


//
//		Name:			MCR_ApiRequestRecalibration()
//
//		Description:	Initiates a time code signal level calibration.
//
//		Returns:		TRUE upon completion, or FALSE if not available or an error
//

BOOL __stdcall MCR_ApiRequestRecalibration(void)
{
	//RETURN_CODE rc;
	UINT8 ucValue = 0;
	INT iStatus;
	BOOL bStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiRequestRecalibration[["));
	#endif

	//AcquireApiMutex();

	iStatus = fnRecalibrate();
	if (iStatus == MCR_SUCCESS)
	{
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("MCR_ApiRequestRecalibration: recalibration started"));
		#endif
		//ReleaseApiMutex();
		bStatus = TRUE;
	} else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR_ApiRequestRecalibration: recalibration request failed: %d"), (int)iStatus);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
		//ReleaseApiMutex();
		bStatus = FALSE;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return bStatus;
}






//
//		Name:			MCR_ApiSetDateTime()
//
//		Description:	Sets the date and time in the MCR PCIe.
//
//		Returns:		TRUE upon completion, or FALSE if not available or an error
//

BOOL __stdcall MCR_ApiSetDateTime(MCRSYNC_IPC_SETTIME *psrSetTime)
{
	//RETURN_CODE rc;
	UINT8 ucValue = 0;
	INT iStatus;
	BOOL bStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiSetDateTime[["));
	#endif

	//AcquireApiMutex();

	iStatus = fnSetDateTime(psrSetTime);
	if (iStatus == MCR_SUCCESS)
	{
		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("MCR_ApiSetDateTime: Completed"));
		#endif
		//ReleaseApiMutex();
		bStatus = TRUE;
	} else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR_ApiSetDateTime request failed: %d"), (int)iStatus);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		//ReleaseApiMutex();
		bStatus = FALSE;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return bStatus;
}




//
//		Name:			MCR_ApiGetTimeCodeType()
//
//		Description:	Returns time code type currently in use/configured by McrSync
//
//		Returns:		One of TC_TYPE_xxx defined in INTAPI.H
//

DWORD __stdcall MCR_ApiGetTimeCodeType(void)
{
	// since this function calls another API function which uses the API mutex,
	// no APi mutex calls are required in this function

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetTimeCodeType[["));
	#endif

	MCR_ApiGetTimeCodeStatus(); // populates g_dwTcType

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return(g_dwTcType);
}


//
//		Name:			MCR_ApiGetTimeCodeGpsStatus()
//
//		Description:	Returns Time code lock status if SMPTE, Leitch date encoding, with Masterclock, Inc proprietary control bits
//
//		Returns:		One of GPS_STATUS_xxx defined in INTAPI.H
//

DWORD __stdcall MCR_ApiGetTimeCodeGpsStatus(void)
{
	DWORD dwTimeCodeGpsStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetTimeCodeGpsStatus[["));
	#endif

	AcquireApiMutex();
	dwTimeCodeGpsStatus = g_devInfo.dwTimeCodeGpsStatus;
	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return(dwTimeCodeGpsStatus);
}


//
//		Name:			MCR_ApiGetDLLVersion()
//
//		Description:	Returns DLL version number
//

void __stdcall MCR_ApiGetDLLVersion(PUINT verMajor, PUINT verMinor, PUINT verRevision)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetDLLVersion[["));
	#endif

	AcquireApiMutex();
	if (verMajor)
		*verMajor = DLL_VERSION_MAJOR;

	if (verMinor)
		*verMinor = DLL_VERSION_MINOR;

	if (verRevision)
		*verRevision = DLL_VERSION_REVISION;

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
}

//
//		Name:			MCR_ApiGetDriverVersion()
//
//		Description:	There is the firmware version of the AVR.
//
//		Returns:		TRUE if DLL initialized correctly
//

BOOL __stdcall MCR_ApiGetDriverVersion(PUINT verMajor, PUINT verMinor, PUINT verRevision)
{
	UINT8	ucDeviceStatus;
	BOOL    bStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetDriverVersion[["));
	#endif

	AcquireApiMutex();

	ucDeviceStatus = g_LastAlarmStatusResponse.ucHwStatus;

	ReleaseApiMutex();

	if (ucDeviceStatus == API_HW_STATUS_NORMAL)
	{
		*verMajor    = g_LastAlarmStatusResponse.verMajor;
		*verMinor    = g_LastAlarmStatusResponse.verMinor;
		*verRevision = g_LastAlarmStatusResponse.verRevision;
		bStatus = API_HW_STATUS_NORMAL;
	} else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!McrGet driver version: bad status: %d"),(int)ucDeviceStatus);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		bStatus = HW_STATUS_FAILED;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return bStatus;
}




//
//		Name:			MCR_ApiSetConfiguration()
//
//		Description:	Sets the configuration and saves it in the registry.
//
//		Returns:		0 if successful
//                      else 32-bit bitmap indicating which parameters were invalid
//
UINT32 __stdcall MCR_ApiSetConfiguration(MCR_CONFIGURATION *pApiSetConfig)
{
	UINT32 ulretval;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiSetConfiguration[["));
	#endif

	AcquireApiMutex();

	CSoftwareConfig	 *pcSaveConfig = new CSoftwareConfig(cDebugOutput);

	ulretval = pcSaveConfig->SaveConfigFromAPI(pApiSetConfig);

	if (!ulretval)
	{
		pcSaveConfig->SaveConfigToRegistry();

		// save the configuration to the AVR
		fnSetAllAvrConfiguration();
	}

	delete pcSaveConfig;

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return ulretval;
}




//
//		Name:			MCR_ApiGetConfiguration()
//
//		Description:	Gets the current configuration.
//
//		Returns:		0 if successful
//                      else 32-bit bitmap indicating which parameters were invalid
//
void __stdcall MCR_ApiGetConfiguration(MCR_CONFIGURATION *pApiSetConfig)
{
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetConfiguration[["));
	#endif

	AcquireApiMutex();
	CSoftwareConfig	 *pcGetConfig = new CSoftwareConfig(cDebugOutput);

	pcGetConfig->GetConfigForAPI(pApiSetConfig);

	delete pcGetConfig;

	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif
}








//
//		Name:			MCR_ApiGetFirmwareVersion()
//
//		Description:	Returns MCR firmware version
//
//		Returns:		TRUE if value retrieved successfully, FALSE if not
//
BOOL __stdcall MCR_ApiGetFirmwareVersion(PUINT verMajor, PUINT verMinor, PUINT verRevision)
{
	UINT8	ucDeviceStatus;
	BOOL bStatus;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetFirmwareVersion[["));
	#endif

	AcquireApiMutex();

	ucDeviceStatus = g_LastAlarmStatusResponse.ucHwStatus; // kpw trace - this is filled in by fnProcessAVR_NOTIFY_ALARM().

	ReleaseApiMutex();

	if (ucDeviceStatus == API_HW_STATUS_NORMAL)
	{
		// kjf xxx yyy zzz need to add returning configuration
		*verMajor    = g_LastAlarmStatusResponse.verMajor;
		*verMinor    = g_LastAlarmStatusResponse.verMinor;
		*verRevision = g_LastAlarmStatusResponse.verRevision;
		bStatus = API_HW_STATUS_NORMAL;
	} else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!McrGet firmware version: bad status: %d"),(int)ucDeviceStatus);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		bStatus = API_HW_STATUS_FAILED;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return bStatus;
}

//
//		Name:			MCR_ApiGetFirmwareChecksum()
//
//		Description:	Returns MCR firmware checksum
//                      Deprecated, always returns 0.
//
//		Returns:		TRUE if value retrieved successfully, FALSE if not
//

BOOL __stdcall MCR_ApiGetFirmwareChecksum(PDWORD pChecksum)
{

	DWORD		checksum = 0x0;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("MCR_ApiGetFirmwareChecksum[["));
	#endif

	AcquireApiMutex();
	*pChecksum = checksum;     // Deprecated, always returns 0.
	ReleaseApiMutex();

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("]]"));
	#endif

	return TRUE;
}
