

#include <stdlib.h>
#include <malloc.h>
#include "IntDLL.h"
#include "tzlib.h"
#include "McrSerialInterface.h"
#include "McrSerialProtocol.h"
#include "DebugOutput.h"




//MC_DLL_SHARED g_srDllShared = {0};

// interprocess DLL instance shared data

#pragma data_seg(".shared")

// Sharing is not working as expected!!! When loaded, the DLL gives each
// process it's own copy of the shared memory. ie. sharing is doing nothing
MC_DLL_SHARED g_srDllShared = {0};

#pragma data_seg()

//#pragma comment(linker, "/section:.shared,RWS") // this line is not needed, see the .def file
//#pragma comment(linker, "/section:.data,RW")    // this line is not needed, see the .def file


// private DLL functions


static DWORD InitDLLInstance(CDebugOutput *cDebugOutput);
static DWORD GetPollingTimeframe(PTIME_CODE_FRAME_REV_A pTimeFrame);


// private DLL data members


DEVICE_INFO		g_devInfo;
DWORD			g_dllInitStatus = DLL_INIT_UNEXPECTED;
DWORD			g_dwTcType      = TC_TYPE_UNDEFINED;
PCALLBACK_NODE	g_cbListHead    = NULL;

DWORD			SerialEventMonitorId;

CDebugOutput *cDebugOutput;

BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
	//DWORD			curThreadId;
	PCALLBACK_NODE	cbNode, cbPrevNode;
	//HINSTANCE      *indirect;
	//boolean         bIsSuccessful;

	//#ifdef _DEBUG_ERROR
	//{
	//	TCHAR tcDebugBuff[256];
	//	swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MC API DLL main: %d"), (int)dwReason);
	//		cDebugOutput->DebugMsg(tcDebugBuff);
	//}
	//#endif

	switch (dwReason)
    {
		case DLL_PROCESS_ATTACH:
			// called when a new application (process) first uses the DLL

			// this pointer is located in the shared data area.
			// We only do the load one time
			if (g_srDllShared.dwDllIsPresent != DLL_IS_PRESENT)
			{
				// We create the class in case something tries to use it.
				// Until DebugInit is called, no output file is opened and no data is output
				cDebugOutput = new CDebugOutput();

				#if defined(_DEBUG_OUTPUT_ENABLE)
				cDebugOutput->DebugInit(_T("MCR.DLL"),_T("C:\\Temp\\MCRDEBUG.TXT"),DEBUG_FILE);
				#endif

				#ifdef _DEBUG_ERROR
				cDebugOutput->DebugMsg(_T("!!!MCR DLL loaded"));
				#endif

				g_srDllShared.dwDllIsPresent = DLL_IS_PRESENT;

				g_dllInitStatus = InitDLLInstance(cDebugOutput);   // should this only be ran once?
			} else
			{
				#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!MCR DLL was already loaded successfully"));
				#endif
			}

			//if (!g_srDllShared.hMcrDllInstanceLib)
			//{
			//	#ifdef _DEBUG_OUTPUT_ENABLE
			//	DebugInit("MCR.DLL", "C:\\TEMP\\MCRDEBUG.TXT", DEBUG_FILE);
			//	#endif

			//	//indirect = &g_srDllShared.hMcrDllInstanceLib;

			//	//bIsSuccessful = fnLoadPlxDll(indirect);
			//} else
			//{
			//	#ifdef _DEBUG_INIT
			//		DebugMsg("MCR DLL was already loaded successfully");
			//	#endif
			//}

			InitializeCriticalSection(&g_srDllShared.csCBListData);

			// we can only run the init if the PLX DLL was successfully loaded
			//if (g_srDllShared.hMcrDllInstanceLib)
			//{
			//	g_dllInitStatus = InitDLLInstance();
			//} else
			//{
			//	#ifdef _DEBUG_ERROR
			//	DebugMsg("!!!MCR DLL not loaded, could not run init instance");
			//	#endif
			//}

			#ifdef _DEBUG_INFO
			cDebugOutput->DebugMsg(_T("MCR DLL process attach critical section start"));
			#endif
			EnterCriticalSection(&g_srDllShared.csCBListData);

			g_srDllShared.dllInstanceCnt++;

			#ifdef _DEBUG_INFO
			cDebugOutput->DebugMsg(_T("MCR DLL process attach critical section end"));
			#endif
			LeaveCriticalSection(&g_srDllShared.csCBListData);


			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!DLL_PROCESS_ATTACH - DLL instance count: %d"),(int)g_srDllShared.dllInstanceCnt);
					cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif

			break;

		case DLL_PROCESS_DETACH:
			// called when an application (process) using the DLL terminates, note that all threads
			// spawned by the DLL have already been terminated (by the OS)

			// check callback list, and remove all the callbacks for
			// all threads that are left

			#ifdef _DEBUG_ERROR
			cDebugOutput->DebugMsg(_T("!!!DLL process detach critical section start"));
			#endif

			EnterCriticalSection(&g_srDllShared.csCBListData);

			cbNode = g_cbListHead;

			while (cbNode)
			{
				// delete this callback node since it will no longer be needed
				#ifdef _DEBUG_ERROR
				cDebugOutput->DebugMsg(_T("!!!MCR dll Deleting a callback node because process terminated w/o disabling its callback."));
				#endif

				// Adjust the expanded interrupt counts, based on removing
				// the callbacks associated with this thread

				fnRemoveCallbackEnableCounts(cbNode->cbHooksMask);

				cbPrevNode = cbNode;

				cbNode = (PCALLBACK_NODE)cbNode->next;

				free(cbPrevNode); // remove node
			}

			g_cbListHead = NULL;	// All modes deleted

			g_srDllShared.dllInstanceCnt--;

			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR DLL_PROCESS_DETACH - DLL instance count: %lu"), (unsigned long)g_srDllShared.dllInstanceCnt);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif

			if (g_srDllShared.dllInstanceCnt == 0)
			{
				// Close the device


				#ifdef _DEBUG_ERROR
				cDebugOutput->DebugMsg(_T("!!!MCR DLL will be leaving memory."));
				#endif
				if (g_srDllShared.hMcrDllInstanceLib)
				{
					//PlxPci_DeviceCloseDynamic(&hPlxDevice);

					//fnFreePlxDll(g_srDllShared.hMcrDllInstanceLib);
					g_srDllShared.hMcrDllInstanceLib = NULL;
				}

				CloseSerialEvents();

			}

			#ifdef _DEBUG_INFO
			cDebugOutput->DebugMsg(_T("MCR DLL process detach critical section end"));
			#endif
			LeaveCriticalSection(&g_srDllShared.csCBListData);

			break;

        case DLL_THREAD_ATTACH:

			// !!! for an unknown reason, DLL_THREAD_ATTACH will get called during normal execution after
			// all initialization is completed. We ignore this state.


			//#ifdef _DEBUG_ERROR
			//cDebugOutput->DebugMsg(_T("!!!MCR DLL_THREAD_ATTACH"));
			//#endif

			break;

        case DLL_THREAD_DETACH:

			// !!! for an unknown reason, DLL_THREAD_DETACH will get called during normal execution after
			// all initialization is completed. We ignore this state.


			// obtain the callback list critical section
			// before doing any access/manipulation

			//#ifdef _DEBUG_INFO
			//cDebugOutput->DebugMsg(_T("MCR DLL_THREAD_DETACH critical section start"));
			//#endif

			//EnterCriticalSection(&g_srDllShared.csCBListData);


			//cbNode = g_cbListHead;
			//curThreadId = GetCurrentThreadId();

			//while (cbNode)
			//{
			//	if (cbNode->cbThreadId == curThreadId)
			//	{
			//		// delete this callback node since it will no longer be needed
			//		#ifdef _DEBUG_ERROR
			//		cDebugOutput->DebugMsg(_T("!!!MCR dll Deleting a callback node because thread terminated w/o disabling its callback."));
			//		#endif

			//		if (cbNode == g_cbListHead)
			//			g_cbListHead = (PCALLBACK_NODE)cbNode->next;
			//		else
			//			((PCALLBACK_NODE)cbNode->prev)->next = cbNode->next;	// route around current node in list

			//		// Adjust the expanded interrupt counts, based on removing
			//		// the callbacks associated with this thread

			//		fnRemoveCallbackEnableCounts(cbNode->cbHooksMask);

			//		free(cbNode); // remove node

			//		// break because a thread can have only one node in our list
			//		break;
			//	}
			//	else
			//	{
			//		#ifdef _DEBUG_INFO
			//		cDebugOutput->DebugMsg(_T("MCR DLL_THREAD_DETACH node thread did not match"));
			//		#endif
			//		cbNode = (PCALLBACK_NODE)cbNode->next;
			//	}
			//}

			//#ifdef _DEBUG_INFO
			//cDebugOutput->DebugMsg(_T("MCR DLL_THREAD_DETACH critical section end"));
			//#endif

			//LeaveCriticalSection(&g_srDllShared.csCBListData);

			//#ifdef _DEBUG_ERROR
			//cDebugOutput->DebugMsg(_T("!!!MCR DLL_THREAD_DETACH done"));
			//#endif
			break;

		default:
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!MCR Unexpected DLLEntryPoint() call with dwReason: %lX"),(unsigned long)dwReason);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
            break;
     }

    return(TRUE);
}


//#define MAX_DEVICES_TO_LIST 50 // 50 is some large number we will never exceed
//
//		Name:			InitDLLInstance()
//
//		Description:	On behalf of each application that uses the DLL perform
//						these one-time initializations
//

DWORD InitDLLInstance(CDebugOutput *cDebugOutput)
{

	// kjf xxx yyy zzz plx was initialized here, should the serial port be done here?


	// Create the Mutexes necessary for controlling access to the shared data.
	// Since the Mutex could be initially created by a privileged process
	// (like McrSync) that won't allow other processes to access the Mutex,
	// we need SECURITY_ATTRIBUTES and a SECURITY_DESCRIPTOR that will allow
	// all process access to the Mutex. Allocating the SECURITY_DESCRIPTOR assures
	// that the structure will be on the correct word boundary for the support
	// functions to work.

    SECURITY_ATTRIBUTES secAttr;
    PSECURITY_DESCRIPTOR pSecDesc = malloc(sizeof(SECURITY_DESCRIPTOR));
    if (pSecDesc == 0)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!MCR InitDLLInstance: malloc failed"));
		#endif

		//PlxPci_DeviceCloseDynamic(&hPlxDevice);
		return DLL_INIT_UNEXPECTED;
	}

	secAttr.nLength = sizeof(secAttr);
	secAttr.lpSecurityDescriptor = (LPVOID) pSecDesc;
	secAttr.bInheritHandle = FALSE; // child processes do not inherit the handle

	InitializeSecurityDescriptor(pSecDesc, SECURITY_DESCRIPTOR_REVISION);

	// Set the "discretionary ACL (access control list)" to provide all access to
	// the object
	SetSecurityDescriptorDacl(pSecDesc, TRUE, NULL, TRUE);


	// Create the mutex that controls access to the CB counts

	g_srDllShared.hCBCountMutex = CreateMutex(&secAttr, FALSE, _T("Mutex_MC_CBCount"));
	if (g_srDllShared.hCBCountMutex == NULL)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!MCR InitDLLInstance: create mutex g_srDllShared.hCBCountMutex failed"));
		#endif

		//PlxPci_DeviceCloseDynamic(&hPlxDevice);
		free(pSecDesc);
		return DLL_INIT_UNEXPECTED;
	}

	// Create the mutex that controls to the LastInt data

	g_srDllShared.hLastIntMutex = CreateMutex(&secAttr, FALSE, _T("Mutex_MC_LastInt"));
	if (g_srDllShared.hLastIntMutex == NULL)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!MCR InitDLLInstance: create mutex g_srDllShared.hLastIntMutex failed"));
		#endif

		//PlxPci_DeviceCloseDynamic(&hPlxDevice);
		free(pSecDesc);
		return DLL_INIT_UNEXPECTED;
	}

	free(pSecDesc); // no longer needed

	// Create the critical section object that controls the access
	// to the callback data.

	InitializeCriticalSection(&g_srDllShared.csCBListData);

	// Create the fnSerialEventMonitorThread for the PCI interrupt

	if (CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)fnSerialEventMonitorThread, (LPVOID)cDebugOutput, 0, &SerialEventMonitorId) == NULL)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!InitDLLInstance: create fnSerialEventMonitorThread failed"));
		#endif

		//PlxPci_DeviceCloseDynamic(&hPlxDevice);
		return DLL_INIT_UNEXPECTED;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("InitDLLInstance completed successfully"));
	#endif

	return DLL_INIT_OK;
}

//
//		Name:			AcquireApiMutex()
//
//		Description:	Waits for the API mutex. If the wait fails, just continue.
//                      It is possible that if a process has acquired the mutex, but, before the process
//                      can free the mutex it is deleted, the mutex would forever be allocated. Instead,
//                      we continue after the timeout.
//
//						We cannot use a true mutex (ie. CreateMutex() and WaitForSingleObject() )
//						Simulate one using this flag.
//
#define MUTEX_RETRY_COUNT 30
#define MUTEX_RETRY_MSEC  10
void AcquireApiMutex(void)
{
	LONG lPriorValue;

	int ii;

	// simulate a 300 msec timeout
	for (ii=0; ii<MUTEX_RETRY_COUNT; ii++)
	{
		if (!g_srDllShared.ldllApiMutex)
		{
			lPriorValue = ::InterlockedExchange(&g_srDllShared.ldllApiMutex, 1);

			if (!lPriorValue)
			{
				// It was a zero and we set it high, we can continue
				break;
			}

			// Another thread beat us to setting it to a one.
			// Wait for that thread to zero the value

		}

		Sleep(MUTEX_RETRY_MSEC);
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!acquiring DLL mutex retry"));
		#endif
	}

	if (ii >= MUTEX_RETRY_COUNT)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!Failed acquiring DLL mutex"));
		#endif
	}
}


//
//		Name:			ReleaseApiMutex()
//
//		Description:	Releases the API mutex.
//
//						We cannot use a true mutex (ie. CreateMutex() and WaitForSingleObject() )
//						Simulate one using this flag.
//

void ReleaseApiMutex(void)
{
	//HANDLE hProcessID;
	//DWORD  dwPriority;

	//hProcessID = GetCurrentProcess();
	//dwPriority = GetPriorityClass(hProcessID);
	//SetPriorityClass(hProcessID,REALTIME_PRIORITY_CLASS);

	::InterlockedExchange(&g_srDllShared.ldllApiMutex, 0);

	//SetPriorityClass(hProcessID,dwPriority);

}













//
//		Name:			GetTCTypeFromTCStatus()
//
//		Description:	Converts the time code status from the device to a time code type.
//
//
//DWORD GetTCTypeFromTCStatus(UINT8 tcStatus)
//{
//	// Set the time code type
//	switch (tcStatus & TIME_CODE_TYPE_MASK) {
//	case SMPTE_30:
//	case SMPTE_25:
//	case SMPTE_24:
//		return TC_TYPE_SMPTE;
//	case IRIG_B_AM:
//		return TC_TYPE_IRIG_B1;
//	case IRIG_B_PWM:
//		return TC_TYPE_IRIG_B;
//	default:
//		return TC_TYPE_UNDEFINED;
//	}
//}

//
//		Name:			fnAddCallbackEnableCounts
//
//		Description:	Adjusts the callback count based on the ending of a thread.
//                      The callbacks can be enabled by multiple sources. We keep a count of
//                      how many times a callback is enabled. When the callback is no longer used
//                      by any source we disable it.
//
//		Returns 0 is success, or ASYNC_MASK_ERROR if there is an error.
//

DWORD fnAddCallbackEnableCounts(DWORD cbToAdd)
{
	UINT8		intsToSet = 0;
	DWORD	dwReturn = MCR_SUCCESS;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() started."));
	#endif

	DWORD dwWaitReason = WaitForSingleObject(g_srDllShared.hCBCountMutex, 1000L);
	if (dwWaitReason == WAIT_OBJECT_0) {

		#ifdef _DEBUG_INFO
		cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() received mutex"));
		#endif

		__try {

			switch (cbToAdd) {

			case ASYNC_PERSEC:
				//if (g_srDllShared.dllCallbackPerSecEnableCnt == 0) {
				//	dwReturn = EnableMcrInterrupt(PER_SECOND_ALARM_ENABLE);

				//	if (dwReturn == API_SUCCESS)
				//	{
				//		#ifdef _DEBUG_INFO
				//		cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_PERSEC successfully added"));
				//		#endif
				//	} else
				//	{
				//		#ifdef _DEBUG_ERROR
				//		{
				//			TCHAR tcDebugBuff[256];
				//			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_PERSEC failed: %d"), (int)dwReturn);
				//			cDebugOutput->DebugMsg(tcDebugBuff);
				//		}
				//		#endif
				//	}
				//} else
				//{
				//	#ifdef _DEBUG_ERROR
				//	cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_PERSEC g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
				//	#endif
				//}
				g_srDllShared.dllCallbackPerSecEnableCnt++;
				break;

			case ASYNC_STARTTC:
				//if (g_srDllShared.dllCallbackStartFrameEnableCnt == 0) {
				//	dwReturn = EnableMcrInterrupt(FRAME_START_ALARM_ENABLE);

				//	if (dwReturn == API_SUCCESS)
				//	{
				//		#ifdef _DEBUG_INFO
				//		cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_STARTTC successfully added"));
				//		#endif
				//	} else
				//	{
				//		#ifdef _DEBUG_ERROR
				//		{
				//			TCHAR tcDebugBuff[256];
				//			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_STARTTC failed: %d"), (int)dwReturn);
				//			cDebugOutput->DebugMsg(tcDebugBuff);
				//		}
				//		#endif
				//	}
				//} else
				//{
				//	#ifdef _DEBUG_ERROR
				//	cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_STARTTC g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
				//	#endif
				//}

				g_srDllShared.dllCallbackStartFrameEnableCnt++;
				break;

			case ASYNC_ENDTC:
				if (g_srDllShared.dllCallbackEndFrameEnableCnt == 0) {
					//dwReturn = EnableMcrInterrupt(FRAME_END_ALARM_ENABLE);

					//if (dwReturn == API_SUCCESS)
					//{
					//	#ifdef _DEBUG_INFO
					//	cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_ENDTC successfully added"));
					//	#endif
					//} else
					//{
					//	#ifdef _DEBUG_ERROR
					//	{
					//		TCHAR tcDebugBuff[256];
					//		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_ENDTC failed: %d"), (int)dwReturn);
					//		cDebugOutput->DebugMsg(tcDebugBuff);
					//	}
					//	#endif
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_ENDTC g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
					#endif
				}

				g_srDllShared.dllCallbackEndFrameEnableCnt++;
				break;

			case ASYNC_TIMESET:
				if (g_srDllShared.dllCallbackPerMinEnableCnt == 0) {
					//dwReturn = EnableMcrInterrupt(MINUTE_ALARM_ENABLE);

					//if (dwReturn == API_SUCCESS)
					//{
					//	#ifdef _DEBUG_INFO
					//	cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_TIMESET successfully added"));
					//	#endif
					//} else
					//{
					//	#ifdef _DEBUG_ERROR
					//	{
					//		TCHAR tcDebugBuff[256];
					//		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMESET failed: %d"), (int)dwReturn);
					//		cDebugOutput->DebugMsg(tcDebugBuff);
					//	}
					//	#endif
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMESET g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
					#endif
				}

				g_srDllShared.dllCallbackPerMinEnableCnt++;
				break;

			case ASYNC_TIMEGAIN:
				if (g_srDllShared.dllCallbackTCDetectEnableCnt == 0) {
					//dwReturn = EnableMcrInterrupt(LOCK_GAIN_ALARM_ENABLE);

					//if (dwReturn == API_SUCCESS)
					//{
					//	#ifdef _DEBUG_INFO
					//	cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_TIMEGAIN successfully added"));
					//	#endif
					//} else
					//{
					//	#ifdef _DEBUG_ERROR
					//	{
					//		TCHAR tcDebugBuff[256];
					//		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMEGAIN failed: %d"), (int)dwReturn);
					//		cDebugOutput->DebugMsg(tcDebugBuff);
					//	}
					//	#endif
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMEGAIN g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
					#endif
				}

				g_srDllShared.dllCallbackTCDetectEnableCnt++;
				break;

			case ASYNC_TIMELOSS:
				if (g_srDllShared.dllCallbackTCLossEnableCnt == 0) {
					//dwReturn = EnableMcrInterrupt(LOCK_LOSS_ALARM_ENABLE);

					//if (dwReturn == API_SUCCESS)
					//{
					//	#ifdef _DEBUG_INFO
					//	cDebugOutput->DebugMsg(_T("fnAddCallbackEnableCounts() ASYNC_TIMELOSS successfully added"));
					//	#endif
					//} else
					//{
					//	#ifdef _DEBUG_ERROR
					//	{
					//		TCHAR tcDebugBuff[256];
					//		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMELOSS failed: %d"), (int)dwReturn);
					//		cDebugOutput->DebugMsg(tcDebugBuff);
					//	}
					//	#endif
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!fnAddCallbackEnableCounts() ASYNC_TIMELOSS g_srDllShared.dllCallbackPerSecEnableCnt not zero, ignore add"));
					#endif
				}

				g_srDllShared.dllCallbackTCLossEnableCnt++;
				break;

			default:
				dwReturn = MCR_FAIL;
				#ifdef _DEBUG_ERROR
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() unknown callback: %d"), (int)cbToAdd);
					cDebugOutput->DebugMsg(tcDebugBuff);
				}
				#endif
				break;
			}
		}
		__finally {
			#ifdef _DEBUG_INFO1
						cDebugOutput->DebugMsg(_T("Add %d %d %d %d %d %d"),
									g_srDllShared.dllCallbackPerSecEnableCnt, g_srDllShared.dllCallbackStartFrameEnableCnt, g_srDllShared.dllCallbackEndFrameEnableCnt,
									g_srDllShared.dllCallbackPerMinEnableCnt, g_srDllShared.dllCallbackTCDetectEnableCnt, g_srDllShared.dllCallbackTCLossEnableCnt);
			#endif
			ReleaseMutex(g_srDllShared.hCBCountMutex);
		}
	}
	else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnAddCallbackEnableCounts() mutex not received: %ld"), (unsigned long)dwWaitReason);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
	}


	if (dwReturn == MCR_SUCCESS)
	{
		return ASYNC_OKAY;
	}
	else
	{
		return ASYNC_MASK_ERROR;
	}
}

//
//		Name:			fnRemoveCallbackEnableCounts
//
//		Description:	Adjusts the callback count based on removing the callbacks
//						specified in the cbToRemove parameter.
//
//		Returns the interrupts left, or ASYNC_MASK_ERROR if there is an error.
//

DWORD fnRemoveCallbackEnableCounts(DWORD cbToRemove)
{
	UINT8		intsToSet = 0;
	DWORD	dwReturn = MCR_SUCCESS;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("fnRemoveCallbackEnableCounts() started"));
	#endif

	DWORD dwWaitReason = WaitForSingleObject(g_srDllShared.hCBCountMutex, 1000L);
	if (dwWaitReason == WAIT_OBJECT_0) {

		__try {

			if (cbToRemove & ASYNC_PERSEC) {
				if (g_srDllShared.dllCallbackPerSecEnableCnt)
				{
					g_srDllShared.dllCallbackPerSecEnableCnt--;
					//if (g_srDllShared.dllCallbackPerSecEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(PER_SECOND_ALARM);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_PERSEC specified, but no callback found"));
					#endif
				}
			}

			if (cbToRemove & ASYNC_STARTTC) {
				if (g_srDllShared.dllCallbackStartFrameEnableCnt)
				{
					g_srDllShared.dllCallbackStartFrameEnableCnt--;
					//if (g_srDllShared.dllCallbackStartFrameEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(FRAME_START_ALARM);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_STARTTC specified, but no callback found"));
					#endif
				}
			}

			if (cbToRemove & ASYNC_ENDTC) {
				if (g_srDllShared.dllCallbackEndFrameEnableCnt)
				{
					g_srDllShared.dllCallbackEndFrameEnableCnt--;
					//if (g_srDllShared.dllCallbackEndFrameEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(FRAME_END_ALARM);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_ENDTC specified, but no callback found"));
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMESET) {
				if (g_srDllShared.dllCallbackPerMinEnableCnt)
				{
					g_srDllShared.dllCallbackPerMinEnableCnt--;
					//if (g_srDllShared.dllCallbackPerMinEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(MINUTE_ALARM);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_TIMESET specified, but no callback found"));
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMEGAIN) {
				if (g_srDllShared.dllCallbackTCDetectEnableCnt)
				{
					g_srDllShared.dllCallbackTCDetectEnableCnt--;
					//if (g_srDllShared.dllCallbackTCDetectEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(MCR_WRITE_TC_GAIN_INT_ENABLE);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_TIMEGAIN specified, but no callback found"));
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMELOSS) {
				if (g_srDllShared.dllCallbackTCLossEnableCnt)
				{
					g_srDllShared.dllCallbackTCLossEnableCnt--;
					//if (g_srDllShared.dllCallbackTCLossEnableCnt == 0) {
					//	dwReturn = DisableMcrInterrupt(MCR_WRITE_TC_LOSS_INT_ENABLE);
					//}
				} else
				{
					#ifdef _DEBUG_ERROR
					cDebugOutput->DebugMsg(_T("!!!Remove callback ASYNC_TIMELOSS specified, but no callback found"));
					#endif
				}
			}
		}

		__finally {
			//#ifdef _DEBUG_ERROR
			//			cDebugOutput->DebugMsg(_T("!!!Rem %d %d %d %d %d %d")),
			//						(int)g_srDllShared.dllCallbackPerSecEnableCnt, (int)g_srDllShared.dllCallbackStartFrameEnableCnt, (int)g_srDllShared.dllCallbackEndFrameEnableCnt,
			//						(int)g_srDllShared.dllCallbackPerMinEnableCnt, (int)g_srDllShared.dllCallbackTCDetectEnableCnt, (int)g_srDllShared.dllCallbackTCLossEnableCnt);
			//#endif
			ReleaseMutex(g_srDllShared.hCBCountMutex);
		}
	}
	else
	{
		dwReturn = MCR_FAIL;
	}

	if (dwReturn == MCR_SUCCESS)
	{
		return ASYNC_OKAY;
	}
	else
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnRemoveCallbackEnableCounts() failed:%d"), (int)dwReturn);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
		return ASYNC_MASK_ERROR;
	}

}














