

#include <stdlib.h>
#include <malloc.h>
#include "IntDLL.h"
#include "tzlib.h"
#include "McrSerialInterface.h"
#include "McrSerialProtocol.h"
#include "DebugOutput.h"


#define ASYNC_TIMING_INTS	(ASYNC_TIMESET | ASYNC_PERSEC | ASYNC_STARTTC | ASYNC_ENDTC)

extern CDebugOutput *cDebugOutput;

HANDLE		g_hSerialEventTimeCodeFrameResponse;
HANDLE		g_hSerialEventTimeCodeFrameAlarm;
HANDLE		g_hSerialEventTopOfSecondStart;
HANDLE		g_hSerialEventTopOfSecondData;
HANDLE		g_hSerialEventResponse;

int g_cCallbackCount;

CMcrSerialInterface *pCMcrSerialInterface;
char                 g_cSerialWaitEventTimeCodeFrameResponse;
char                 g_cSerialWaitEventTopOfSecond;
char                 g_cSerialWaitEventResponse;
DWORD			     ThreadIdSerialTopOfSecond;
DWORD                ThreadIdTimeCodeFrameAlarm;

char g_cLastLockStatus;

LARGE_INTEGER            g_liPerfCountValueTopOfSecond;
LARGE_INTEGER            g_liPerfCountValueFrame;
WCATIME                  g_LastDateTimeResponse;
TIME_CODE_FRAME_COMPLETE g_LastTimeCodeResponse;
LAST_ALARM_STATUS        g_LastAlarmStatusResponse;

DWORD g_dwPciCommunicationsStatus, g_dwPciCommunicationsStatusLast;

void  fnScanCallbacks(UCHAR interruptReason, MCR_TIME_REV_A *plMcrTime);
void  CreateSerialEvents(void);
void  SetSerialEventTimeCodeFrameResponse(void);
void  SetSerialEventTimeCodeFrameAlarm(void);
void  SetSerialEventTopOfSecondStart(void);
void  SetSerialEventTopOfSecondData(void);
void  SetSerialEventResponse(void);
BOOL  WaitForSerialEventTimeCodeFrameResponse(BOOL bIsForever);
BOOL  WaitForSerialEventTimeCodeFrameAlarm(BOOL bIsForever);
BOOL  WaitForSerialEventTopOfSecondStart(BOOL bIsForever);
BOOL  WaitForSerialEventTopOfSecondData(BOOL bIsForever);
BOOL  WaitForSerialEventResponse(BOOL bIsForever);
void  AcquireSerialEventMutex(void);
void  ReleaseSerialEventMutex(void);
void  fnSerialRxBufferInit(void);
void  fnSerialProcessRxData(void);
void  fnSerialProcessRxPacket(void);
int   fnSerialReadRxData(void);
int   fnGetIntegerValue(char *piBuffer, int *piIndex);
int   fnGetStringValue(char *piBufferDest, int iMaxLengthDest, char *piBufferSrc, int *piIndex);
void  fnProcessDATE_TIME_REQUEST(void);
void  fnProcessDATE_TIME_RESPONSE(void);
void  fnProcessAVR_NOTIFY_ALARM(void);
void  fnProcessPC_READ_TIME_CODE_RESPONSE(void);
void  fnCopyDateTimeFromResponse(SYSTEMTIME *psrDateTime);
void  fnSendNAK(char cCommand, int iReason);
void  fnSendACK(void);
void  fnSerialTimeCodeFrameAlarm(void);
void  fnSerialInterruptDsr(void);

BOOL  DecodeTimeFrame(TIME_CODE_FRAME_REV_A *timeFrame, MCR_TIME_REV_A *mcrTime);

// needed by IRIG
// for values 0-999
//#define bin_to_BCD(val) ( ((val/100)<<8) | ((val/10)<<4) | (val%10))                     //Convert a binary value to BCD

// for values 0-99
#define BCD_to_bin(val) (((((val >> 4) & 0x0F) * 10)) + (val & 0x0F))  //Convert a BCD value to  binary

//for values 0-999
unsigned short bin_to_BCD(unsigned short usBIN)
{
	unsigned short usVal1, usVal2, usVal3;
	unsigned short usResult;

	usVal1 = usVal2 = usVal3 = usBIN;
	usVal1 = (usVal1/100)<<8;
	usVal2 = usVal2 % 100;   // range 0-99
	usVal2 = (usVal2/10) <<4;
	usVal3 = usVal3%10;

	usResult = usVal1 | usVal2 | usVal3;

	return usResult;
}

// for values 0-99
//unsigned short BCD_to_bin(unsigned short usBCD)
//{
//	unsigned short usResult;
//
//	usResult = usBCD;
//	usResult = (usResult >> 4) & 0x0F);
//	usResult = usResult * 10;
//	usResult = usResult + (usBCD & 0x0F);
//
//	usResult = usVal1 | usVal2 | usVal3;
//
//	return usResult;
//}





//
//		Name:			fnSerialEventMonitorThread()
//
//		Description:	Monitors MCR driver for asynchronous notifications via the serial port, process them, and calls
//						callback functions with required data.
//
//		Returns:		When hMonitorExit is signaled.
//
void fnSerialEventMonitorThread(void)
{
	int iPortNum;
	CSoftwareConfig	*pswConfig;

	LONG lLastError;

	DWORD dwObjectReceived;
	DWORD dwBytesWrite = 0;

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("Event Monitor() start delay"));
	#endif
	Sleep(1000);
	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("Event Monitor() start"));
	#endif

	g_cCallbackCount = 0;

	g_liPerfCountValueTopOfSecond.QuadPart  = 0;
	g_cSerialWaitEventTimeCodeFrameResponse = 0;
	g_cSerialWaitEventTopOfSecond           = 0;
	g_cSerialWaitEventResponse              = 0;
	g_cLastLockStatus                       = LOCK_INVALID;

	memset(&g_LastAlarmStatusResponse, 0, sizeof(g_LastAlarmStatusResponse));
	g_LastAlarmStatusResponse.ucHwStatus                    = HW_STATUS_FAILED;
	g_LastAlarmStatusResponse.srLockStatus.ucLockStatus     = LOCK_INVALID;
	g_LastAlarmStatusResponse.srLockStatus.ucHsoStatus      = LOCK_INVALID;
	g_LastAlarmStatusResponse.srLockStatus.ucGpsStatus      = LOCK_INVALID;
	g_LastAlarmStatusResponse.srLockStatus.ucRtcStatus      = LOCK_INVALID;
	g_LastAlarmStatusResponse.srLockStatus.ucTimeCodeStatus = LOCK_INVALID;

	fnSerialRxBufferInit();

	pCMcrSerialInterface = new CMcrSerialInterface();

	// If no COM port has been defined, this will set set it.
	pswConfig = new CSoftwareConfig(cDebugOutput);
	
	if (!fnValidateComPort(pCMcrSerialInterface, pswConfig))
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("fnSerialEventMonitorThread() is terminating, no valid COM port found."));
		#endif

		if (pCMcrSerialInterface->mcrSerial.IsOpen())
			pCMcrSerialInterface->mcrSerial.csClose();

		delete pCMcrSerialInterface;

		return;
	}

	//
	iPortNum = pswConfig->GetRegistrySerialPortNum();
	if (!fnSerialComPortInit(&pCMcrSerialInterface->mcrSerial, iPortNum))
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("fnSerialEventMonitorThread() is terminating, failed to initialize COM port"));
		#endif

		g_dwPciCommunicationsStatus = PCI_COMMUNICATIONS_LOST;

		if (pCMcrSerialInterface->mcrSerial.IsOpen())
		{
			pCMcrSerialInterface->mcrSerial.csClose();
		}

		delete pCMcrSerialInterface;

		return;
	}
	
	delete pswConfig;

	// Register for the events we want to receive from the serial port
	lLastError = pCMcrSerialInterface->mcrSerial.SetMask(
		CSerial::EEventBreak |
		CSerial::EEventCTS   |
		CSerial::EEventDSR   |
		CSerial::EEventError |
		CSerial::EEventRing  |
		CSerial::EEventRLSD  |
		CSerial::EEventRecv);

	#ifdef _DEBUG_ERROR
	if (lLastError != ERROR_SUCCESS)
	{
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread error setting handshake parameters: %ld"), lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);

		g_dwPciCommunicationsStatus = PCI_COMMUNICATIONS_LOST;

		if (pCMcrSerialInterface->mcrSerial.IsOpen())
		{
			pCMcrSerialInterface->mcrSerial.csClose();
		}

		delete pCMcrSerialInterface;

		return;

	}
	#endif



	if (g_dwPciCommunicationsStatus != PCI_COMMUNICATIONS_LOST)
	{
		g_dwPciCommunicationsStatus = PCI_COMMUNICATIONS_OKAY;
	}


	if (pswConfig->m_srKeys.dwNmeaMessagesBitMap)
	{
		// validate the PCI NMEA virtual COM port number

		// NMEA is being output, is it the the virtual port on this card, or is it an external port?
		if (pswConfig->m_srKeys.dwNmeaComPortNumber != pswConfig->m_srKeys.pciNmeaVirtualComPort)
		{
			// it is an external port, open the port

			// Open the serial port, if we get an error and cPrimaryPericomPortFound is different
			// then our current value, try using the value from cPrimaryPericomPortFound

			if (!fnOpenComPort(&pCMcrSerialInterface->nmeaSerial, pswConfig->m_srKeys.dwNmeaComPortNumber, pswConfig->m_srKeys.dwNmeaBaudrate, pswConfig->m_srKeys.dwNmeaDataBits, pswConfig->m_srKeys.dwNmeaParity, pswConfig->m_srKeys.dwNmeaStopBits))
			{
				#ifdef _DEBUG_ERROR
					TCHAR tcDebugBuff[256];
					LONG lLastError;
					lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread failed opening NMEA port: %d"), (int)pswConfig->m_srKeys.dwNmeaComPortNumber);
					cDebugOutput->DebugMsg(tcDebugBuff);
				#endif
			}
		}
	}


	// Create a handle for the overlapped operations
	HANDLE hevtOverlapped = ::CreateEvent(0,TRUE,FALSE,0);;
	#ifdef _DEBUG_ERROR
	if (hevtOverlapped == 0)
	{
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread error creating event for overlapped I/O.: %ld"), lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	// Setup the overlapped structure
	OVERLAPPED ov = {0};
	ov.hEvent = hevtOverlapped;

	// Open the "STOP" handle
	HANDLE hevtStop = ::CreateEvent(0,TRUE,FALSE,_T("Overlapped_Stop_Event"));
	#ifdef _DEBUG_ERROR
	if (hevtStop == 0)
	{
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread error creating stop event: %ld"), lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif


	CreateSerialEvents();
	if (CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)fnSerialTopOfSecondThread, (LPVOID)cDebugOutput, 0, &ThreadIdSerialTopOfSecond) == NULL)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!InitDLLInstance: create fnSerialTopOfSecondThread failed"));
		#endif

		return;
	}

	if (CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)fnSerialTimeCodeFrameAlarmThread, (LPVOID)cDebugOutput, 0, &ThreadIdTimeCodeFrameAlarm) == NULL)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!InitDLLInstance: create fnSerialTimeCodeFrameAlarmThread failed"));
		#endif

		return;
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("Event Monitor() serial port initialized."));
	#endif

	bool waitActive = true;
	while (waitActive)
	{
		// Wait for an event
		lLastError = pCMcrSerialInterface->mcrSerial.WaitEvent(&ov);
		#ifdef _DEBUG_ERROR
		if (lLastError != ERROR_SUCCESS)
		{
			TCHAR tcDebugBuff[256];
			LONG lLastError;
			lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread Unable to wait for a COM-port event: %ld"), lLastError);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		// Setup array of handles in which we are interested
		HANDLE ahObjectWait[2];
		ahObjectWait[0] = hevtOverlapped;
		ahObjectWait[1] = hevtStop;


		// Wait until something happens
		dwObjectReceived = ::WaitForMultipleObjects(sizeof(ahObjectWait)/sizeof(*ahObjectWait),ahObjectWait,FALSE,INFINITE);
		switch (dwObjectReceived)
		{
			case WAIT_OBJECT_0:
				{
					// Save event
					const CSerial::EEvent eEvent = pCMcrSerialInterface->mcrSerial.GetEventType();

					// Handle break event
					if (eEvent & CSerial::EEventBreak)
					{
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[256];
							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!BREAK event received"));
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif
					}

					//
					//
					// Modem control line summary PCIe chip
					// DTR - output
					// RTS - output
					// DSR - input
					// CTS - input
					//
					//

					// Handle CTS event
					if (eEvent & CSerial::EEventCTS)
					{
						// Currently, we do nothing with this event
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[256];
							TCHAR *pStr;
							pStr = pCMcrSerialInterface->mcrSerial.GetCTS()?_T("on"):_T("off");
							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CTS com event: %s"), pStr);
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif
					}

					// Handle DSR event
					if (eEvent & CSerial::EEventDSR)
					{
						fnSerialInterruptDsr();
					}

					// Handle error event
					if (eEvent & CSerial::EEventError)
					{
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[512];
							TCHAR *pStr;

							switch (pCMcrSerialInterface->mcrSerial.GetError())
							{
								case CSerial::EErrorBreak:		pStr = _T("Break condition");			break;
								case CSerial::EErrorFrame:		pStr = _T("Framing error");				break;
								case CSerial::EErrorIOE:		pStr = _T("IO device error");			break;
								case CSerial::EErrorMode:		pStr = _T("Unsupported mode");			break;
								case CSerial::EErrorOverrun:	pStr = _T("Buffer overrun");			break;
								case CSerial::EErrorRxOver:		pStr = _T("Input buffer overflow");		break;
								case CSerial::EErrorParity:		pStr = _T("Input parity error");		break;
								case CSerial::EErrorTxFull:		pStr = _T("Output buffer full");		break;
								default:						pStr = _T("Unknown");					break;
							}

							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Error COM event: %s"), pStr);
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif
					}

					// Handle ring event
					if (eEvent & CSerial::EEventRing)
					{
						// Currently, we do nothing with this event
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[256];
							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!RING com event"));
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif
					}

					// Handle RLSD/CD event
					if (eEvent & CSerial::EEventRLSD)
					{
						// Currently, we do nothing with this event
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[256];
							TCHAR *pStr;
							pStr = pCMcrSerialInterface->mcrSerial.GetRLSD()?_T("on"):_T("off");
							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!RLSD com event: %s"), pStr);
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif
					}

					// Handle data receive event
					if (eEvent & CSerial::EEventRecv)
					{
						fnSerialProcessRxData();
					}
				}
				break;

			case WAIT_OBJECT_0+1:
				{
					// Set the continue bit to false, so we'll exit
					waitActive = false;
				}
				break;

			default:
				{
					// Something went wrong
					#ifdef _DEBUG_ERROR
					{
						TCHAR tcDebugBuff[256];
						lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
						swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Error while calling WaitForMultipleObjects: %ld"), lLastError);
						cDebugOutput->DebugMsg(tcDebugBuff);
					}
					#endif
				}
				break;
		}
	}

	#ifdef _DEBUG_INFO
	cDebugOutput->DebugMsg(_T("fnSerialEventMonitorThread() is terminating."));
	#endif

	pCMcrSerialInterface->mcrSerial.csClose();

	delete pCMcrSerialInterface;
}




//
//		Name:			fnSerialComPortInit()
//
//		Description:	Open the COM port, set the baud rate
//                      Must give the port number to open, port number must be > 0
//
//		Returns:		FALSE if successfully processed
//						!= 0, failure
BOOL fnSerialComPortInit(CSerial *pCSerialPort, int iPortNum)
{
	BOOL bIsOkay = true;

	LONG lLastError;

	if (iPortNum <= 0)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialComPortInit - invalid port %d"), iPortNum);
		cDebugOutput->DebugMsg(tcDebugBuff);

		return 0;

	}

	// Attempt to open the serial port
	if (!fnOpenComPort(pCSerialPort, iPortNum, 115200, 8, NMEA_PARITY_NONE, 1))
		return 0;

	// Setup handshaking (default is no handshaking)
	lLastError = pCSerialPort->SetupHandshaking(CSerial::EHandshakeOff);
	
	#ifdef _DEBUG_ERROR
	if (lLastError != ERROR_SUCCESS)
	{
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCSerialPort->GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread error setting handshake parameters: %d %ld"), iPortNum, lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);

		if (lLastError != ERROR_SUCCESS)
		{
			return 0;
		}
	}
	#endif

	// Use 'non-blocking' reads, because we don't know how many bytes
	// will be received. This is normally the most convenient mode
	// (and also the default mode for reading data).
	lLastError = pCSerialPort->SetupReadTimeouts(CSerial::EReadTimeoutNonblocking);
	
	#ifdef _DEBUG_ERROR
	if (lLastError != ERROR_SUCCESS)
	{
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCSerialPort->GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialEventMonitorThread error setting read timeout: %d %ld"), iPortNum, lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);

		if (lLastError != ERROR_SUCCESS)
		{
			return 0;
		}
	}
	#endif

	return bIsOkay;
}





//
//		Name:			fnSerialTopOfSecondThread()
//
//		Description:	Waits for a top of second event (initiated by the DSR modem control line).
//                      When this event is received,
//                      do the top of second processing needed.
//
//		Returns:		none
//
void fnSerialTopOfSecondThread(void)
{
	MCR_TIME_REV_A  lMcrTime;
	TIME_CODE_FRAME_REV_A srTimeFrameToReturn;

	char     tcCmdBuff[8];
	DWORD    dwBytesWrite;
	BOOL     bStatus;
	UCHAR    interruptReason;

	// When the DLL is loaded make sure we have the current status of the AVR
	AcquireSerialEventMutex();

	g_cSerialWaitEventTopOfSecond = PC_STATUS_RESPONSE;
	sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c",PC_STATUS_REQUEST);

	dwBytesWrite = (DWORD)strlen(tcCmdBuff);
	pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

	ReleaseSerialEventMutex();

	// wait here for the first top of second, then ignore it
	WaitForSerialEventTopOfSecondData(FALSE);

	// At the end of the first top of second, send configuration information
	// to the AVR.
	//
	// We wait until now to make sure we have communication with the AVR.
	//
	// These settings should have been saved in the AVR, we only send them
	// to make sure the PC and the AVR agree
	fnSetAllAvrConfiguration();


	// wait for a top of second event. when one occurs, perform the top of second processing
	while (1)
	{

		bStatus = WaitForSerialEventTopOfSecondStart(TRUE);
		if (bStatus != MCR_SUCCESS)
			continue;

		interruptReason = ASYNC_PERSEC;

		AcquireSerialEventMutex();

		g_cSerialWaitEventTopOfSecond = DATE_TIME_RESPONSE;
		sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c",DATE_TIME_REQUEST);

		dwBytesWrite = (DWORD)strlen(tcCmdBuff);
		pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

		ReleaseSerialEventMutex();

		bStatus = WaitForSerialEventTopOfSecondData(FALSE);
		if (bStatus != MCR_SUCCESS)
			continue;

		AcquireSerialEventMutex();
		g_cSerialWaitEventTopOfSecond = PC_STATUS_RESPONSE;
		sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c",PC_STATUS_REQUEST); // kpw trace - shouldn't this get lock info?

		dwBytesWrite = (DWORD)strlen(tcCmdBuff);
		pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

		ReleaseSerialEventMutex();

		bStatus = WaitForSerialEventTopOfSecondData(FALSE);
		if (bStatus != MCR_SUCCESS)
			continue;

		// default to all zeros
		memset(&lMcrTime, 0, sizeof(lMcrTime));

		// we know this was saved at the top of second since ASYNC_PER_SEC is set
		lMcrTime.intPerfCount.QuadPart      = g_liPerfCountValueTopOfSecond.QuadPart;

		// in case the user uses it, get the time code data. This is for compatibility
		// with old TCR1000 API calls that the user may have adapted to this card.
		ReadAndProcessTimeFrame(&srTimeFrameToReturn, &lMcrTime.intPerfCount, &lMcrTime);

		// We overwrite the time values returned in the ReadAndProcessTimeFrame() with the last values
		// we received at top of second.
		lMcrTime.tcrValid          = true;
		lMcrTime.wMonth            = g_LastDateTimeResponse.wMonth;
		lMcrTime.wDay              = g_LastDateTimeResponse.wDay;
		lMcrTime.wYear             = g_LastDateTimeResponse.wYear;
		lMcrTime.wHour             = g_LastDateTimeResponse.wHour;
		lMcrTime.wMinute           = g_LastDateTimeResponse.wMinute;
		lMcrTime.wSecond           = g_LastDateTimeResponse.wSecond;

		if (lMcrTime.wSecond == 30)
		{
			// at the top of second when seconds == 30, perform this callback if it is enabled
			interruptReason |= ASYNC_TIMESET;
		}

		if (g_cLastLockStatus != g_LastAlarmStatusResponse.srLockStatus.ucLockStatus)
		{

			// There was a lock status change, We only do the callback at the top of second
			if ((g_cLastLockStatus                                   == LOCK_RESTORED) && // If we were locked before
				(g_LastAlarmStatusResponse.srLockStatus.ucLockStatus != LOCK_RESTORED))   // and now we are not locked
			{
				interruptReason |= ASYNC_TIMELOSS;
			}

			if ((g_cLastLockStatus                                   != LOCK_RESTORED) && // If we were not locked before
				(g_LastAlarmStatusResponse.srLockStatus.ucLockStatus == LOCK_RESTORED))   // and now we are locked
			{
				interruptReason |= ASYNC_TIMEGAIN;
			}

			g_cLastLockStatus = g_LastAlarmStatusResponse.srLockStatus.ucLockStatus;
		}

		// We received all of the information we need for top of second
		// Time to execute the callbacks

		fnScanCallbacks(interruptReason, &lMcrTime);

	}
}



//
//		Name:			fnSerialTimeCodeFrameAlarmThread()
//
//		Description:	Initiated when the alarm is received which says a new time code frame
//                      is available to be read from the AVR.
//
//		Returns:		none
//
void fnSerialTimeCodeFrameAlarmThread(void)
{
	MCR_TIME_REV_A  lMcrTime;
	TIME_CODE_FRAME_REV_A srTimeFrameToReturn;

	BOOL     bStatus;
	UCHAR    interruptReason;

	// wait for a time code frame available event
	while (1)
	{

		bStatus = WaitForSerialEventTimeCodeFrameAlarm(TRUE);
		if (bStatus != MCR_SUCCESS)
			continue;

		interruptReason = ASYNC_ENDTC | ASYNC_STARTTC;

		// default to all zeros
		memset(&lMcrTime, 0, sizeof(lMcrTime));

		// we know this was saved at the top of second since ASYNC_PER_SEC is set
		lMcrTime.intPerfCount.QuadPart      = g_liPerfCountValueFrame.QuadPart;

		// get the time code data
		ReadAndProcessTimeFrame(&srTimeFrameToReturn, &lMcrTime.intPerfCount, &lMcrTime);

		// We received all of the information we need for the frame alarm
		// Time to execute the callbacks

		fnScanCallbacks(ASYNC_ENDTC | ASYNC_STARTTC, &lMcrTime);

	}
}




//
//		Name:			fnScanCallbacks()
//
//		Description:	If there was any callbacks that need to be called, do it now.
//                      The flags get set 
//
//
//
void fnScanCallbacks(UCHAR interruptReason, MCR_TIME_REV_A *plMcrTime)
{
	DWORD			msToAdd;
	LARGE_INTEGER	curPerfCount,perfFreq,liTempPerfCounter;

	PCALLBACK_NODE	cbNode;


	// If the interrupt is a timing interrupt, process the interrupt data
	QueryPerformanceFrequency(&perfFreq);
	
	// the current high performance counter value
	QueryPerformanceCounter(&liTempPerfCounter);

	// asynchronous time completed

	// scan callback list for threads which have requested callbacks for this asynchronous
	// notification

	// protect ourselves from accessing the callback list while it is being changed by
	// obtaining the mutex

	#ifdef _DEBUG_INFO1
	cDebugOutput->DebugMsg(_T("Callback event critical section start"));
	#endif

	EnterCriticalSection(&g_srDllShared.csCBListData);

	cbNode = g_cbListHead;

	while (cbNode)
	{
		// Check for each of the 6 different kinds of interrupts,
		// and call any attached functions

		if (interruptReason & ASYNC_PERSEC &&
			cbNode->cbHooksMask & ASYNC_PERSEC)
		{

			#ifdef _DEBUG_INFO
			{
				//TCHAR tcDebugBuff[256];
				//swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("callback ASYNC_PERSEC %d/%d/%d %d:%d:%d"),
				//	(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear,
				//	(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond);
				//cDebugOutput->DebugMsg(tcDebugBuff);
				cDebugOutput->DebugMsg(_T("callback ASYNC_PERSEC"));
			}
			#endif

			// determine the msec. offset we'll have to apply
			QueryPerformanceCounter(&curPerfCount);

			msToAdd = (ULONG)((curPerfCount.QuadPart - plMcrTime->intPerfCount.QuadPart) / (perfFreq.QuadPart / (__int64)1000));
			msToAdd = msToAdd % 1000;

			plMcrTime->wMilliseconds     = (WORD)msToAdd;

			CopyMemory(&cbNode->timeData,plMcrTime,sizeof(MCR_TIME_REV_A));

			(*(cbNode->cbFunction))(ASYNC_PERSEC,&cbNode->timeData,sizeof(MCR_TIME_REV_A));

			#ifdef _DEBUG_INFO1
			cDebugOutput->DebugMsg(_T("callback ASYNC_PERSEC done"));
			#endif

			if (g_dwPciCommunicationsStatus != g_dwPciCommunicationsStatusLast)
			{
				if (g_dwPciCommunicationsStatus == PCI_COMMUNICATIONS_LOST)
				{
					(*(cbNode->cbFunction))(PCI_COMMUNICATIONS_LOST,NULL,0);

					#ifdef _DEBUG_INFO1
					cDebugOutput->DebugMsg(_T("callback PCI_COMMUNICATIONS_LOST done"));
					#endif
				}

				if (g_dwPciCommunicationsStatus == PCI_COMMUNICATIONS_OKAY)
				{
					(*(cbNode->cbFunction))(PCI_COMMUNICATIONS_OKAY,NULL,0);

					#ifdef _DEBUG_INFO1
					cDebugOutput->DebugMsg(_T("callback PCI_COMMUNICATIONS_OKAY done"));
					#endif
				}

				g_dwPciCommunicationsStatusLast = g_dwPciCommunicationsStatus;
			}

		}

		if (interruptReason & ASYNC_TIMESET &&
			cbNode->cbHooksMask & ASYNC_TIMESET)
		{

			#ifdef _DEBUG_INFO
			{
				cDebugOutput->DebugMsg(_T("callback ASYNC_TIMESET"));
			}
			#endif

			// determine the msec. offset we'll have to apply
			QueryPerformanceCounter(&curPerfCount);

			msToAdd = (ULONG)((curPerfCount.QuadPart - plMcrTime->intPerfCount.QuadPart) / (perfFreq.QuadPart / (__int64)1000));
			msToAdd = msToAdd % 1000;

			plMcrTime->wMilliseconds     = (WORD)msToAdd;

			CopyMemory(&cbNode->timeData,plMcrTime,sizeof(MCR_TIME_REV_A));

			(*(cbNode->cbFunction))(ASYNC_TIMESET,&cbNode->timeData,sizeof(MCR_TIME_REV_A));
		}

		if (interruptReason & ASYNC_STARTTC &&
			cbNode->cbHooksMask & ASYNC_STARTTC)
		{


			#ifdef _DEBUG_INFO
			{
				cDebugOutput->DebugMsg(_T("callback ASYNC_STARTTC completed"));
			}
			#endif

			// determine the msec. offset we'll have to apply
			QueryPerformanceCounter(&curPerfCount);

			msToAdd = (ULONG)((curPerfCount.QuadPart - plMcrTime->intPerfCount.QuadPart) / (perfFreq.QuadPart / (__int64)1000));
			msToAdd = msToAdd % 1000;

			plMcrTime->wMilliseconds     = (WORD)msToAdd;

			CopyMemory(&cbNode->timeData,plMcrTime,sizeof(MCR_TIME_REV_A));

			(*(cbNode->cbFunction))(ASYNC_STARTTC,&cbNode->timeData,sizeof(MCR_TIME_REV_A));
		}

		if (interruptReason & ASYNC_ENDTC &&
			cbNode->cbHooksMask & ASYNC_ENDTC)
		{

			#ifdef _DEBUG_INFO
			{
				//TCHAR tcDebugBuff[256];
				//swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("TC ASYNC_ENDTC %d/%d/%d %d:%d:%d"),
				//	(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear,
				//	(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond);
				//cDebugOutput->DebugMsg(tcDebugBuff);
				cDebugOutput->DebugMsg(_T("callback ASYNC_ENDTC"));
			}
			#endif

			// determine the msec. offset we'll have to apply
			QueryPerformanceCounter(&curPerfCount);

			msToAdd = (ULONG)((curPerfCount.QuadPart - plMcrTime->intPerfCount.QuadPart) / (perfFreq.QuadPart / (__int64)1000));
			msToAdd = msToAdd % 1000;

			plMcrTime->wMilliseconds     = (WORD)msToAdd;

			CopyMemory(&cbNode->timeData,plMcrTime,sizeof(MCR_TIME_REV_A));

			(*(cbNode->cbFunction))(ASYNC_ENDTC,&cbNode->timeData,sizeof(MCR_TIME_REV_A));
		}

		if (interruptReason & ASYNC_TIMELOSS &&
			cbNode->cbHooksMask & ASYNC_TIMELOSS)
		{
			#ifdef _DEBUG_INFO
			cDebugOutput->DebugMsg(_T("TC ASYNC_TIMELOSS"));
			#endif

			(*(cbNode->cbFunction))(ASYNC_TIMELOSS,NULL,0);

		}

		if (interruptReason & ASYNC_TIMEGAIN &&
			cbNode->cbHooksMask & ASYNC_TIMEGAIN)
		{
			#ifdef _DEBUG_INFO
			cDebugOutput->DebugMsg(_T("TC ASYNC_TIMEGAIN"));
			#endif

			(*(cbNode->cbFunction))(ASYNC_TIMEGAIN,NULL,0);
		}

		cbNode = (PCALLBACK_NODE)cbNode->next;
	}

	//#ifdef _DEBUG_INFO1
	//cDebugOutput->DebugMsg(_T("Callback event critical section end"));
	//#endif
	LeaveCriticalSection(&g_srDllShared.csCBListData);

	//#ifdef _DEBUG_INFO
	//cDebugOutput->DebugMsg(_T("Callback scan complete"));
	//#endif
}







//
//		Name:			ReadAndProcessTimeFrame()
//
//		Description:	Read and process a time frame from the AVR
//
//		Returns:		0 if data read OK
//

DWORD ReadAndProcessTimeFrame(TIME_CODE_FRAME_REV_A *ptimeFrameToReturn, LARGE_INTEGER *pliPerfCountIn, MCR_TIME_REV_A *pmcrTime)
{
	DWORD           dwReturn;
	TIME_CODE_FRAME_COMPLETE srCompleteTimeFrame;
	TIME_CODE_FRAME_REV_A		srResponseTimeframe;

	int             ii;
	UINT16          usDayOfYear;

	UINT8           ucTcType;

	memset(&srResponseTimeframe, 0, sizeof(srResponseTimeframe));

	memset(&srCompleteTimeFrame, 0, sizeof(srCompleteTimeFrame));

	// Read the last time code frame from the AVR
	dwReturn = fnReadTimeCodeFrame(&srCompleteTimeFrame);

	if (dwReturn)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnScanCallbacks: populate fail: %d"), (int)dwReturn);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return DLL_TC_STATUS_DRVFAIL;	// success
	}


	// Get the interrupt status and data
	// Data is only populated if the interrupt is a timing interrupt
	if ((srCompleteTimeFrame.tcStatus & REF_LOCKED_MASK) == REF_IS_LOCKED)
	{
		// time code is locked
		ucTcType = srCompleteTimeFrame.tcStatus & TIME_CODE_TYPE_MASK;
		switch (ucTcType)
		{
		case SMPTE_24:
		case SMPTE_25:
		case SMPTE_30:
			srResponseTimeframe.tcBuffer[0]  = (UINT8)(srCompleteTimeFrame.tcFrame[0] & 0x0F);        // frame units
			srResponseTimeframe.tcBuffer[0] |= (UINT8)(srCompleteTimeFrame.tcFrame[1] & 0x0F) << 4;   // frame tens
			srResponseTimeframe.tcBuffer[1]  = (UINT8)(srCompleteTimeFrame.tcFrame[2] & 0x0F);        // seconds units
			srResponseTimeframe.tcBuffer[1] |= (UINT8)(srCompleteTimeFrame.tcFrame[3] & 0x0F) << 4;   // seconds tens
			srResponseTimeframe.tcBuffer[2]  = (UINT8)(srCompleteTimeFrame.tcFrame[4] & 0x0F);        // minutes units
			srResponseTimeframe.tcBuffer[2] |= (UINT8)(srCompleteTimeFrame.tcFrame[5] & 0x0F) << 4;   // minutes tens
			srResponseTimeframe.tcBuffer[3]  = (UINT8)(srCompleteTimeFrame.tcFrame[6] & 0x0F);        // hours units
			srResponseTimeframe.tcBuffer[3] |= (UINT8)(srCompleteTimeFrame.tcFrame[7] & 0x0F) << 4;   // hours tens
			srResponseTimeframe.tcBuffer[4]  = (UINT8)(srCompleteTimeFrame.tcFrame[0] & 0xF0) >> 4;   // user group 1
			srResponseTimeframe.tcBuffer[4] |= (UINT8)(srCompleteTimeFrame.tcFrame[1] & 0xF0);        // user group 2
			srResponseTimeframe.tcBuffer[5]  = (UINT8)(srCompleteTimeFrame.tcFrame[2] & 0xF0) >> 4;   // user group 3
			srResponseTimeframe.tcBuffer[5] |= (UINT8)(srCompleteTimeFrame.tcFrame[3] & 0xF0);        // user group 4
			srResponseTimeframe.tcBuffer[6]  = (UINT8)(srCompleteTimeFrame.tcFrame[4] & 0xF0) >> 4;   // user group 5
			srResponseTimeframe.tcBuffer[6] |= (UINT8)(srCompleteTimeFrame.tcFrame[5] & 0xF0);        // user group 6
			srResponseTimeframe.tcBuffer[7]  = (UINT8)(srCompleteTimeFrame.tcFrame[6] & 0xF0) >> 4;   // user group 7
			srResponseTimeframe.tcBuffer[7] |= (UINT8)(srCompleteTimeFrame.tcFrame[7] & 0xF0);        // user group 8
			break;

		case IRIG_B_AM:
		case IRIG_B_PWM:
			srResponseTimeframe.tcBuffer[0]  = (UINT8)(srCompleteTimeFrame.tcFrame[0] & 0x1E) >> 1;   // Seconds
			srResponseTimeframe.tcBuffer[0] |= (UINT8)(srCompleteTimeFrame.tcFrame[0] & 0xC0) >> 2;   //
			srResponseTimeframe.tcBuffer[0] |= (UINT8)(srCompleteTimeFrame.tcFrame[1] & 0x01) << 6;   //

			srResponseTimeframe.tcBuffer[1]  = (UINT8)(srCompleteTimeFrame.tcFrame[1] & 0x3C) >> 2;   // Minutes
			srResponseTimeframe.tcBuffer[1] |= (UINT8)(srCompleteTimeFrame.tcFrame[1] & 0x80) >> 3;   //
			srResponseTimeframe.tcBuffer[1] |= (UINT8)(srCompleteTimeFrame.tcFrame[2] & 0x03) << 5;   //

			srResponseTimeframe.tcBuffer[2]  = (UINT8)(srCompleteTimeFrame.tcFrame[2] & 0xF0) >> 4;   // Hours
			srResponseTimeframe.tcBuffer[2] |= (UINT8)(srCompleteTimeFrame.tcFrame[3] & 0x06) << 3;   //

			usDayOfYear                     = (UINT8)(srCompleteTimeFrame.tcFrame[3] & 0xC0) >> 6;   // Day Of Year 
			usDayOfYear                    |= (UINT8)(srCompleteTimeFrame.tcFrame[4] & 0x03) << 2;   // 
			usDayOfYear                    |= (UINT8)(srCompleteTimeFrame.tcFrame[4] & 0x78) << 1;   //
			usDayOfYear                     = BCD_to_bin(usDayOfYear);
			usDayOfYear                    += ((srCompleteTimeFrame.tcFrame[5] & 0x03) >> 0) * 100;
			usDayOfYear                     = bin_to_BCD(usDayOfYear);

			srResponseTimeframe.tcBuffer[3]  = (UINT8)(usDayOfYear & 0x00FF);
			srResponseTimeframe.tcBuffer[4]  = (UINT8)((usDayOfYear & 0xFF00) >> 8);

			srResponseTimeframe.tcBuffer[4] |= srCompleteTimeFrame.tcFrame[6] & 0xF0;       // Control Function 1 
			srResponseTimeframe.tcBuffer[5]  = srCompleteTimeFrame.tcFrame[6] & 0x0F;       // Control Function 2 
			srResponseTimeframe.tcBuffer[5] |= srCompleteTimeFrame.tcFrame[7] & 0xF0;       // Control Function 3 
			srResponseTimeframe.tcBuffer[6]  = srCompleteTimeFrame.tcFrame[7] & 0x0F;       // Control Function 4 
			srResponseTimeframe.tcBuffer[6] |= srCompleteTimeFrame.tcFrame[8] & 0xF0;       // Control Function 5 
			srResponseTimeframe.tcBuffer[7]  = srCompleteTimeFrame.tcFrame[8] & 0x0F;       // Control Function 6 
			srResponseTimeframe.tcBuffer[7] |= srCompleteTimeFrame.tcFrame[9] & 0xF0;       // Control Function 7 

			break;
		}
	} else
	{
		// time code is not locked
		memset(srResponseTimeframe.tcBuffer, 0, sizeof(srResponseTimeframe.tcBuffer));
	}

	UINT8 tmpSum = srResponseTimeframe.tcBuffer[0];
	for (ii=1; ii<=7; ii++)
		tmpSum += srResponseTimeframe.tcBuffer[ii];

	srResponseTimeframe.tcChecksum            = tmpSum;

	srResponseTimeframe.tcStatus              = srCompleteTimeFrame.tcStatus;
	srResponseTimeframe.intPerfCount.QuadPart = pliPerfCountIn->QuadPart;
	srResponseTimeframe.hwStatus              = srCompleteTimeFrame.hwStatus;
	srResponseTimeframe.ucLockStatus          = srCompleteTimeFrame.srLockStatus.ucLockStatus;

	// Data read, now process it into the correct format
	DecodeTimeFrame(&srResponseTimeframe, pmcrTime);

	*ptimeFrameToReturn = srResponseTimeframe;

	// kjf xxx yyy zzz
	return 0;
}









//
//		Name:			DecodeTimeFrame()
//
//		Description:	Build a MCR_TIME_REV_A structure from the t structure returned by MCR kernel-mode
//						driver.
//
//		Returns:		TRUE if time code successfully processed
//						FALSE if no time code present
//
BOOL DecodeTimeFrame(TIME_CODE_FRAME_REV_A *timeFrame, MCR_TIME_REV_A *mcrTime)
{
	WORD	tcnt;
	BOOL	tmpflag;
	BOOL    bStatus = MCR_FAIL;

	// Decode the time value, based on if we have SMPTE or IRIG

	switch (timeFrame->tcStatus & TIME_CODE_TYPE_MASK)
	{
	case SMPTE_30:
	case SMPTE_25:
	case SMPTE_24:
		// SMPTE
		mcrTime->dwTimeCodeType = TC_TYPE_SMPTE;

		// Populate SMPTE structure from tcBuffer
		//MSF fixed bits that should not be used for time
		mcrTime->rawTc.smpte.fu =  timeFrame->tcBuffer[0] & 0x0F;
		mcrTime->rawTc.smpte.ft = (timeFrame->tcBuffer[0] >> 4) & 0x03;  //0x0F;
		mcrTime->rawTc.smpte.su =  timeFrame->tcBuffer[1] & 0x0F;
		mcrTime->rawTc.smpte.st = (timeFrame->tcBuffer[1] >> 4) & 0x07;  //0x0F;
		mcrTime->rawTc.smpte.mu =  timeFrame->tcBuffer[2] & 0x0F;
		mcrTime->rawTc.smpte.mt = (timeFrame->tcBuffer[2] >> 4) & 0x07;  //0x0F;
		mcrTime->rawTc.smpte.hu =  timeFrame->tcBuffer[3] & 0x0F;
		mcrTime->rawTc.smpte.ht = (timeFrame->tcBuffer[3] >> 4) & 0x03;  //0x0F;
		mcrTime->rawTc.smpte.u1 =  timeFrame->tcBuffer[4] & 0x0F;
		mcrTime->rawTc.smpte.u2 = (timeFrame->tcBuffer[4] >> 4) & 0x0F;
		mcrTime->rawTc.smpte.u3 =  timeFrame->tcBuffer[5] & 0x0F;
		mcrTime->rawTc.smpte.u4 = (timeFrame->tcBuffer[5] >> 4) & 0x0F;
		mcrTime->rawTc.smpte.u5 =  timeFrame->tcBuffer[6] & 0x0F;
		mcrTime->rawTc.smpte.u6 = (timeFrame->tcBuffer[6] >> 4) & 0x0F;
		mcrTime->rawTc.smpte.u7 =  timeFrame->tcBuffer[7] & 0x0F;
		mcrTime->rawTc.smpte.u8 = (timeFrame->tcBuffer[7] >> 4) & 0x0F;

		// Populate the time data (raw data is BCD)
		mcrTime->wHour	  = (mcrTime->rawTc.smpte.ht * 10) + mcrTime->rawTc.smpte.hu;
		mcrTime->wMinute = (mcrTime->rawTc.smpte.mt * 10) + mcrTime->rawTc.smpte.mu;
		mcrTime->wSecond = (mcrTime->rawTc.smpte.st * 10) + mcrTime->rawTc.smpte.su;
		{
			WORD frameNumWord = (WORD)((mcrTime->rawTc.smpte.ft * 10) + mcrTime->rawTc.smpte.fu) * 1000;
			if ((timeFrame->tcStatus & TIME_CODE_TYPE_MASK) == 3) {
				mcrTime->wMilliseconds = frameNumWord / 24;
			} else if (timeFrame->tcStatus == 2) {
				mcrTime->wMilliseconds = frameNumWord / 25;
			} else {
				mcrTime->wMilliseconds = frameNumWord / 30;
			}
		}

		// Decode Leitch-format date from user groups
		// Note: the value of these members will be undetermined if
		// the Leitch-format date isn't present
		// Day tens are two LSB's of group 4. Day units are group 2
		mcrTime->wDay = ((mcrTime->rawTc.smpte.u4 & 0x03) * 10) + mcrTime->rawTc.smpte.u2;

		// Month ten is 3rd LSB of group 4. Month units are group 3
		mcrTime->wMonth = (((mcrTime->rawTc.smpte.u4 >> 2) & 0x01) * 10) + mcrTime->rawTc.smpte.u3;

		// Year tens is group 8. Year units is group 6.
		// This code has a Y2.1K problem..
		mcrTime->wYear	= 2000 + ((mcrTime->rawTc.smpte.u8) * 10) + mcrTime->rawTc.smpte.u6;

		// Pass on the interrupt value

		mcrTime->intPerfCount.QuadPart = timeFrame->intPerfCount.QuadPart;

		// decode lock status
		// note: the value of this member will be undetermined if
		// the time code source is not a Masterclock device with SMPTE, Leitch date encoding, with Masterclock proprietary control bits

		g_devInfo.dwTimeCodeGpsStatus = (mcrTime->rawTc.smpte.u5 << 4) | mcrTime->rawTc.smpte.u1;

		// however, we're only supporting freewheeling right now

		g_devInfo.dwTimeCodeGpsStatus &= 0x01;

		// Data is valid

		mcrTime->tcrValid = TRUE;

		// Perform a rough validity check of the date/time
		if ((mcrTime->wMonth < 1) || (mcrTime->wMonth > 12))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid month #1: %d"), (int)mcrTime->wMonth);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wDay < 1) || (mcrTime->wDay > 31))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid date #1: %d"), (int)mcrTime->wDay);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wYear < 1900) || (mcrTime->wYear > 2100))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid year #1: %d"), (int)mcrTime->wYear);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wHour < 0) || (mcrTime->wHour > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid hour #1: %d"), (int)mcrTime->wHour);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wMinute < 0) || (mcrTime->wMinute > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid minute #1: %d"), (int)mcrTime->wMinute);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wSecond < 0) || (mcrTime->wSecond > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid Second #1: %d"), (int)mcrTime->wSecond);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wMilliseconds < 0) || (mcrTime->wMilliseconds > 999))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid msec #1: %d"), (int)mcrTime->wMilliseconds);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if (mcrTime->tcrValid == TRUE)
			bStatus = MCR_SUCCESS;
		else
			bStatus = MCR_FAIL;

		break;

	case IRIG_B_PWM:
	case IRIG_B_AM:
		// IRIG

		if ((timeFrame->tcStatus & TIME_CODE_TYPE_MASK)  == IRIG_B_AM)
			mcrTime->dwTimeCodeType = TC_TYPE_IRIG_B1;
		else
			mcrTime->dwTimeCodeType = TC_TYPE_IRIG_B;

		// Populate IRIG structure from tcBuffer
		mcrTime->rawTc.irig.su  =  timeFrame->tcBuffer[0] & 0x0F;
		mcrTime->rawTc.irig.st  = (timeFrame->tcBuffer[0] >> 4) & 0x0F;
		mcrTime->rawTc.irig.mu  =  timeFrame->tcBuffer[1] & 0x0F;
		mcrTime->rawTc.irig.mt  = (timeFrame->tcBuffer[1] >> 4) & 0x0F;
		mcrTime->rawTc.irig.hu  =  timeFrame->tcBuffer[2] & 0x0F;
		mcrTime->rawTc.irig.ht  = (timeFrame->tcBuffer[2] >> 4) & 0x0F;
		mcrTime->rawTc.irig.du  =  timeFrame->tcBuffer[3] & 0x0F;
		mcrTime->rawTc.irig.dt  = (timeFrame->tcBuffer[3] >> 4) & 0x0F;
		mcrTime->rawTc.irig.dh  =  timeFrame->tcBuffer[4] & 0x0F;
		mcrTime->rawTc.irig.cf1 = (timeFrame->tcBuffer[4] >> 4) & 0x0F;
		mcrTime->rawTc.irig.cf2 =  timeFrame->tcBuffer[5] & 0x0F;
		mcrTime->rawTc.irig.cf3 = (timeFrame->tcBuffer[5] >> 4) & 0x0F;
		mcrTime->rawTc.irig.cf4 =  timeFrame->tcBuffer[6] & 0x0F;
		mcrTime->rawTc.irig.cf5 = (timeFrame->tcBuffer[6] >> 4) & 0x0F;
		mcrTime->rawTc.irig.cf6 =  timeFrame->tcBuffer[7] & 0x0F;
		mcrTime->rawTc.irig.cf7 = (timeFrame->tcBuffer[7] >> 4) & 0x0F;

		// Populate the time data (raw data is BCD)
		mcrTime->wHour	  = (mcrTime->rawTc.irig.ht * 10) + mcrTime->rawTc.irig.hu;
		mcrTime->wMinute = (mcrTime->rawTc.irig.mt * 10) + mcrTime->rawTc.irig.mu;
		mcrTime->wSecond = (mcrTime->rawTc.irig.st * 10) + mcrTime->rawTc.irig.su;
		mcrTime->wMilliseconds = 0;	// Irig cannot do subsecond timing.

		// decode IEEE1344-format date from control functions
		// note: the value of these members will be undetermined if
		// the IEEE1344 encoding isn't present

		mcrTime->wYear = mcrTime->rawTc.irig.cf1 & 0x0f;
		mcrTime->wYear += (((mcrTime->rawTc.irig.cf2 >> 1) | ((mcrTime->rawTc.irig.cf3 & 0x01) << 3)) * 10);

		// Adjust to the 21st century. This code has a Y2.1K problem.

		mcrTime->wYear += 2000;

		// Compute the day-of-year, and then get the month and day-of-month
		// (depends on if this is a leap year.

		mcrTime->wDay = (mcrTime->rawTc.irig.dh * 100) +
			(mcrTime->rawTc.irig.dt * 10) +
			mcrTime->rawTc.irig.du;

		tmpflag = !(mcrTime->wYear & 0x03);

		// determine month and day

		if (tmpflag)
		{
			for(tcnt = 1; tcnt < 13; tcnt++)
			{
				if (_daysMonthElapsedLeapYear[tcnt] >= mcrTime->wDay)
				{
					mcrTime->wMonth = tcnt;
					mcrTime->wDay   = mcrTime->wDay - _daysMonthElapsedLeapYear[tcnt-1];
					break;
				}
			}
		}
		else
		{
			for(tcnt = 1; tcnt < 13; tcnt++)
			{
				if (_daysMonthElapsedNonLeapYear[tcnt] >= mcrTime->wDay)
				{
					mcrTime->wMonth = tcnt;
					mcrTime->wDay   = mcrTime->wDay - _daysMonthElapsedNonLeapYear[tcnt-1];
					break;
				}
			}
		}

		// Pass on the interrupt value

		mcrTime->intPerfCount.QuadPart = timeFrame->intPerfCount.QuadPart;

		mcrTime->tcrValid = TRUE;

		// Perform a rough validity check of the date/time
		if ((mcrTime->wMonth < 1) || (mcrTime->wMonth > 12))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid month: %d"), (int)mcrTime->wMonth);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wDay < 1) || (mcrTime->wDay > 31))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid date: %d"), (int)mcrTime->wDay);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wYear < 1900) || (mcrTime->wYear > 2100))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid year #2: %d"), (int)mcrTime->wYear);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wHour < 0) || (mcrTime->wHour > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid hour #2: %d"), (int)mcrTime->wHour);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wMinute < 0) || (mcrTime->wMinute > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid minute #2: %d"), (int)mcrTime->wMinute);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wSecond < 0) || (mcrTime->wSecond > 60))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid Second #2: %d"), (int)mcrTime->wSecond);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if ((mcrTime->wMilliseconds < 0) || (mcrTime->wMilliseconds > 999))
		{
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid msec #2: %d"), (int)mcrTime->wMilliseconds);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			mcrTime->tcrValid = FALSE;
		}

		if (mcrTime->tcrValid == TRUE)
			bStatus = MCR_SUCCESS;
		else
			bStatus = MCR_FAIL;
		break;

	default:
		bStatus = MCR_FAIL;	// TC not present (either lost or never there)
	}

	return bStatus;
}









//
//		Name:			CreateSerialEvents()
//
//		Description:	Creates the Serial event used during serial communications.
//                      A mutex is used to make sure only a single process, thread
//                      or dll callback can perform a task which uses the PCIe serial
//                      communications port to the AVR on the PCIe card.
//
//
//
#define SERIAL_EVENT_RETRY_COUNT 30
#define SERIAL_EVENT_RETRY_MSEC  10
void CreateSerialEvents(void)
{
	// Create a manual-reset event object.

	g_hSerialEventTimeCodeFrameResponse = CreateEvent(
		NULL,						// default security attributes
		TRUE,						// manual-reset event
		FALSE,						// initial state is nonsignaled
		TEXT("ObjSerialEventTimeCodeFrameResponse")  // object name
		);

	if (g_hSerialEventTimeCodeFrameResponse == NULL)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CreateSerialEvent time code frame response failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}

	g_hSerialEventTimeCodeFrameAlarm = CreateEvent(
		NULL,						// default security attributes
		TRUE,						// manual-reset event
		FALSE,						// initial state is nonsignaled
		TEXT("ObjSerialEventTimeCodeFrameAlarm")  // object name
		);

	if (g_hSerialEventTimeCodeFrameAlarm == NULL)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CreateSerialEvent time code frame alarm failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}

	g_hSerialEventTopOfSecondStart = CreateEvent(
		NULL,						// default security attributes
		TRUE,						// manual-reset event
		FALSE,						// initial state is nonsignaled
		TEXT("ObjSerialEventTopOfSecondStart")  // object name
		);

	if (g_hSerialEventTopOfSecondStart == NULL)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CreateSerialEventStart top of second failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}

	g_hSerialEventTopOfSecondData = CreateEvent(
		NULL,						// default security attributes
		TRUE,						// manual-reset event
		FALSE,						// initial state is nonsignaled
		TEXT("ObjSerialEventTopOfSecondData")  // object name
		);

	if (g_hSerialEventTopOfSecondData == NULL)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CreateSerialEventData top of second failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}

	g_hSerialEventResponse = CreateEvent(
		NULL,						// default security attributes
		TRUE,						// manual-reset event
		FALSE,						// initial state is nonsignaled
		TEXT("ObjSerialEventResponse")  // object name
		);

	if (g_hSerialEventResponse == NULL)
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CreateSerialEvent response failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}









//
//		Name:		 fnSetAllAvrConfiguration()
//
//		Description: Sends all of the configuration data to the AVR.
//                   If any of the configuration changes, we just send it all.
//
//
//
void fnSetAllAvrConfiguration(void)
{
	CSoftwareConfig	*pswConfig;

	char  cConfigBuff[128];
	DWORD dwBytesWrite;
	int   iVirtual;

	#ifdef _DEBUG_INFO
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("Set all AVR configuration\n"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif


	// save off the service pointer so that we can call routines from the
	// callback routine

	pswConfig = new CSoftwareConfig(cDebugOutput);

	AcquireSerialEventMutex();

	// Write the configuration to the PCIe serial port
	// We review the registry keys and set local variables to make sure
	// the AVR gets appropriate values.

	// Set misc. configuration
	// - includes time zone, DST, etc.
	{
		UINT8 ucTimeCodeSourceSmpte, ucTimeCodeSourceIrig;
		bool bIsTimeCodeUTC;
		bool bIsTimeZoneBiasPositive;
		bool bDstIsUtcReferenced;
		UINT32 ulTimeZoneBias;
		UINT32 ulDstBias;

		UINT uDstMonth,uDstDate,uDstYear,uDstDayOfWeek,uDstHour,uDstMin,uDstSec;
		UINT uStdMonth,uStdDate,uStdYear,uStdDayOfWeek,uStdHour,uStdMin,uStdSec;
		
		ucTimeCodeSourceSmpte = (UINT8)pswConfig->m_srKeys.timeCodeSourceSmpte;
		ucTimeCodeSourceIrig  = (UINT8)pswConfig->m_srKeys.timeCodeSourceIrig;

		if (pswConfig->IsTzDescUTC())
			bIsTimeCodeUTC = true;
		else
			bIsTimeCodeUTC = false;

		// The PC configuration doesn't tell us if DST is UTC referenced. Instead it just changes
		// the start and end time to match local time.
		bDstIsUtcReferenced = false;

		// The PC's view of bias is the opposite of ours, use negative logic
		if (pswConfig->m_srKeys.tzInfo.Bias >= 0)
		{
			bIsTimeZoneBiasPositive = false;
			ulTimeZoneBias          = pswConfig->m_srKeys.tzInfo.Bias*60;
		} else
		{
			bIsTimeZoneBiasPositive = true;
			ulTimeZoneBias          = (-pswConfig->m_srKeys.tzInfo.Bias)*60; // negate to make it a positive value
		}

		// PC units are minutes, convert to seconds
		if (pswConfig->m_srKeys.tzInfo.DaylightBias >= 0)
		{
			ulDstBias = pswConfig->m_srKeys.tzInfo.DaylightBias*60;
		} else
		{
			ulDstBias = (-pswConfig->m_srKeys.tzInfo.DaylightBias)*60;  // negate to make it a positive value
		}


		uDstMonth     = pswConfig->m_srKeys.tzInfo.DaylightDate.wMonth;
		uDstDate      = pswConfig->m_srKeys.tzInfo.DaylightDate.wDay;
		uDstYear      = pswConfig->m_srKeys.tzInfo.DaylightDate.wYear;
		uDstDayOfWeek = pswConfig->m_srKeys.tzInfo.DaylightDate.wDayOfWeek;
		uDstHour      = pswConfig->m_srKeys.tzInfo.DaylightDate.wHour;
		uDstMin       = pswConfig->m_srKeys.tzInfo.DaylightDate.wMinute;
		uDstSec       = pswConfig->m_srKeys.tzInfo.DaylightDate.wSecond;

		uStdMonth     = pswConfig->m_srKeys.tzInfo.StandardDate.wMonth;
		uStdDate      = pswConfig->m_srKeys.tzInfo.StandardDate.wDay;
		uStdYear      = pswConfig->m_srKeys.tzInfo.StandardDate.wYear;
		uStdDayOfWeek = pswConfig->m_srKeys.tzInfo.StandardDate.wDayOfWeek;
		uStdHour      = pswConfig->m_srKeys.tzInfo.StandardDate.wHour;
		uStdMin       = pswConfig->m_srKeys.tzInfo.StandardDate.wMinute;
		uStdSec       = pswConfig->m_srKeys.tzInfo.StandardDate.wSecond;

		sprintf_s(cConfigBuff,sizeof(cConfigBuff),"%c<%u,%u,%u,%u,%lu,%lu,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u>",
			PC_SET_CONFIG,
			ucTimeCodeSourceSmpte,
			ucTimeCodeSourceIrig,
			bIsTimeCodeUTC,
			bIsTimeZoneBiasPositive,
			ulTimeZoneBias,
			ulDstBias,
			bDstIsUtcReferenced,
			uDstMonth,uDstDate,uDstYear,uDstDayOfWeek,uDstHour,uDstMin,uDstSec,
			uStdMonth,uStdDate,uStdYear,uStdDayOfWeek,uStdHour,uStdMin,uStdSec);

		dwBytesWrite = (DWORD)strlen(cConfigBuff);
		pCMcrSerialInterface->mcrSerial.Write(cConfigBuff, &dwBytesWrite);
	}
	ReleaseSerialEventMutex();
	Sleep(100); // Make sure the AVR has a time to process the data

	// Set NMEA configuration
	if (pswConfig->m_srKeys.dwNmeaComPortNumber == pswConfig->m_srKeys.pciNmeaVirtualComPort)
	{
		iVirtual = 1;
	} else
	{
		iVirtual = 0;
	}

	fnConfigureNmeaMessages(&pCMcrSerialInterface->mcrSerial, pswConfig->m_srKeys.dwNmeaBaudrate, pswConfig->m_srKeys.dwNmeaMessagesBitMap, iVirtual);
	Sleep(100); // Make sure the AVR has a time to process the data

	// Set GPS configuration
	AcquireSerialEventMutex();
	{

	}
	ReleaseSerialEventMutex();

	delete pswConfig;
}








//
//		Name:			SetSerialEventTimeCodeFrameResponse()
//
//		Description:	Sets the event that a time code frame response has been received.
//
//
//
void SetSerialEventTimeCodeFrameResponse(void)
{
    // Set g_hSerialEventTimeCodeFrameResponse to signaled

	if (! SetEvent(g_hSerialEventTimeCodeFrameResponse) )
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetSerialEventTimeCodeFrameResponse failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}






//
//		Name:			SetSerialEventTimeCodeFrameAlarm()
//
//		Description:	Sets the event that a time code frame alarm has beeen received.
//
//
//
void SetSerialEventTimeCodeFrameAlarm(void)
{
	// Set gg_hSerialEventTimeCodeFrame to signaled

	if (! SetEvent(g_hSerialEventTimeCodeFrameAlarm) )
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetSerialEventTimeCodeFrameAlarm failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}


//
//		Name:			SetSerialEventTopOfSecondStart()
//
//		Description:	Sets the event that a top of second has occurred (DSR modem control line).
//
//
//
void SetSerialEventTopOfSecondStart(void)
{

	if (! SetEvent(g_hSerialEventTopOfSecondStart) )
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetSerialEventTopOfSecondStart failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}











//
//		Name:			SetSerialEventTopOfSecondData()
//
//		Description:	Set the event that we have received top of second data.	
//
//
//
void SetSerialEventTopOfSecondData(void)
{

	if (! SetEvent(g_hSerialEventTopOfSecondData) )
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetSerialEventTopOfSecondData failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}












//
//		Name:			SetSerialEventResponse()
//
//		Description:	Set the event that we have received a response to data we requested.
//
//
//
void SetSerialEventResponse(void)
{

	if (! SetEvent(g_hSerialEventResponse) )
	{
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetSerialEventResponse failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		return;
	}
}







//
//		Name:			WaitForSerialEventTimeCodeFrameResponse()
//
//		Description:	Wait for a time code frame response data.
//
//
//
BOOL WaitForSerialEventTimeCodeFrameResponse(BOOL bIsForever)
{
	DWORD dwWaitResult;

	//dwWaitResult = WaitForSingleObject(g_hSerialEventTimeCodeFrame, SERIAL_EVENT_RETRY_COUNT*SERIAL_EVENT_RETRY_MSEC);
	dwWaitResult = WaitForSingleObject(g_hSerialEventTimeCodeFrameResponse, 5000);

	switch (dwWaitResult)
	{
	case WAIT_OBJECT_0:
		// Event object was signaled
		ResetEvent(g_hSerialEventTimeCodeFrameResponse);
		return MCR_SUCCESS;
		break;

	default:
		// An error occurred
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!WaitForSerialEventTimeCodeFrameResponse failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		if (g_dwPciCommunicationsStatus != PCI_COMMUNICATIONS_LOST)
		{
			g_dwPciCommunicationsStatus = PCI_COMMUNICATIONS_LOST;
		}

		return MCR_FAIL;
	}
}







//
//		Name:			WaitForSerialEventTimeCodeFrameAlarm()
//
//		Description:	Wait for the event that a time code frame alarm has been received.
//
//
//
BOOL WaitForSerialEventTimeCodeFrameAlarm(BOOL bIsForever)
{
	DWORD dwWaitResult;

	//dwWaitResult = WaitForSingleObject(g_hSerialEventTimeCodeFrame, SERIAL_EVENT_RETRY_COUNT*SERIAL_EVENT_RETRY_MSEC);
	dwWaitResult = WaitForSingleObject(g_hSerialEventTimeCodeFrameAlarm, 5000);

	switch (dwWaitResult)
	{
		case WAIT_OBJECT_0:
			// Event object was signaled
			ResetEvent(g_hSerialEventTimeCodeFrameAlarm);
			return MCR_SUCCESS;
			break;

		default:
			// Frame alarms will normally occur since
			// reading at a frame level s disabled by default

			// An error occurred
			//#ifdef _DEBUG_ERROR
			//{
			//	TCHAR tcDebugBuff[256];
			//	swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!WaitForSerialEventTimeCodeFrameAlarm failed (%d)\n"), GetLastError());
			//	cDebugOutput->DebugMsg(tcDebugBuff);
			//}
			//#endif
			return MCR_FAIL;
	}
}













//
//		Name:			WaitForSerialEventTopOfSecondStart()
//
//		Description:	Wait for the top of second event.
//
//
//
BOOL WaitForSerialEventTopOfSecondStart(BOOL bIsForever)
{
	DWORD dwWaitResult;

	//if (bIsForever)
	//{
	//	//dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondStart, 86400*1000); // wait up to one day
	//	dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondStart, 5000); // kjf xxx yyy zzz
	//} else
	//{
		dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondStart, 5000);       // wait up to 100 milleseconds
	//}

	switch (dwWaitResult)
	{
	case WAIT_OBJECT_0:
		// Event object was signaled
		ResetEvent(g_hSerialEventTopOfSecondStart);
		return MCR_SUCCESS;
		break;

	default:
		// An error occurred
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!WaitForSerialEventTopOfSecondStart failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
		return MCR_FAIL;
	}
}














//
//		Name:			WaitForSerialEventTopOfSecondData()
//
//		Description:	Wait for the event that top of second data has been received.
//
//
//
BOOL WaitForSerialEventTopOfSecondData(BOOL bIsForever)
{
	DWORD dwWaitResult;

	//if (bIsForever)
	//{
	//	//dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondData, 86400*1000); // wait up to one day
	//	dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondData, 15000); // kjf xxx yyy zzz
	//} else
	//{
		dwWaitResult = WaitForSingleObject(g_hSerialEventTopOfSecondData, 5000);       // wait up to 100 milleseconds
	//}

	switch (dwWaitResult)
	{
		case WAIT_OBJECT_0:
			// Event object was signaled
			ResetEvent(g_hSerialEventTopOfSecondStart);
			return MCR_SUCCESS;
			break;

		default:
			// An error occurred
			#ifdef _DEBUG_ERROR
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!WaitForSerialEventTopOfSecondData failed (%d)\n"), GetLastError());
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
			return MCR_FAIL;
	}
}








//
//		Name:			WaitForSerialEventResponse()
//
//		Description:	Wait for the event generated when response data to a request we sent is received.
//
//
//
BOOL WaitForSerialEventResponse(BOOL bIsForever)
{
	DWORD dwWaitResult;

	//if (bIsForever)
	//{
	//	//dwWaitResult = WaitForSingleObject(g_hSerialEventResponse, 86400*1000); // wait up to one day
	//	dwWaitResult = WaitForSingleObject(g_hSerialEventResponse, 15000); // kjf xxx yyy zzz
	//} else
	//{
		dwWaitResult = WaitForSingleObject(g_hSerialEventResponse, 5000);       // wait up to 1.5 seconds
	//}

	switch (dwWaitResult)
	{
	case WAIT_OBJECT_0:
		// Event object was signaled
		ResetEvent(g_hSerialEventResponse);
		return MCR_SUCCESS;
		break;

	default:
		// An error occurred
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!WaitForSerialEventResponse failed (%d)\n"), GetLastError());
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
		return MCR_FAIL;
	}
}






//
//		Name:			AcquireSerialEventMutex()
//
//		Description:	Waits for the Serial Event mutex. If the wait fails, just continue.
//                      It is possible that if a process has acquired the mutex, but, before the process
//                      can free the mutex it is deleted, the mutex would forever be allocated. Instead,
//                      we continue after the timeout.
//
//						We cannot use a true mutex (ie. CreateMutex() and WaitForSingleObject() )
//						Simulate one using this flag.
//
#define SERIAL_EVENT_MUTEX_RETRY_COUNT 30
#define SERIAL_EVENT_MUTEX_RETRY_MSEC  10
void AcquireSerialEventMutex(void)
{
	LONG lPriorValue;

	int ii;

	// simulate a 300 msec timeout
	for (ii=0; ii<SERIAL_EVENT_MUTEX_RETRY_COUNT; ii++)
	{
		if (!g_srDllShared.ldllSerialEventMutex)
		{
			lPriorValue = ::InterlockedExchange(&g_srDllShared.ldllSerialEventMutex, 1);

			if (!lPriorValue)
			{
				// It was a zero and we set it high, we can continue
				break;
			}

			// Another thread beat us to setting it to a one.
			// Wait for that thread to zero the value

		}

		Sleep(SERIAL_EVENT_MUTEX_RETRY_MSEC);
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!acquiring serial event mutex retry"));
		#endif
	}

	if (ii >= SERIAL_EVENT_MUTEX_RETRY_COUNT)
	{
		#ifdef _DEBUG_ERROR
		cDebugOutput->DebugMsg(_T("!!!Failed acquiring serial event mutex"));
		#endif
	}
}


//
//		Name:			ReleaseSerialEventMutex()
//
//		Description:	Releases the Serial Event mutex.
//
//						We cannot use a true mutex (ie. CreateMutex() and WaitForSingleObject() )
//						Simulate one using this flag.
//
void ReleaseSerialEventMutex(void)
{
	//HANDLE hProcessID;
	//DWORD  dwPriority;

	//hProcessID = GetCurrentProcess();
	//dwPriority = GetPriorityClass(hProcessID);
	//SetPriorityClass(hProcessID,REALTIME_PRIORITY_CLASS);

	::InterlockedExchange(&g_srDllShared.ldllSerialEventMutex, 0);

	//SetPriorityClass(hProcessID,dwPriority);

}


//
//		Name:			CloseSerialEvents()
//
//		Description:	Close the event handles. We tried doing it here
//                      but Windows generated an error. Comment out and let Windows
//                      do our housekeeping.
//
//		Returns:		
//
void CloseSerialEvents(void)
{
	//HANDLE hProcessID;
	//DWORD  dwPriority;

	//hProcessID = GetCurrentProcess();
	//dwPriority = GetPriorityClass(hProcessID);
	//SetPriorityClass(hProcessID,REALTIME_PRIORITY_CLASS);

	//CloseHandle(g_srDllShared.g_hSerialEventTimeCodeFrameResponse);
	//CloseHandle(g_srDllShared.g_hSerialEventTimeCodeFrameAlarm);
	//CloseHandle(g_srDllShared.g_hSerialEventTopOfSecondStart);
	//CloseHandle(g_srDllShared.g_hSerialEventTopOfSecondData);

	//SetPriorityClass(hProcessID,dwPriority);

}

















////////////////////////////////////////////////
//
// Process the serial data received
//
////////////////////////////////////////////////

#define WAITING_FOR_COMMAND         1
#define WAITING_FOR_LEFT_DELIMITER  2
#define WAITING_FOR_DATA            3

#define MAX_RX_RESP_BUFFER 1024
char acSerialRxBuffer[MAX_RX_RESP_BUFFER];  // buffer to save the characters as we receive them
char acSerialRxResponse[200];               // buffer to save the last response parsed from SerialRxBuffer
int  iNextIndex;
int  iCurrentRxState;
char cCurrentCommand;



//
//		Name:			fnSerialRxBufferInit()
//
//		Description:	Initialize the variables used for processing the response data.
//
//		Returns:		none
//
void fnSerialRxBufferInit(void)
{
	memset(acSerialRxBuffer, 0, sizeof(acSerialRxBuffer));
	memset(acSerialRxResponse, 0, sizeof(acSerialRxResponse));
	iNextIndex = 0;
	iCurrentRxState = WAITING_FOR_COMMAND;
	cCurrentCommand = 0;
}






//
//		Name:			fnSerialReadRxData()
//
//		Description:	Reads all data pending in the serial RX buffer
//                      and places it in a local buffer.
//
//		Returns:		number of characters read
//
int fnSerialReadRxData(void)
{
	// Read data, until there is nothing left
	char acTempBuffer[MAX_RX_RESP_BUFFER];

	DWORD dwTotalBytesRead = 0;
	DWORD dwBytesRead      = 0;
	int   iIndex           = 0;

	LONG lLastError;

	memset(acSerialRxBuffer, 0, sizeof(acSerialRxBuffer));

	do
	{
		// We leave the last 8 bytes unused in each buffer just to make sure nothing ever goes past the end of the buffer.

		// Read data from the COM-port
		lLastError = pCMcrSerialInterface->mcrSerial.Read(acTempBuffer,sizeof(acTempBuffer)-8,&dwBytesRead);
		#ifdef _DEBUG_ERROR
		if (lLastError != ERROR_SUCCESS)
		{
			TCHAR tcDebugBuff[256];
			lLastError = pCMcrSerialInterface->mcrSerial.GetLastError();
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Unable to read from COM-port: %ld"), lLastError);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		if ((dwBytesRead > 0) && (dwTotalBytesRead < (sizeof(acSerialRxBuffer)-8)))
		{
			memcpy(&acSerialRxBuffer[dwTotalBytesRead], acTempBuffer, dwBytesRead);
			dwTotalBytesRead += dwBytesRead;
		}
	}
	while (dwBytesRead > 0);

	return (int)dwTotalBytesRead;
}





//
//		Name:			fnSerialProcessRxData()
//
//		Description:	Processes the data we receive from the AVR.
//                      Triggers events and/or saves the data as needed.
//
//		Returns:		none
//

void fnSerialProcessRxData(void)
{
	int ii, iTotalBytesRead;
	char cNextCharacter;

	iTotalBytesRead = fnSerialReadRxData();

	if (iTotalBytesRead)
	{
		for (ii=0; ii<iTotalBytesRead; ii++)
		{
			cNextCharacter = acSerialRxBuffer[ii];

			switch (iCurrentRxState)
			{
				case WAITING_FOR_COMMAND:
					switch (cNextCharacter)
					{
						case DATE_TIME_REQUEST:
							// send the response
							fnProcessDATE_TIME_REQUEST();
							break;

						case DATE_TIME_RESPONSE:
							// wait for the rest of the packet
							cCurrentCommand = DATE_TIME_RESPONSE;
							iCurrentRxState = WAITING_FOR_LEFT_DELIMITER;
							break;

						case PC_STATUS_RESPONSE:
							// wait for the rest of the packet
							cCurrentCommand = PC_STATUS_RESPONSE;
							iCurrentRxState = WAITING_FOR_LEFT_DELIMITER;
							break;

						case AVR_NMEA_PACKET:
							// wait for the rest of the packet
							cCurrentCommand = AVR_NMEA_PACKET;
							iCurrentRxState = WAITING_FOR_LEFT_DELIMITER;
							break;

						case AVR_NOTIFY_ALARM:
							// wait for the rest of the packet
							cCurrentCommand = AVR_NOTIFY_ALARM;
							iCurrentRxState = WAITING_FOR_LEFT_DELIMITER;
							break;

						case AVR_FRAME_ALARM:
							fnSerialTimeCodeFrameAlarm();
							break;

						case AVR_NOTIFY_TOP_OF_SECOND:
							// read and save the current high performance counter value
							// performs the same function as toggling the modem control line
							fnSerialInterruptDsr();
							break;


						case PC_READ_TIME_CODE_RESPONSE:
							// wait for the rest of the packet
							cCurrentCommand = PC_READ_TIME_CODE_RESPONSE;
							iCurrentRxState = WAITING_FOR_LEFT_DELIMITER;
							break;
					}
					break;

				case WAITING_FOR_LEFT_DELIMITER:
					if (cNextCharacter == '<')
						iCurrentRxState = WAITING_FOR_DATA;
					else
					{
						// out of sync, start over
						#ifdef _DEBUG_ERROR
						{
							TCHAR tcDebugBuff[256];
							swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialProcessRxData() out of sync"));
							cDebugOutput->DebugMsg(tcDebugBuff);
						}
						#endif

						cCurrentCommand = 0;
						iCurrentRxState = WAITING_FOR_COMMAND;
						iNextIndex = 0;
						memset(acSerialRxResponse, 0, sizeof(acSerialRxResponse));
					}
					break;

				case WAITING_FOR_DATA:
					if (cNextCharacter == '>')
					{
						fnSerialProcessRxPacket();
						cCurrentCommand = 0;
						iCurrentRxState = WAITING_FOR_COMMAND;
						iNextIndex      = 0;
						memset(acSerialRxResponse, 0, sizeof(acSerialRxResponse));
					} else
					{
						// save the character
						acSerialRxResponse[iNextIndex++] = cNextCharacter;

						if (iNextIndex >= (sizeof(acSerialRxResponse)-1))
						{
							// buffer overflow, start over

							#ifdef _DEBUG_ERROR
							{
								TCHAR tcDebugBuff[256];
								swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialProcessRxData() response bufffer overflow"));
								cDebugOutput->DebugMsg(tcDebugBuff);
							}
							#endif

							cCurrentCommand = 0;
							iCurrentRxState = WAITING_FOR_COMMAND;
							iNextIndex      = 0;
							iTotalBytesRead = 0;
							memset(acSerialRxResponse, 0, sizeof(acSerialRxResponse));
						}
					}
					break;

				default:
					#ifdef _DEBUG_ERROR
					{
						TCHAR tcDebugBuff[256];
						swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnSerialProcessRxData() invalid command state"));
						cDebugOutput->DebugMsg(tcDebugBuff);
					}
					#endif

					// Should never get here, set back to waiting for a command
					cCurrentCommand = 0;
					iCurrentRxState = WAITING_FOR_COMMAND;
					iNextIndex      = 0;
					iTotalBytesRead = 0;
					memset(acSerialRxResponse, 0, sizeof(acSerialRxResponse));

					break;
			}
		}
	}
}





//
//		Name:			fnSerialProcessRxPacket()
//
//		Description:	We have a comma delimited response packet. We know what command
//                      the data is for, process and save the data.
//
//		Returns:		none
//
void fnSerialProcessRxPacket(void)
{
	switch (cCurrentCommand)
	{
		case DATE_TIME_RESPONSE:
			fnProcessDATE_TIME_RESPONSE();


			if (g_cSerialWaitEventTimeCodeFrameResponse == DATE_TIME_RESPONSE)
			{
				g_cSerialWaitEventTimeCodeFrameResponse = 0;
				SetSerialEventTimeCodeFrameResponse();
			}

			if (g_cSerialWaitEventTopOfSecond == DATE_TIME_RESPONSE)
			{
				g_cSerialWaitEventTopOfSecond = 0;
				SetSerialEventTopOfSecondData();
			}

			if (g_cSerialWaitEventResponse == DATE_TIME_RESPONSE)
			{
				g_cSerialWaitEventResponse = 0;
				SetSerialEventResponse();
			}
			break;

		case AVR_NOTIFY_ALARM:
		case PC_STATUS_RESPONSE:
			fnProcessAVR_NOTIFY_ALARM();

			if (g_cSerialWaitEventTimeCodeFrameResponse == PC_STATUS_RESPONSE)
			{
				g_cSerialWaitEventTimeCodeFrameResponse = 0;
				SetSerialEventTimeCodeFrameResponse();
			}

			if (g_cSerialWaitEventTopOfSecond == PC_STATUS_RESPONSE)
			{
				g_cSerialWaitEventTopOfSecond = 0;
				SetSerialEventTopOfSecondData();
			}


			if (g_cSerialWaitEventResponse == PC_STATUS_RESPONSE)
			{
				g_cSerialWaitEventResponse = 0;
				SetSerialEventResponse();
			}
			break;

		case PC_READ_TIME_CODE_RESPONSE:
			fnProcessPC_READ_TIME_CODE_RESPONSE();

			if (g_cSerialWaitEventTimeCodeFrameResponse == PC_READ_TIME_CODE_RESPONSE)
			{
				g_cSerialWaitEventTimeCodeFrameResponse = 0;
				SetSerialEventTimeCodeFrameResponse();
			}

			if (g_cSerialWaitEventTopOfSecond == PC_READ_TIME_CODE_RESPONSE)
			{
				g_cSerialWaitEventTopOfSecond = 0;
				SetSerialEventTopOfSecondData();
			}

			if (g_cSerialWaitEventResponse == PC_READ_TIME_CODE_RESPONSE)
			{
				g_cSerialWaitEventResponse = 0;
				SetSerialEventResponse();
			}
			break;

		case AVR_NMEA_PACKET:
			// write the NMEA packet to the configured output port
			DWORD dwLength;

			dwLength = (DWORD)strlen(acSerialRxResponse);

			// NMEA packets can never be this long
			if (dwLength < 100)
			{
				if (pCMcrSerialInterface->nmeaSerial.IsOpen())
					pCMcrSerialInterface->nmeaSerial.Write(acSerialRxResponse, &dwLength);
			} else
			{
				#ifdef _DEBUG_ERROR
				{
					cDebugOutput->DebugMsg(_T("!!!NMEA length error"));
				}
				#endif
			}

			#ifdef _DEBUG_NMEA
			{
				int ii;
				TCHAR tcDebugBuff[256];

				// Only print the first 5 NMEA message characters
				// must convert char to TCHAR
				//for (ii=0; ii<strlen(acSerialRxResponse); ii++)
				for (ii=0; ii<6; ii++)
				{
					tcDebugBuff[ii] = acSerialRxResponse[ii];
				}
				tcDebugBuff[ii] = 0; //make sure it is null terminated
				//swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("%s"), acSerialRxResponse);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif

			break;

		default:
			// unknown command
			break;
	}
}






//
//		Name:			fnProcessDATE_TIME_REQUEST()
//
//		Description:	The AVR requested the current PC date and time.
//                      Send the data to the AVR.
//
//		Returns:		
//
void fnProcessDATE_TIME_REQUEST(void)
{
	SYSTEMTIME srSystemTime;
	char tcTimeBuff[128];
	DWORD dwBytesWrite;

	AcquireSerialEventMutex();

	// Write the current data/time to the PCIe serial port
	// Note: due to Windows time slicing and how the time gets updated, this could be
	//       off by as much as one system time slice tick (probably 15.9 msec).
	GetSystemTime(&srSystemTime);


	sprintf_s(tcTimeBuff,sizeof(tcTimeBuff),"%c<%d,%d,%d,%d,%d,%d,%d>",
		DATE_TIME_RESPONSE,
		srSystemTime.wMonth, srSystemTime.wDay, srSystemTime.wYear,
		srSystemTime.wHour, srSystemTime.wMinute, srSystemTime.wSecond, srSystemTime.wMilliseconds);

	dwBytesWrite = (DWORD)strlen(tcTimeBuff);
	pCMcrSerialInterface->mcrSerial.Write(tcTimeBuff, &dwBytesWrite);

	ReleaseSerialEventMutex();

}







//
//		Name:			fnProcessDATE_TIME_RESPONSE()
//
//		Description:	We requested the Date and Time from the AVR. Process the response.	
//
//		Returns:		
//
void fnProcessDATE_TIME_RESPONSE(void)
{
	WCATIME tempDateTime;

	int  iCurrentIndex = 0;
	int  iNextValue = 0;
	BOOL bError = false;

	memset(&tempDateTime, 0, sizeof(tempDateTime));

	// save the month
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wMonth = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 12)) // crude validity check
		bError = true;

	// save the day
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wDay = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 31)) // crude validity check
		bError = true;

	// save the year
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wYear = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue < 1900) || (iNextValue > 2100)) // crude validity check
		bError = true;


	// save the hour
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wHour = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 24)) // crude validity check
		bError = true;


	// save the minute
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wMinute = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 60)) // crude validity check
		bError = true;

	// save the second
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempDateTime.wSecond = (unsigned short)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 60)) // crude validity check
		bError = true;

	// save the milleseconds
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	if (iNextValue < 0)
		bError = true;

	if (iNextValue >= 1000)
		bError = true;

	tempDateTime.wMilliseconds = (unsigned short)iNextValue;

	if (bError)
	{
		// send the NAK
		fnSendNAK(DATE_TIME_RESPONSE, BAD_PARAMETER);
	} else
	{
		// send the ACK
		fnSendACK();

		// save the data
		g_LastDateTimeResponse = tempDateTime;
	}
}





//
//		Name:			fnProcessAVR_NOTIFY_ALARM()
//
//		Description:	The AVR sent us an Alarm message. Process it.	
//
//		Returns:		
//
void  fnProcessAVR_NOTIFY_ALARM(void)
{
	LAST_ALARM_STATUS tempAlarmStatus;

	int  iCurrentIndex = 0;
	int  iNextValue = 0;
	BOOL bError = false;

	memset(&tempAlarmStatus, 0, sizeof(tempAlarmStatus));

	// The index is incremented by fnGetIntegerValue().

	// ucLockStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucLockStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > VALID_ALARM_ENTRY)) // crude validity check - WARNING - new code may send the value over the limit. Extend the limit if needed. That define has no other use.
		bError = true;

	// ucRtcStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucRtcStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > VALID_ALARM_ENTRY)) // crude validity check
		bError = true;

	// ucGpsStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucGpsStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > VALID_ALARM_ENTRY)) // crude validity check
		bError = true;

	// ucTimeCodeStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucTimeCodeStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > VALID_ALARM_ENTRY)) // crude validity check
		bError = true;

	// ucHsoStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucHsoStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > VALID_ALARM_ENTRY)) // crude validity check
		bError = true;

	// ucHsoAchievementLevel
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucHsoAchievementLevel = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	// ucReserved1
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucReserved1 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// ucReserved2
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucReserved2 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// ucReserved3
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucReserved3 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// ucReserved4
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.srLockStatus.ucReserved4 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// ucHwStatus
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.ucHwStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	// version major
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.verMajor = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	// version minor
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.verMinor = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	// version revision
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.verRevision = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	// reserved
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.ucReserved1 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// reserved
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.ucReserved2 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// reserved
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.ucReserved3 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	// reserved
	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempAlarmStatus.ucReserved4 = (UCHAR)iNextValue;
	//if (iNextValue < 0) // crude validity check
	//	bError = true;

	if (bError)
	{
		// send the NAK
		fnSendNAK(AVR_NOTIFY_ALARM, BAD_PARAMETER);
	} else
	{
		// send the ACK
		fnSendACK();

		// save the data
		g_LastAlarmStatusResponse = tempAlarmStatus;
	}

}









//
//		Name:			fnProcessPC_READ_TIME_CODE_RESPONSE()
//
//		Description:	Process the time code response.	
//
//		Returns:		
//
void fnProcessPC_READ_TIME_CODE_RESPONSE(void)
{
	TIME_CODE_FRAME_COMPLETE tempTimeCodeResponse;
	LARGE_INTEGER tempPerfCountValue;
	int  ii, iNumberBytes;
	int  iCurrentIndex = 0;
	int  iNextValue = 0;
	BOOL bError = false;

	memset(&tempTimeCodeResponse, 0, sizeof(tempTimeCodeResponse));

	// Save the value in case the high performance counter was read before this was started

	// compare the two copies in case the interrupt modified it during a copy;
	tempPerfCountValue.QuadPart                = g_liPerfCountValueTopOfSecond.QuadPart;
	tempTimeCodeResponse.intPerfCount.QuadPart = g_liPerfCountValueTopOfSecond.QuadPart;

	if (tempPerfCountValue.QuadPart != g_liPerfCountValueTopOfSecond.QuadPart)
	{
		// it changed, do it one more time
		tempTimeCodeResponse.intPerfCount.QuadPart = g_liPerfCountValueTopOfSecond.QuadPart;
	}

	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempTimeCodeResponse.tcStatus = (UCHAR)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
	tempTimeCodeResponse.hwStatus = (UINT8)iNextValue;
	if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		bError = true;

	iNumberBytes = sizeof(tempTimeCodeResponse.tcFrame);
	for (ii=0; ii<iNumberBytes; ii++)
	{
		iNextValue = fnGetIntegerValue(&acSerialRxResponse[iCurrentIndex], &iCurrentIndex);
		tempTimeCodeResponse.tcFrame[ii] = (UCHAR)iNextValue;
		if ((iNextValue < 0) || (iNextValue > 255)) // crude validity check
		{
			bError = true;
			break;
		}
	}

	if (bError)
	{
		// send the NAK
		fnSendNAK(DATE_TIME_RESPONSE, BAD_PARAMETER);
	} else
	{
		// send the ACK
		fnSendACK();

		// save the data
		tempTimeCodeResponse.srLockStatus = g_LastAlarmStatusResponse.srLockStatus;

		g_LastTimeCodeResponse = tempTimeCodeResponse;

	}
}



//
//		Name:			fnSendNAK()
//
//		Description:	Send a NAK to the AVR. While some of the functions
//                      will send this. The ACK/NAK has not been completely implemented.
//
//		Returns:		
//
void fnSendNAK(char cCommand, int iReason)
{
	DWORD dwBytesWrite;
	char cNakBuff[16];

	AcquireSerialEventMutex();

	sprintf_s(cNakBuff, sizeof(cNakBuff), "%c<%c,%d>", PC_NAK, cCommand, iReason);

	dwBytesWrite = (DWORD)strlen(cNakBuff);
	pCMcrSerialInterface->mcrSerial.Write(cNakBuff, &dwBytesWrite); // PC_NAK

	ReleaseSerialEventMutex();

	#ifdef _DEBUG_ERROR
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!PC_NAK AVR data:%d"), iReason);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif
}




//
//		Name:			fnSendACK()
//
//		Description:	Send a ACK to the AVR. While some of the functions
//                      will send this. The ACK/NAK has not been completely implemented.	
//
//		Returns:		
//
void fnSendACK(void)
{
	DWORD dwBytesWrite;
	char cAckBuff[16];

	AcquireSerialEventMutex();

	sprintf_s(cAckBuff, sizeof(cAckBuff), "%c", PC_ACK);

	dwBytesWrite = (DWORD)strlen(cAckBuff);
	pCMcrSerialInterface->mcrSerial.Write(cAckBuff, &dwBytesWrite); // PC_ACK

	ReleaseSerialEventMutex();
}







//
//		Name:			fnSerialTimeCodeFrameAlarm()
//
//		Description:	Do the processing when a time code frame alarm is received.	
//
//		Returns:		
//
void fnSerialTimeCodeFrameAlarm(void)
{
	HANDLE  hProcessHandle;
	DWORD   dwPriorityClass;

	//QueryPerformanceFrequency(&perfFreq);
	hProcessHandle  = GetCurrentProcess();
	dwPriorityClass = GetPriorityClass(hProcessHandle);

	SetPriorityClass(hProcessHandle,HIGH_PRIORITY_CLASS);
	//CRITICAL_SECTION	m_csEvents;
	//EnterCriticalSection(&m_csEvents);
	QueryPerformanceCounter(&g_liPerfCountValueFrame);

	// restore priority execution
	//LeaveCriticalSection(&m_csEvents);
	SetPriorityClass(hProcessHandle,dwPriorityClass);

	SetSerialEventTimeCodeFrameAlarm();
}


//
//		Name:			fnSerialInterruptDsr()
//
//		Description:	At top of second, the AVR will toggle the DSR modem control line.
//                      That triggers this interrupt. We capture the high performance counter
//                      so that we can calculate the PC delays and derive an accurate time.
//
//		Returns:		none
//
void fnSerialInterruptDsr(void)
{
	LARGE_INTEGER newPerfCount;
	LARGE_INTEGER perfFreq;
	DWORD         msDiff;

	HANDLE  hProcessHandle;
	DWORD   dwPriorityClass;
	BOOL    bIgnore;

	bIgnore = false;

	//QueryPerformanceFrequency(&perfFreq);
	hProcessHandle  = GetCurrentProcess();
	dwPriorityClass = GetPriorityClass(hProcessHandle);

	SetPriorityClass(hProcessHandle,HIGH_PRIORITY_CLASS);
	//CRITICAL_SECTION	m_csEvents;
	//EnterCriticalSection(&m_csEvents);

	QueryPerformanceFrequency(&perfFreq);

	QueryPerformanceCounter(&newPerfCount);

	if (newPerfCount.QuadPart > g_liPerfCountValueTopOfSecond.QuadPart)
	{
		msDiff = (ULONG)((newPerfCount.QuadPart - g_liPerfCountValueTopOfSecond.QuadPart) / (perfFreq.QuadPart / (__int64)1000));

		if (msDiff < 300)
		{
			// ignore glitches or top of second pulses which are too close together
			bIgnore = true;

			//#ifdef _DEBUG_ERROR
			//{
			//	TCHAR tcDebugBuff[256];
			//	swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Ignore TOS: %d"), msDiff);
			//	cDebugOutput->DebugMsg(tcDebugBuff);
			//}
			//#endif
		}
	}

	if (!bIgnore)
	{
		pCMcrSerialInterface->mcrSerial.SetDTR();

		g_liPerfCountValueTopOfSecond = newPerfCount;

		SetSerialEventTopOfSecondStart();

		// restore priority execution
		//LeaveCriticalSection(&m_csEvents);
		SetPriorityClass(hProcessHandle,dwPriorityClass);

		Sleep(2);

		pCMcrSerialInterface->mcrSerial.ClearDTR();
	} else
	{
		// restore priority execution
		//LeaveCriticalSection(&m_csEvents);
		SetPriorityClass(hProcessHandle,dwPriorityClass);
	}
}















//
//		Name:			McrSerialInterruptDSR_GetTime
//
//		Description:	The AVR will toggle the Data Set Ready (DSR)
//                      modem control line at top of second. This causes
//                      an event in fnSerialEventMonitorThread() - CSerial::EEventDSR.
//                      We read and save the current contents of the high performance
//                      counter when this event occurs.
//
//		Returns MCR_SUCCESS (0), or an error code
//
INT McrSerialInterruptDSR_GetTime(LARGE_INTEGER *pliPerfCountIn)
{
	pliPerfCountIn->QuadPart = g_liPerfCountValueTopOfSecond.QuadPart;

	return MCR_SUCCESS;
}





//
//		Name:			fnReadTimeCodeFrame()
//
//		Description:	Reads time code frame from the AVR. This is the
//                      complete time code frame. NOT the time code frame saved in TIME_CODE_FRAME_REV_A
//                      which is only a part of a time code frame.
//
//                      This function is called from a thread and waits for the serial
//                      thread to process the response.
//
//		Returns:		TRUE if time code successfully processed
//						FALSE if no time code present
//
BOOL  fnReadTimeCodeFrame(TIME_CODE_FRAME_COMPLETE *psrPreviousSecondTimeFrame)
{
	BOOL bStatus;
	char tcCmdBuff[8];
	DWORD dwBytesWrite;

	AcquireSerialEventMutex();

	g_cSerialWaitEventTimeCodeFrameResponse = PC_READ_TIME_CODE_RESPONSE; // we are waiting for this response
	sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c",PC_READ_TIME_CODE_REQUEST);

	dwBytesWrite = (DWORD)strlen(tcCmdBuff);
	pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

	ReleaseSerialEventMutex();

	bStatus = WaitForSerialEventTimeCodeFrameResponse(FALSE);

	*psrPreviousSecondTimeFrame = g_LastTimeCodeResponse;

	// wait for the response
	if (!bStatus)
	{
		// we received the event
		return MCR_SUCCESS;
	} else
	{
		// event was not received
		return MCR_FAIL;
	}
}


//
//		Name:			fnValidateComPort()
//
//		Description:	Test the COM port selected for communications. If the currently
//                      selected COM port is not valid, scan the COM ports
//                      looking for a COM port that commujnicates with the AVR. If
//                      scanning is successful, then scan for the internal, virtual
//                      COM port connected to the GPS receiver NMEA port.
//
//		Returns:		true if communications are successful, false if not.
//
BOOL  fnValidateComPort(CMcrSerialInterface *pCMcrSerialInterface, CSoftwareConfig *pswConfig)
{
	//TCHAR tcPortFomattedWithPath[64];
	//TCHAR tcPortFomattedCom[64];
	int iPCIeComPortNumber, iNmeaComPortNumber, iTempComPortNumber;
	int iPrimaryComIndex, iNmeaComIndex;
	BOOL bIsValid;


	iPCIeComPortNumber = pswConfig->GetRegistrySerialPortNum();
	if (iPCIeComPortNumber)
	{
		bIsValid = fnIsComPortValid(pCMcrSerialInterface, iPCIeComPortNumber);
		if (bIsValid)
		{
			// The COM port number in the registry for PCIe communications is valid.
			// Assume the NMEA port is also valid.
			return true;
		}
	}

	// If the COM port number in the registry is not valid, we need to scan the COM ports to find a valid port.
	// Once a valid COM port is found, we also save the COM port number to use for NMEA output.

	// Clear the area used to save the Pericom COM ports we have found.
	iNumberPericomPortsFound = 0;
	for (iPrimaryComIndex=0; iPrimaryComIndex<MAX_PERICOM_PORTS_FOUND; iPrimaryComIndex++)
	{
		cPrimaryPericomPortFound[iPrimaryComIndex].m_nPortNum = 0;
	}

	// The EnumCallbackWrapper will save the Pericom COM ports it finds in the cPrimaryPericomPortFound array.
	pCMcrSerialInterface->EnumAllDevices();

	for (iPrimaryComIndex=0; iPrimaryComIndex<iNumberPericomPortsFound; iPrimaryComIndex++)
	{
		// remember to clear the  bits identifying the type of port it is
		iPCIeComPortNumber = cPrimaryPericomPortFound[iPrimaryComIndex].m_nPortNum & DATA_PORT_MASK;

		if (fnIsComPortValid(pCMcrSerialInterface, iPCIeComPortNumber))
		{
			// We have found a valid COM port
			break;
		}
	}


	if (iPrimaryComIndex == iNumberPericomPortsFound)
	{
		// we did not find a valid Pericom PCIe COM port to use
		return false;
	}

	// we need the port to the AVR open so we can send the NMEA commands
	fnOpenComPort(&pCMcrSerialInterface->mcrSerial, iPCIeComPortNumber, 115200, 8, NMEA_PARITY_NONE, 1);

	// Using the valid COM port as a reference, select the NMEA COM port to use
	// Scan the list looking for a Pericom port that has a COM port number that is
	// one higher or one lower than the iComPortNumber we found.
	iNmeaComPortNumber = 0;
	for (iNmeaComIndex=0; iNmeaComIndex<iNumberPericomPortsFound; iNmeaComIndex++)
	{
		// remember to clear the  bits identifying the type of port it is
		iTempComPortNumber = cPrimaryPericomPortFound[iNmeaComIndex].m_nPortNum & DATA_PORT_MASK;
		if (iPCIeComPortNumber != iTempComPortNumber)
		{
			if (fnIsNmeaComPortValid(pCMcrSerialInterface, iTempComPortNumber))
			{
				iNmeaComPortNumber = iTempComPortNumber;
				break;
			}
		}
	}

	if (pCMcrSerialInterface->mcrSerial.IsOpen())
		pCMcrSerialInterface->mcrSerial.csClose();

	if (iPCIeComPortNumber)
	{
		// save the primary Pericom COM port number
		pswConfig->SetRegistrySerialPortNum(iPCIeComPortNumber);

		// save the Pericom COM NMEA port number to use when the virtual output on this card is selected
		pswConfig->SetRegistryNmeaVirtualSerialPortNum(iNmeaComPortNumber);
	}

	return true;
}



//
//		Name:			fnIsComPortValid()
//
//		Description:	Tests if the selected COM port is connected to the AVR.	
//
//		Returns:		true if communications are successful, false if not.
//
BOOL fnIsComPortValid(CMcrSerialInterface *pCMcrSerialInterface, int iComPortNumber)
{
	BOOL  bIsValid;
	int   ii, iWaitState;
	DWORD dwBytesRead, dwBytesWrite;
	char  tempBuff[8];

	// send a command and wait up to one second for a response
	if (pCMcrSerialInterface->mcrSerial.IsOpen())
	{
		// At this time this should not have any devices open.
		// Flag an error.
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!COM port was already open: %d"), iComPortNumber);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		pCMcrSerialInterface->mcrSerial.csClose();
		//return false;
	}

	// open the com port to send/receive data
	// events connected to the COM port are not initialized.
	if (!fnSerialComPortInit(&pCMcrSerialInterface->mcrSerial, iComPortNumber))
	{
		if (pCMcrSerialInterface->mcrSerial.IsOpen())
			pCMcrSerialInterface->mcrSerial.csClose();

		// error opening the port, return an error
		return false;
	}

	bIsValid   = true;
	iWaitState = 0; // waiting for the first character

	// write the request
	sprintf_s(tempBuff,sizeof(tempBuff),"%c",PC_STATUS_REQUEST);

	dwBytesWrite = (DWORD)strlen(tempBuff);
	pCMcrSerialInterface->mcrSerial.Write(tempBuff, &dwBytesWrite);

	// wait up to 500 msec for the port to respond with the first two characters of valid data
	for (ii=0; ii<500;)
	{
		pCMcrSerialInterface->mcrSerial.Read(tempBuff, 1, &dwBytesRead);
		if (dwBytesRead)
		{
			if (iWaitState == 0)
			{
				// waiting for the first character of the response, DATE_TIME_RESPONSE ('b')
				if (tempBuff[0] == PC_STATUS_RESPONSE)
				{
					// we got the first character of the response
					// The next character must be a '<'
					iWaitState = 1;
				}
			} else
			{
				// waiting for the second character of the response, '<'
				if (tempBuff[0] == '<')
				{
					// we got the second character of the response, '<'
					// we have good communications with the AVR

					break;
				} else
				{
					// did not get the next character, start over
					iWaitState = 0;
				}

			}

		} else
		{
			ii++;
			Sleep(1);
		}
	}

	if (pCMcrSerialInterface->mcrSerial.IsOpen())
		pCMcrSerialInterface->mcrSerial.csClose();

	if (ii >= 500)
	{
		// we timed out waiting, return error
		return false;
	}
	
	return true;
}




//
//		Name:			fnIsNmeaComPortValid()
//
//		Description:	Test the NMEA COM port selected for communications. Do this by waiting for a "$GP"
//                      NOTE:: this requires that the user does not have an application
//                             that has already connected to the port.
//
//		Returns:		true if communications are successful, false if not.
//
#define NMEA_DELAY 2000
BOOL fnIsNmeaComPortValid(CMcrSerialInterface *pCMcrSerialInterface, int iComPortNumber)
{
	BOOL  bIsValid;
	int   ii, iWaitState;
	DWORD dwBytesRead;
	char  tempBuff[8];

	CSoftwareConfig	*pswConfig;

	pswConfig = new CSoftwareConfig(cDebugOutput);

	// send a command and wait up to one second for a response
	if (pCMcrSerialInterface->nmeaSerial.IsOpen())
	{
		// At this time this should not have any devices open.
		// Flag an error.
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnIsNmeaComPortValid() NMEA COM port was already open: %d"), iComPortNumber);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		pCMcrSerialInterface->nmeaSerial.csClose();
		//return false;
	}

	// open the com port to send/receive data
	// events connected to the COM port are not initialized.
	// Open the serial port, if we get an error and cPrimaryPericomPortFound is different
	// then our current value, try using the value from cPrimaryPericomPortFound
	if (!fnOpenComPort(&pCMcrSerialInterface->nmeaSerial, iComPortNumber, pswConfig->m_srKeys.dwNmeaBaudrate, pswConfig->m_srKeys.dwNmeaDataBits, pswConfig->m_srKeys.dwNmeaParity, pswConfig->m_srKeys.dwNmeaStopBits))
	{
		// At this time this should not have any devices open.
		// Flag an error.
		#ifdef _DEBUG_ERROR
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnIsNmeaComPortValid(), NMEA COM port, error opening port: %d"), iComPortNumber);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		delete pswConfig;
		return false;
	}

	// turn on the NMEA message(s) we want. We really don't care what we receive. We just want the $GP at the start.
	// The GGA and VTG are always output with or without lock.
	fnConfigureNmeaMessages(&pCMcrSerialInterface->mcrSerial, pswConfig->m_srKeys.dwNmeaBaudrate, NMEA_MESSAGE_BIT_GGA | NMEA_MESSAGE_BIT_VTG, 1);

	bIsValid   = true;
	iWaitState = 0; // waiting for the first character

	// wait up to 5000 msec for the port to respond with the first two characters of valid data
	for (ii=0; ii<NMEA_DELAY;)
	{
		pCMcrSerialInterface->nmeaSerial.Read(tempBuff, 1, &dwBytesRead);
		if (dwBytesRead)
		{
			switch (iWaitState)
			{
				case 0:
					// waiting for the first character of the response, DATE_TIME_RESPONSE ('b')
					if (tempBuff[0] == '$')
					{
						// we got the first character of the response
						// The next character must be a 'G'
						iWaitState = 1;
					}
					break;

				case 1:
					// waiting for the first character of the response, DATE_TIME_RESPONSE ('b')
					if (tempBuff[0] == 'G')
					{
						// we got the first character of the response
						// The next character must be a 'G'
						iWaitState = 2;
					} else
					{
						iWaitState = 0;
					}
					break;

				case 2:
					// waiting for the first character of the response, DATE_TIME_RESPONSE ('b')
					if (tempBuff[0] == 'P')
					{
						// we got the third character of the response
						// done
						iWaitState = 3; // successful
					} else
					{
						iWaitState = 0;
					}
					break;
			}

			if (iWaitState == 3)
			{
				break; // successful
			}

		} else
		{
			ii++;
			Sleep(1);
		}
	}

	// turn off all NMEA messages
	fnConfigureNmeaMessages(&pCMcrSerialInterface->mcrSerial, pswConfig->m_srKeys.dwNmeaBaudrate, 0, 1);

	delete pswConfig;

	if (pCMcrSerialInterface->nmeaSerial.IsOpen())
		pCMcrSerialInterface->nmeaSerial.csClose();

	if (ii >= NMEA_DELAY)
	{
		// we timed out waiting, return error
		return false;
	}

	return true;
}





//
//		Name:			fnOpenComPort()
//
//		Description:	Opens the selected port, sets the baud rate, data bits, parity, and stop bits.	
//
//		Returns:		true if successful, false if not
//
BOOL  fnOpenComPort(CSerial *pCSerial, int iComPortNumber, int iBaudRate, int iDataBits, int iParity, int iStopBits)
{
	TCHAR tcPortFomattedWithPath[64];
	TCHAR tcPortFomattedCom[64];
	LONG lLastError;

	// open the com port to send/receive data
	// events connected to the COM port are not initialized.

	// Attempt to open the serial port
	swprintf_s(tcPortFomattedWithPath,sizeof(tcPortFomattedWithPath)/sizeof(TCHAR),_T("\\\\.\\COM%d"), iComPortNumber);
	swprintf_s(tcPortFomattedCom,sizeof(tcPortFomattedCom)/sizeof(TCHAR),_T("COM%d"), iComPortNumber);

	lLastError = pCSerial->csOpen(tcPortFomattedCom,tcPortFomattedWithPath,0,0,true);

	if (lLastError != ERROR_SUCCESS)
	{
		#ifdef _DEBUG_ERROR
		TCHAR tcDebugBuff[256];
		LONG lLastError;
		lLastError = pCSerial->GetLastError();
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnOpenComPort - Unable to open COM port %s: %ld"), tcPortFomattedCom, lLastError);
		cDebugOutput->DebugMsg(tcDebugBuff);
		#endif

		return false;
	}

	if (pCSerial->IsOpen())
	{
		// Setup the serial port
		CSerial::EBaudrate eBaudrate;
		CSerial::EDataBits eDataBits;
		CSerial::EParity   eParity;
		CSerial::EStopBits eStopBits;

		switch (iBaudRate)
		{
		case 4800:
			eBaudrate = CSerial::EBaud4800;
			break;

		case 9600:
			eBaudrate = CSerial::EBaud9600;
			break;

		case 19200:
			eBaudrate = CSerial::EBaud19200;
			break;

		case 38400:
			eBaudrate = CSerial::EBaud38400;
			break;

		case 57600:
			eBaudrate = CSerial::EBaud57600;
			break;

		case 115200:
			eBaudrate = CSerial::EBaud115200;
			break;

		default:
			eBaudrate = CSerial::EBaud9600;
			break;
		}

		switch (iDataBits)
		{
		case 7:
			eDataBits = CSerial::EData7;
			break;

		case 8:
			eDataBits = CSerial::EData8;
			break;

		default:
			eDataBits = CSerial::EData8;
			break;
		}

		switch (iParity)
		{
		case NMEA_PARITY_NONE:
			eParity   = CSerial::EParNone;
			break;

		case NMEA_PARITY_ODD:
			eParity   = CSerial::EParOdd;
			break;

		case NMEA_PARITY_EVEN:
			eParity   = CSerial::EParEven;
			break;

		default:
			eParity   = CSerial::EParNone;
			break;
		}


		switch (iStopBits)
		{
		case 1:
			eStopBits = CSerial::EStop1;
			break;

		case 2:
			eStopBits = CSerial::EStop2;
			break;

		default:
			eStopBits = CSerial::EStop1;
			break;
		}

		lLastError = pCSerial->csSetup(eBaudrate,eDataBits,eParity,eStopBits);
		if (lLastError != ERROR_SUCCESS)
		{
			#ifdef _DEBUG_ERROR
			TCHAR tcDebugBuff[256];
			LONG lLastError;
			lLastError = pCSerial->GetLastError();
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!fnOpenComPort: error setting port parameters port:%d error:%ld"), iComPortNumber, lLastError);
			cDebugOutput->DebugMsg(tcDebugBuff);
			#endif

			//delete pswConfig;
			return false;
		}
	}

	//delete pswConfig;
	return true;
}






//
//		Name:			fnConfigureNmeaMessages()
//
//		Description:	Assumes the serial port used to communicate with the AVR is open.
//                      Configures the NMEA messages for the GPS receiver to output, if the messages
//                      are to be output on the GPS receiver secondary port (ie. NMEA port), it sets
//                      the ports baud rate and tells the GPS receiver to output the messages on the secondary port.
//
//		Returns:		none, Assumes sending the message is successful
//
void  fnConfigureNmeaMessages(CSerial *pCSerialAvr, int iBaudRate, DWORD dwNmeaMessagesBitMap, int iIsVirtualComPort)
{
	char cConfigBuff[128];
	DWORD dwBytesWrite;

	// Set NMEA configuration
	AcquireSerialEventMutex();
	{
		int iZDA,iGGA,iGLL,iVTG,iGSA,iGSV,iGST,iRMC;
		int iVirtual, iNmeaOnOff;

		iZDA = iGGA = iGLL = iVTG = iGSA = iGSV = iGST = iRMC = 0;
		iVirtual = iNmeaOnOff = 0;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_ZDA)
			iZDA = 1;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_GGA)
			iGGA = 2;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_GLL)
			iGLL = 3;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_VTG)
			iVTG = 4;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_GSA)
			iGSA = 5;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_GSV)
			iGSV = 6;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_GST)
			iGST = 7;

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_BIT_RMC)
			iRMC = 8;

		// did they select the virtual com port on this PCIe card?
		if (iIsVirtualComPort)
		{
			iVirtual = 1;
		}

		if (dwNmeaMessagesBitMap & NMEA_MESSAGE_MASK)
		{
			iNmeaOnOff = 1;
		}

		sprintf_s(cConfigBuff,sizeof(cConfigBuff),"%c<%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d>",
			PC_SET_NMEA_CONFIG,
			iVirtual,
			iBaudRate,
			iNmeaOnOff,
			iZDA, iGGA, iGLL, iVTG, iGSA, iGSV, iGST, iRMC);

		dwBytesWrite = (DWORD)strlen(cConfigBuff);
		pCSerialAvr->Write(cConfigBuff, &dwBytesWrite);

	}
	ReleaseSerialEventMutex();
	Sleep(100); // Make sure the AVR has a time to process the data

}





//
//		Name:			fnRecalibrate()
//
//		Description:	Tells the AVR to recalibrate time code.
//
//		Returns:		FALSE if successfully processed
//						!= 0, failure
//
BOOL  fnRecalibrate(void)
{
	char tcCmdBuff[8];
	DWORD dwBytesWrite;

	AcquireSerialEventMutex();

	sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c",PC_RECALIBRATE_REQUEST);

	dwBytesWrite = (DWORD)strlen(tcCmdBuff);
	pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

	ReleaseSerialEventMutex();

	return MCR_SUCCESS;
}





//
//		Name:			fnSetDateTime()
//
//		Description:	Tells the AVR to set it's date and time to this value.
//
//		Returns:		FALSE if successfully processed
//						!= 0, failure
//
BOOL  fnSetDateTime(MCRSYNC_IPC_SETTIME *psrSetTime)
{
	char tcCmdBuff[256];
	DWORD dwBytesWrite;

	AcquireSerialEventMutex();

	sprintf_s(tcCmdBuff,sizeof(tcCmdBuff),"%c<%d,%d,%d,%d,%d,%d,%d>",
		PC_SET_DATE_TIME,
		psrSetTime->utcTimeToSet.wMonth,
		psrSetTime->utcTimeToSet.wDay,
		psrSetTime->utcTimeToSet.wYear,
		psrSetTime->utcTimeToSet.wHour,
		psrSetTime->utcTimeToSet.wMinute,
		psrSetTime->utcTimeToSet.wSecond,
		psrSetTime->utcTimeToSet.wMilliseconds);

	dwBytesWrite = (DWORD)strlen(tcCmdBuff);
	pCMcrSerialInterface->mcrSerial.Write(tcCmdBuff, &dwBytesWrite);

	ReleaseSerialEventMutex();

	return MCR_SUCCESS;
}




//
//		Name:			fnGetIntegerValue()
//
//		Description:	returns the next integer value in the comma delimited string
//
//		Returns:		none
//
int fnGetIntegerValue(char *piBuffer, int *piIndex)
{
	int  ii;
	int  iCurrentIndex = 0;
	int  iNextValue = -1;
	char cValueBuffer[32];

	memset(cValueBuffer, 0, sizeof(cValueBuffer));
	for (ii=0; ii<(sizeof(cValueBuffer)-1); ii++)
	{
		cValueBuffer[ii] = *piBuffer++;

		if (!((cValueBuffer[ii] >= '0') && (cValueBuffer[ii] <= '9')))
		{
			// we stop on anything that is not a '0' through '9'
			cValueBuffer[ii] = 0;
			break;
		}
	}

	// update the index with the number of characters we processed
	*piIndex = *piIndex + ii + 1;

	if (ii > 0)
	{
		// if we found any data, convert the ascii string to an integer
		iNextValue = atoi(cValueBuffer);
	}

	return iNextValue; 
}












//
//		Name:			fnGetStringValue()
//
//		Description:	returns the next string value in the comma delimited string
//                      saved in piBufferDest and the number of characters in the string
//
//		Returns:		none
//
int fnGetStringValue(char *piBufferDest, int iMaxLengthDest, char *piBufferSrc, int *piIndex)
{
	int  ii;
	char cNextChar;

	for (ii=0; ii<iMaxLengthDest; ii++)
	{
		cNextChar = *piBufferSrc++;

		// it must be a printable character
		if (!((cNextChar >= ' ') && (cNextChar >= '~')))
		{
			// we stop on anything that is not a printable character
			break;
		}

		// comma delimited, stop on a comma.
		if (cNextChar == ',')
		{
			// we stop on a comma
			break;
		}

		// Our data packet delimiters are '<' and '>', stop on either on of those
		if ((cNextChar == '<') || (cNextChar == '>'))
		{
			// we stop on a packet delimiter
			break;
		}

		if (ii >= iMaxLengthDest-1)
		{
			// we ran out of room
			break;
		}

		// save the character
		*piBufferDest++ = cNextChar;

	}

	// null terminate the string we saved
	*piBufferDest++ = 0;

	// update the index with the number of characters we processed
	*piIndex = *piIndex + ii + 1;

	return ii; 
}















