
// Mcr MonitorDlg.cpp : implementation file
//

#include "stdafx.h"
#include "mcrcommon.h"
#include "Mcr Monitor.h"
#include "Mcr MonitorDlg.h"
#include "afxdialogex.h"
#include "PipeInterface.h"
#include "DebugOutput.h"
#include "SetTimeAndDate.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

BYTE month[12][4] = 
{
	"Jan","Feb","Mar","Apr",
	"May","Jun","Jul","Aug",
	"Sep","Oct","Nov","Dec"
};

BYTE dayofweek[7][4] = 
{
	"Sun","Mon","Tue","Wed",
	"Thu","Fri","Sat"
};


// Functions for dynamically loading and using the MCR DLL
#ifdef _DEBUG_DLL
HINSTANCE hMcrInstanceLib = NULL;

boolean fnLoadMcrDll(HINSTANCE *hMcrInstanceLib); // must use double indirection
boolean fnFreeMcrDll(HINSTANCE hMcrInstanceLib);

void  MCR_ApiGetDLLVersionDynamic(HINSTANCE hMcrInstanceLib, UINT *pMajor, UINT *pMinor, UINT *pRevision);
#endif

void fnAddOneSecondSystime(SYSTEMTIME *pTime);
void fnAddOneSecondWcatime(WCATIME *pTime);

// CAboutDlg dialog used for App About

class CAboutDlg : public CDialogEx
{
public:
	CAboutDlg();

// Dialog Data
	enum { IDD = IDD_ABOUTBOX };

	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support

// Implementation
protected:
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialogEx(CAboutDlg::IDD)
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialogEx::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialogEx)
END_MESSAGE_MAP()


// CMcrMonitorDlg dialog




CMcrMonitorDlg::CMcrMonitorDlg(CWnd* pParent /*=NULL*/)
	: CDialogEx(CMcrMonitorDlg::IDD, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	curInterruptFreq = INTERRUPT_FREQ_PERSEC;

	hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_SETINTFREQ,
					(PVOID)&curInterruptFreq,
					sizeof(DWORD),
					NULL,
					0);


	ZeroMemory(&syncHist,sizeof(MCRSYNC_IPC_SYNCHIST));
	ZeroMemory(&syncHistOld,sizeof(MCRSYNC_IPC_SYNCHIST));

}

CMcrMonitorDlg::~CMcrMonitorDlg()
{
#ifdef _DEBUG_DLL
	if (hMcrInstanceLib)
	{
		fnFreeMcrDll(hMcrInstanceLib);
	}
#endif


}


void CMcrMonitorDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialogEx::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_SYNCHIST, m_SyncHist);
	DDX_Control(pDX, IDC_MONITORING2, m_Monitoring);
	DDX_Control(pDX, IDC_MCRREAD_TIME2, display_McrRead_Time);
	DDX_Control(pDX, IDC_MCRREAD_DATE2, display_McrRead_Date);
	DDX_Control(pDX, IDC_UTC_TIME2, display_UTC_Time);
	DDX_Control(pDX, IDC_UTC_DATE2, display_UTC_Date);
	DDX_Control(pDX, IDC_SYSTEM_TIME2, display_System_Time);
	DDX_Control(pDX, IDC_SYSTEM_DATE2, display_System_Date);
	DDX_Control(pDX, IDC_SYNCVER2, display_mcrsync_Version);
	DDX_Control(pDX, IDC_SYNCSTATUS2, display_Time_Status);
	DDX_Control(pDX, IDC_FIRMWARE_NUMBERS, display_mcrsync_FirmwareVersion);
	DDX_Control(pDX, IDC_HSO_STATUS, display_mcrsync_HSO_Status);
	DDX_Control(pDX, IDC_HSO_LEVEL, display_mcrsync_HSO_Level);
}

BEGIN_MESSAGE_MAP(CMcrMonitorDlg, CDialogEx)
	ON_WM_SYSCOMMAND()
	ON_WM_DESTROY()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_TIMER()
	ON_WM_HELPINFO()
	ON_BN_CLICKED(ID_RECALIBRATE, &CMcrMonitorDlg::OnBnClickedRecalibrate)
	ON_BN_CLICKED(IDOK, &CMcrMonitorDlg::OnBnClickedOk)
	ON_BN_CLICKED(IDCANCEL, &CMcrMonitorDlg::OnBnClickedCancel)
	ON_BN_CLICKED(IDC_SET_TIME, &CMcrMonitorDlg::OnBnClickedSetTime)
END_MESSAGE_MAP()


// CMcrMonitorDlg message handlers

BOOL CMcrMonitorDlg::OnInitDialog()
{
	char tmp[256];

	CDialogEx::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		BOOL bNameValid;
		CString strAboutMenu;
		bNameValid = strAboutMenu.LoadString(IDS_ABOUTBOX);
		ASSERT(bNameValid);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// TODO: Add extra initialization here
	if (hPipeInterface->monitorSystemIsLocal)
		m_Monitoring.SetWindowText("the local system.");
	else
	{
		sprintf_s(tmp,sizeof(tmp),"\\\\%s",hPipeInterface->monitorSystem);
		m_Monitoring.SetWindowText(tmp);
	}


	hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_VERSION,
					NULL,
					0,
					(PVOID)&syncVersion,
					sizeof(MCRSYNC_IPC_VERSION));


	sprintf_s(tmp,sizeof(tmp),"%u.%u",syncVersion.verMaj,syncVersion.verMin);
	display_mcrsync_Version.SetWindowText(tmp);




	// kpw - do pipe to get firmware version from service. copied from above.
	hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_FIRMWARE_VERSION,
					NULL,
					0,
					(PVOID)&syncFirmwareVersion,
					sizeof(MCRSYNC_IPCCMD_FIRMWARE_VERSION));

	sprintf_s( tmp,sizeof(tmp), "%u.%u.%u", syncFirmwareVersion.verMaj, syncFirmwareVersion.verMin, syncFirmwareVersion.verRev );
	display_mcrsync_FirmwareVersion.SetWindowText(tmp);




	// include for testing two applications accessing the .dll concurrently
#ifdef _DEBUG_DLL
	{
		HINSTANCE *indirect;

		hMcrInstanceLib = NULL;

		indirect = &hMcrInstanceLib;

		//m_bServiceIsRunning = TRUE;


		if (!fnLoadMcrDll(indirect))
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!Failed loading MCR API dll");
			#endif
		} else
		{
			if (hMcrInstanceLib)
			{
				#ifdef _DEBUG_INFO
				DebugMsg("MCR API dll successfully loaded");
				#endif
			} else
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!MCR API dll says it loaded, but the pointer is a null");
				#endif
			}					
		}
	}
#endif


	// this timer runs 20 times per second.
	// but, the only time we actually do something is after we see the seconds on system time change
	SetTimer(1,50,NULL);
	

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CMcrMonitorDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialogEx::OnSysCommand(nID, lParam);
	}
}

void CMcrMonitorDlg::OnDestroy()
{
//	WinHelp(0L, HELP_QUIT);
	CDialog::OnDestroy();
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CMcrMonitorDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialogEx::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CMcrMonitorDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}


void CMcrMonitorDlg::OnTimer(UINT_PTR nIDEvent) 
{
	static SYSTEMTIME	sysTimeNew, sysTimePrevious; 
	char	tmp[256];
	DWORD	t;

	DWORD tcType, tcSubType;

	#ifdef _DEBUG_INFO
	DebugMsg("Timer event");
	#endif

	if (nIDEvent == 1)
	{

		// is it time to do something?
		// this timer runs 20 times per second.
		// but, the only time we actually do something is after we see the seconds on system time change
		GetSystemTime(&sysTimeNew);

		if (sysTimeNew.wSecond == sysTimePrevious.wSecond)
		{
			return;
		}

		if (sysTimeNew.wMilliseconds < 10)
		{
			// wait until after top of second
			return;
		}

		sysTimePrevious = sysTimeNew;


		hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_CURTIME,
						NULL,
						0,
						(PVOID)&syncCurTime,
						sizeof(MCRSYNC_IPC_CURTIME));


		//#ifdef _DEBUG_INFO
		//		DebugMsg("%02u:%02u:%02u %02u/%02u/%04u Projected time = %lu",
		//				syncCurTime.rawTime.wHour,
		//				syncCurTime.rawTime.wMinute,
		//				syncCurTime.rawTime.wSecond,
		//				syncCurTime.rawTime.wMonth,
		//				syncCurTime.rawTime.wDay,
		//				syncCurTime.rawTime.wYear,
		//				syncCurTime.projectedTime);
		//#endif

		// show time code reader time/date in window

		//fnAddOneSecondSystime(&syncCurTime.localTime);
		//fnAddOneSecondWcatime(&syncCurTime.rawTime);
		//fnAddOneSecondWcatime(&syncCurTime.utcTime);

		if (syncCurTime.timeStatus == CURTIME_INVALID)
		{
			tmp[0] = '\0';
			display_McrRead_Time.SetWindowText(tmp);
			display_McrRead_Date.SetWindowText(tmp);
			display_UTC_Time.SetWindowText(tmp);
			display_UTC_Date.SetWindowText(tmp);

			sprintf_s(tmp,sizeof(tmp),"%02d:%02d:%02d",
				syncCurTime.localTime.wHour,
				syncCurTime.localTime.wMinute,
				syncCurTime.localTime.wSecond);

			display_System_Time.SetWindowText(tmp);

			sprintf_s(tmp,sizeof(tmp),"%s, %s %02d, %04d",
				dayofweek[syncCurTime.localTime.wDayOfWeek],
				month[syncCurTime.localTime.wMonth-1],
				syncCurTime.localTime.wDay,
				syncCurTime.localTime.wYear);

			display_System_Date.SetWindowText(tmp);

			display_Time_Status.SetWindowText("UNHEALTHY (Time has never synchronized)");
		}
		else
		{
			if (syncCurTime.timeStatus == CURTIME_HEALTHY)
			{
				sprintf_s(tmp,sizeof(tmp),"%02d:%02d:%02d",
					syncCurTime.rawTime.wHour,
					syncCurTime.rawTime.wMinute,
					syncCurTime.rawTime.wSecond);

				display_McrRead_Time.SetWindowText(tmp);

				if (!syncCurTime.rawTime.wMonth &&
					!syncCurTime.rawTime.wDay &&
					!syncCurTime.rawTime.wYear)
				{
					sprintf_s(tmp,sizeof(tmp),"    N/A");
				}
				else
				{
					sprintf_s(tmp,sizeof(tmp),"%d/%d/%d",
						syncCurTime.rawTime.wMonth,
						syncCurTime.rawTime.wDay,
						syncCurTime.rawTime.wYear);
				}

				display_McrRead_Date.SetWindowText(tmp);

				sprintf_s(tmp,sizeof(tmp),"%02u:%02u:%02u",
					syncCurTime.utcTime.wHour,
					syncCurTime.utcTime.wMinute,
					syncCurTime.utcTime.wSecond);

				display_UTC_Time.SetWindowText(tmp);

				//sprintf_s(tmp,sizeof(tmp),"%s, %s %02d, %04d",
				//	dayofweek[syncCurTime.utcTime.wDayOfWeek],
				//	month[syncCurTime.utcTime.wMonth-1],
				//	syncCurTime.utcTime.wDay,
				//	syncCurTime.utcTime.wYear);

				sprintf_s(tmp,sizeof(tmp),"%d/%d/%d",
					syncCurTime.utcTime.wMonth,
					syncCurTime.utcTime.wDay,
					syncCurTime.utcTime.wYear);
							
				display_UTC_Date.SetWindowText(tmp);
			}
			else
			{
				tmp[0] = '\0';
				display_McrRead_Time.SetWindowText(tmp);
				display_McrRead_Date.SetWindowText(tmp);
				display_UTC_Time.SetWindowText(tmp);
				display_UTC_Date.SetWindowText(tmp);
			}

			sprintf_s(tmp,sizeof(tmp),"%02d:%02d:%02d",
				syncCurTime.localTime.wHour,
				syncCurTime.localTime.wMinute,
				syncCurTime.localTime.wSecond);

			display_System_Time.SetWindowText(tmp);

			sprintf_s(tmp,sizeof(tmp),"%s, %s %d, %d",
				dayofweek[syncCurTime.localTime.wDayOfWeek],
				month[syncCurTime.localTime.wMonth-1],
				syncCurTime.localTime.wDay,
				syncCurTime.localTime.wYear);

			display_System_Date.SetWindowText(tmp);

			//hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_FIRMWARE_VERSION,
			//		NULL,
			//		0,
			//		(PVOID)&syncFirmwareVersion,
			//		sizeof(MCRSYNC_IPCCMD_FIRMWARE_VERSION));

			// kpw - ask for status info.
			hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_LOCK_STATUS,
				NULL,
				0,
				(PVOID)&syncLockStatus,
				sizeof(MCRSYNC_IPCCMD_LOCK_STATUS));

			// kpw - interpret and show the hso info:
			// The defines are from the AVR code.
			// From examining the AVR code, it appears that only 2 and 3 were originally returned.
			// I added the not-installed.
			switch( syncLockStatus.ucHsoStatus )
			{
			case LOCK_INVALID:
				sprintf_s(tmp,sizeof(tmp), "%s", "Never Locked");
				break;

			case LOCK_NOT_PRESENT:
				sprintf_s(tmp,sizeof(tmp), "%s", "Lock Not Present");
				break;

			case LOCK_LOST: // 2 This is in use.
				sprintf_s(tmp,sizeof(tmp), "%s", "Not Locked");
				//sprintf_s(tmp,sizeof(tmp), "%s", "Lost Lock");
				break;

			case LOCK_RESTORED: // 3 This is in use.
				sprintf_s(tmp,sizeof(tmp), "%s", "Locked");
				//sprintf_s(tmp,sizeof(tmp), "%s", "Lock Restored");
				break;

			case LOCK_PRESENT:
				sprintf_s(tmp,sizeof(tmp), "%s", "Locked");
				break;

			case HSO_NOT_INSTALLED: // 5 This is in use.
				sprintf_s(tmp,sizeof(tmp), "%s", "Not Installed");
				break;
			}

			//sprintf_s(tmp,sizeof(tmp),"%d", syncLockStatus.ucHsoStatus);
			display_mcrsync_HSO_Status.SetWindowText(tmp);

			sprintf_s(tmp,sizeof(tmp),"%d", syncLockStatus.ucHsoAchievementLevel);
			display_mcrsync_HSO_Level.SetWindowText(tmp);
			// kpw end


			hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_INFO,
				NULL,
				0,
				(PVOID)&syncInfo,
				sizeof(MCRSYNC_IPC_INFO));



			switch(syncCurTime.timeStatus)
			{
				case CURTIME_HEALTHY:
					sprintf_s(tmp, sizeof(tmp), "HEALTHY");

					tcType = syncInfo.TimeCodeType & TC_TYPE_MASK;

					// note:: the device might not return a sub type
					tcSubType = syncInfo.TimeCodeType & TC_SUBTYPE_MASK;

					if (tcType == TC_TYPE_SMPTE)
					{
						strcat_s(tmp, sizeof(tmp), ", SMPTE time code");
						if (tcSubType == TC_SUBTYPE_SMPTE30FPS)
						{
							strcat_s(tmp, sizeof(tmp), ", 30 fps");
						} else
						if (tcSubType == TC_SUBTYPE_EBU25FPS)
						{
							strcat_s(tmp, sizeof(tmp), ", 25 fps");
						} else
						if (tcSubType == TC_SUBTYPE_FILM24FPS)
						{
							strcat_s(tmp, sizeof(tmp), ", 24 fps");
						}
					} else
					if (tcType == TC_TYPE_IRIG) // also TC_TYPE_IRIG_B1
					{
						strcat_s(tmp, sizeof(tmp), ", IRIG B1 (Amplitude Modulated)");
					} else
					if (tcType == TC_TYPE_IRIG_B)
					{
						strcat_s(tmp, sizeof(tmp), ", IRIG B0 (Pulse Width Coded)");
					}

					display_Time_Status.SetWindowText(tmp);
					break;

				case CURTIME_UNHEALTHY:
					strcpy_s(tmp,sizeof(tmp),"UNHEALTHY ");
					switch(syncCurTime.extStatus)
					{
						case REASON_TIMELOSS:
							strcat_s(tmp,sizeof(tmp),"(Time code loss)");
							break;

						case REASON_SERIOUSTIMELOSS:
							strcat_s(tmp,sizeof(tmp),"(Serious time code loss)");
							break;

						case REASON_GPSLOSS:
							strcat_s(tmp,sizeof(tmp),"(Time code generator is freewheeling)");
							break;

						case REASON_SERIOUSGPSLOSS:
							strcat_s(tmp,sizeof(tmp),"(Serious time code generator lock problem)");
							break;
					}

					display_Time_Status.SetWindowText(tmp);
					break;
			}

		}

		hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_SYNCHIST,
			NULL,
			0,
			(PVOID)&syncHist,
			sizeof(MCRSYNC_IPC_SYNCHIST));



		if (syncHist.numSyncRecords)
		{
			if (memcmp(&syncHist,&syncHistOld,sizeof(MCRSYNC_IPC_SYNCHIST)))
			{
				// the synchronization history list has changed

				m_SyncHist.ResetContent();

				t = syncHist.numSyncRecords;

				do {
					t--;

					sprintf_s(tmp,sizeof(tmp),"(S)%02u:%02u:%02u.%03u %02u/%02u/%04u <- (T)%02u:%02u:%02u.%03u %02u/%02u/%04u",
						syncHist.systemTime[t].wHour,syncHist.systemTime[t].wMinute,syncHist.systemTime[t].wSecond,syncHist.systemTime[t].wMilliseconds,
						syncHist.systemTime[t].wMonth,syncHist.systemTime[t].wDay,syncHist.systemTime[t].wYear,
						syncHist.utcTime[t].wHour,syncHist.utcTime[t].wMinute,syncHist.utcTime[t].wSecond,syncHist.utcTime[t].wMilliseconds,
						syncHist.utcTime[t].wMonth,syncHist.utcTime[t].wDay,syncHist.utcTime[t].wYear);
					
					m_SyncHist.AddString(tmp);
				
				} while(t);
			}
		}

		// include for testing having two applications accessing the .dll concurrently
#ifdef _DEBUG_DLL
		UINT Major, Minor, Revision;
		if (hMcrInstanceLib)
		{
			//MCR_ApiGetDLLVersionDynamic(hMcrInstanceLib, &Major, &Minor, &Revision);
			//DebugMsg("DLL version: %d.%d.%d",Major, Minor, Revision);
		} else
		{
			DebugMsg("!!!MCR DLL version not read, dll not loaded");
		}

#endif
	}
	
	CDialog::OnTimer(nIDEvent);
}

void fnAddOneSecondSystime(SYSTEMTIME *pTime)
{

	pTime->wSecond++;
	if (pTime->wSecond > 59)
	{
		pTime->wSecond = 0;

		pTime->wMinute++;

		if (pTime->wMinute > 59)
		{
			pTime->wMinute = 0;

			pTime->wHour++;

			if (pTime->wHour > 23)
			{
				// So we don't have to account for day/month/year rollover,
				// once a day at 23:59:59 we don't add one second
				pTime->wSecond = 59;
				pTime->wMinute = 59;
				pTime->wHour   = 23;
			}
		}
	}
}


void fnAddOneSecondWcatime(WCATIME *pTime)
{

	pTime->wSecond++;
	if (pTime->wSecond > 59)
	{
		pTime->wSecond = 0;

		pTime->wMinute++;

		if (pTime->wMinute > 59)
		{
			pTime->wMinute = 0;

			pTime->wHour++;

				if (pTime->wHour > 23)
				{
					// So we don't have to account for day/month/year rollover,
					// once a day at 23:59:59 we don't add one second
					pTime->wSecond = 59;
					pTime->wMinute = 59;
					pTime->wHour   = 23;
				}
		}
	}
}

#ifdef _DEBUG_DLL
//
// This function load the MCR DLL into memory
// When using a CLR based project, the DLL must be loaded dynamically.
//
boolean fnLoadMcrDll(HINSTANCE *hinstLib)
{
	HINSTANCE hLocalInstLib;

	boolean bLoadIsSuccessful = true;

	// kjf xxx yyy zzz this will need to use a path determined at run time or, 
	// we must force the end user to install the DLL into a known location
	// This is done using two lines of code to make it easier to understand the double indirection

#ifdef _DEBUG
	hLocalInstLib = LoadLibrary("C:\\mcr\\Mcr API\\Mcr API\\Mcr API\\Debug\\x64\\mcr_x64_rev_a.dll"); 
#else
	hLocalInstLib = LoadLibrary("C:\\mcr\\Mcr API\\Mcr API\\Mcr API\\Release\\mcr_rev_a.dll"); 
#endif

	//hLocalInstLib = LoadLibrary(_T("C:\\MCR\\McrSyncservice and API vc10\\Debug\\mcr_rev_a.dll"));
	*hinstLib     = hLocalInstLib;

	// If the handle is valid, try to get the function address.

	if (*hinstLib == NULL) 
	{ 
		// error loading the DLL
		bLoadIsSuccessful = false;
	}

	return bLoadIsSuccessful;
}

//
// This function load the MCR DLL into memory
// When using a CLR based project, the DLL must be loaded dynamically.
//
boolean fnFreeMcrDll(HINSTANCE hinstLib)
{
	BOOL fFreeIsSuccessful;

	// Free the DLL module.

	fFreeIsSuccessful = FreeLibrary(hinstLib);

	if (!fFreeIsSuccessful)
	{
		// error freeing DLL
	}

	return fFreeIsSuccessful;
}


//void  __stdcall MCR_ApiGetDLLVersion(PUINT verMaj, PUINT verMin, PUINT verRev)
typedef void (__stdcall *MYPROC2)(UINT *Major, UINT *Minor, UINT *Revision);

void MCR_ApiGetDLLVersionDynamic(HINSTANCE hinstLib, UINT *pMajor, UINT *pMinor, UINT *pRevision)
{
	MYPROC2 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE;

	// If the handle is valid, try to get the function address.

	if (hinstLib != NULL) 
	{ 
		ProcAdd = (MYPROC2) GetProcAddress(hinstLib, "MCR_ApiGetDLLVersion"); 

		// If the function address is valid, call the function.

		if (NULL != ProcAdd) 
		{
			fRunTimeLinkSuccess = TRUE;
			(ProcAdd)(pMajor, pMinor, pRevision); // this is the dll call 
		} 
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 
}
#endif

void CMcrMonitorDlg::OnBnClickedRecalibrate()
{

		DWORD dwData = 0;

		hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_RECALIBRATE,
			(PVOID)&dwData,
			sizeof(DWORD),
			NULL,
			0);

		MessageBox("Recalibration may take up to 2 minutes to complete.\r\n\r\nPlease wait at least two minutes before clicking the Recalibration button again.", "Recalibration");

}


void CMcrMonitorDlg::OnBnClickedOk()
{
	// TODO: Add your control notification handler code here
	CDialogEx::OnOK();
}


void CMcrMonitorDlg::OnBnClickedCancel()
{
	// TODO: Add your control notification handler code here
	CDialogEx::OnCancel();
}


void CMcrMonitorDlg::OnBnClickedSetTime()
{
	bool bIsInvalid;
	CSetTimeAndDate *pSetDateTimedlg;
	MCRSYNC_IPC_SETTIME srSetTime;

	pSetDateTimedlg = new CSetTimeAndDate();

	INT_PTR nResponse = pSetDateTimedlg->DoModal();
	if (nResponse == IDOK)
	{
		memset(&srSetTime, 0, sizeof(srSetTime));

		//  dismissed with OK

		// get the time and set the PCIe time with it
		srSetTime.utcTimeToSet.wIsDstActive  = false;
		srSetTime.utcTimeToSet.wDayOfWeek    = 0;
		srSetTime.utcTimeToSet.wMilliseconds = 0;

		srSetTime.utcTimeToSet.wHour         = pSetDateTimedlg->iHour;
		srSetTime.utcTimeToSet.wMinute       = pSetDateTimedlg->iMinute;
		srSetTime.utcTimeToSet.wSecond       = pSetDateTimedlg->iSecond;

		srSetTime.utcTimeToSet.wMonth        = pSetDateTimedlg->iMonth;
		srSetTime.utcTimeToSet.wDay          = pSetDateTimedlg->iDayOfMonth;
		srSetTime.utcTimeToSet.wYear         = pSetDateTimedlg->iYear;

		// do a crude check
		bIsInvalid = false;

		if ((srSetTime.utcTimeToSet.wHour < 0) || (srSetTime.utcTimeToSet.wHour > 23))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if ((srSetTime.utcTimeToSet.wMinute < 0) || (srSetTime.utcTimeToSet.wMinute > 59))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if ((srSetTime.utcTimeToSet.wSecond < 0) || (srSetTime.utcTimeToSet.wSecond > 59))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if ((srSetTime.utcTimeToSet.wMonth < 1) || (srSetTime.utcTimeToSet.wMonth > 12))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if ((srSetTime.utcTimeToSet.wDay < 1) || (srSetTime.utcTimeToSet.wDay > 31))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if ((srSetTime.utcTimeToSet.wYear < 2000) || (srSetTime.utcTimeToSet.wYear > 2050))
		{
			bIsInvalid = true;
			MessageBox(NULL, "Invalid hour, enter 0-23", MB_ICONWARNING);
		}

		if (!bIsInvalid)
		{
			hPipeInterface->PipeTransaction(MCRSYNC_IPCCMD_SETTIME,
				(PVOID)&srSetTime,
				sizeof(MCRSYNC_IPC_SETTIME),
				NULL,
				0);
		}
	}
	else if (nResponse == IDCANCEL)
	{
		//  dismissed with Cancel
	}

	delete pSetDateTimedlg;
}
