#include <string.h>
#include <wtypes.h>
#include <winbase.h>
#include <winreg.h>
#include "RegClass.h"
#include "DebugOutput.h"

//
//		Name:			::CClarkReg()
//
//		Description:	Constructor
//

CClarkReg::CClarkReg()
{
	key1 = CRYPT_KEY1;
	key2 = CRYPT_KEY2;
}


//
//		Name:			::OpenRegFile()
//
//		Description:	Load in the registration information binary.
//
//		Returns:		One of the REGIS_xx statuses for the target executable.
//

WORD CClarkReg::OpenRegFile(char *fspec)
{
	DWORD	checksum, bytesRead;
	HANDLE	regHandle;
	BOOL	resp;

	regHandle = CreateFile(	fspec,
							GENERIC_READ,
							0,
							NULL,
							OPEN_EXISTING,
							0,
							NULL);


	if (regHandle == INVALID_HANDLE_VALUE) return(REGIS_LOADERR);

	resp = ReadFile(regHandle,
					&regInfo,
					sizeof(CLARKREGINFO),
					&bytesRead,
					NULL);

	CloseHandle(regHandle);

	if (!resp || (bytesRead != sizeof(CLARKREGINFO)))
		return(REGIS_CORRUPT);
	
	switch(regInfo.isregistered)
	{
		case STAMP_REGISTERED:
	
			// verify checksum on the registration information
			
			checksum = 0;

			ChecksumString((char *)&regInfo.user,&checksum,USER_SIZE);
			ChecksumString((char *)&regInfo.org,&checksum,ORG_SIZE);
			ChecksumString((char *)&regInfo.originUser,&checksum,USER_SIZE);
			ChecksumString((char *)&regInfo.originOrg,&checksum,ORG_SIZE);

			if (checksum != regInfo.checksum) 
				return(REGIS_CORRUPT);

			memcpy(user,&regInfo.user,USER_SIZE);
			CryptString(user,USER_SIZE);

			memcpy(org,&regInfo.org,ORG_SIZE);
			CryptString(org,ORG_SIZE);

			memcpy(originUser,&regInfo.originUser,USER_SIZE);
			CryptString(originUser,USER_SIZE);

			memcpy(originOrg,&regInfo.originOrg,ORG_SIZE);
			CryptString(originOrg,ORG_SIZE);

			return(REGIS_YES);

		case STAMP_UNREGISTERED:
			return(REGIS_NO);

		default:
			return(REGIS_CORRUPT);
	}
}


//
//		Name:			::WriteRegistration()
//
//		Description:	Write registration information
//
//		Returns:		TRUE on success
//

BOOL CClarkReg::WriteRegistration(char *fspec, char *user, char *org)
{
	HKEY		enumKey;
	char		buffer[512];
	DWORD		bufferSize;
	DWORD		keyType;
	DWORD		checksum;
	HANDLE		tmpFile, regHandle;
	char		keyPath[256];
	BOOL		resp;
	DWORD		bytesWritten;

	if (GetVersion() & 0x80000000)
		strcpy_s(keyPath,sizeof(keyPath),"SOFTWARE\\Microsoft\\Windows\\CurrentVersion");
	else
		strcpy_s(keyPath,sizeof(keyPath),"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion");

	// write a temporary file to insure we have write access to drive
	// containing file to register

	tmpFile = CreateFile(	"@@@@@@@@.WCA",
							GENERIC_WRITE,
							0,
							NULL,
							CREATE_ALWAYS,
							FILE_ATTRIBUTE_NORMAL,
							NULL);
	
	if (tmpFile == INVALID_HANDLE_VALUE) return(FALSE);

	CloseHandle(tmpFile);

	DeleteFile("@@@@@@@@.WCA");

	// retrieve computer name and organization from registry

	RegOpenKeyEx(HKEY_LOCAL_MACHINE,keyPath,0,KEY_READ,&enumKey);

	bufferSize = sizeof(buffer);
	keyType = REG_SZ;

	if (RegQueryValueEx(enumKey,
						"RegisteredOwner",
						NULL,
						&keyType,
						(BYTE *)buffer,
						&bufferSize) == ERROR_SUCCESS)
	{
		memcpy(originUser,buffer,USER_SIZE);
	}
	else
		strcpy_s(originUser,sizeof(originUser),"[None Found]");

	bufferSize = sizeof(buffer);
	keyType = REG_SZ;

	if (RegQueryValueEx(enumKey,
						"RegisteredOrganization",
						NULL,
						&keyType,
						(BYTE *)buffer,
						&bufferSize) == ERROR_SUCCESS)
	{
		memcpy(originOrg,buffer,ORG_SIZE);
	}
	else
		strcpy_s(originOrg,sizeof(originOrg),"[None Found]");

	// fill in our registration structure

	strncpy_s((char *)&regInfo.user,sizeof(regInfo.user),user,USER_SIZE);
	strncpy_s((char *)&regInfo.org,sizeof(regInfo.org),org,ORG_SIZE);
	strncpy_s((char *)&regInfo.originUser,sizeof(regInfo.originUser),originUser,USER_SIZE);
	strncpy_s((char *)&regInfo.originOrg,sizeof(regInfo.originOrg),originOrg,ORG_SIZE);

	regInfo.isregistered = STAMP_REGISTERED;

	// encrypt all the registration info

	CryptString((char *)&regInfo.user,USER_SIZE);
	CryptString((char *)&regInfo.org,ORG_SIZE);
	CryptString((char *)&regInfo.originUser,USER_SIZE);
	CryptString((char *)&regInfo.originOrg,ORG_SIZE);

	// build checksum based on encrypted registration info

	checksum = 0;

	ChecksumString((char *)regInfo.user,&checksum,USER_SIZE);
	ChecksumString((char *)regInfo.org,&checksum,ORG_SIZE);
	ChecksumString((char *)regInfo.originUser,&checksum,USER_SIZE);
	ChecksumString((char *)regInfo.originOrg,&checksum,ORG_SIZE);

	regInfo.checksum = checksum;

	// update registration with new info

	regHandle = CreateFile(	fspec,
							GENERIC_WRITE,
							0,
							NULL,
							OPEN_EXISTING,
							0,
							NULL);

	if (regHandle == INVALID_HANDLE_VALUE) return(FALSE);

	resp = WriteFile(	regHandle,
						&regInfo,
						sizeof(CLARKREGINFO),
						&bytesWritten,
						NULL);

	CloseHandle(regHandle);

	if (!resp || (bytesWritten != sizeof(CLARKREGINFO)))
		return(FALSE);

	return(TRUE);
}


//
//		Name:			::CryptString()
//
//		Description:	Simple encryption/decryption routine.  Passing a plaintext
//						string to this routine encrypts it.  Passing a string
//						already encrypted once by CryptString() decrypts it.
//

void CClarkReg::CryptString(char *s, WORD size)
{
	WORD t;
	char *keyp;

	for(keyp = key1, t = 0; t < size; t++)
	{
		s[t] ^= *keyp;

		keyp++;

		if ((size_t)(keyp - key1) > (strlen(key1)-1)) keyp = key1;
	}

	for(keyp = key2, t = 0; t < size; t++)
	{
		s[t] ^= *keyp;

		keyp++;

		if ((size_t)(keyp - key2) > (strlen(key2)-1)) keyp = key2;
	}
}


//
//		Name:			::ChecksumString()
//
//		Description:	Creates simple doubleword checksum of a string of characters.
//						
//		Note:			Does not initialize checksum value
//

void CClarkReg::ChecksumString(char *s, DWORD *checksum, WORD size)
{
	WORD t1, t2;

	for(t1 = 0, t2 = 0; t1 < size; t1++)
	{
		*(((BYTE *)checksum)+t2) ^= s[t1];
	
		if (++t2 == sizeof(DWORD)) t2 = 0;
	}
}
