#pragma once


#include "tzlib.h"
//#include <wtypes.h>

typedef struct _REGISTRY_TIMEZONE_INFO
{
	LONG				Bias;
	LONG				StandardBias;
	LONG				DaylightBias;
	SYSTEMTIME			StandardDate;
	SYSTEMTIME			DaylightDate;

} REGISTRY_TIMEZONE_INFO, *PREGISTRY_TIMEZONE_INFO;

//#define SERVICE_KERNEL_DRIVER_NAME		"TCRDrv"
//#define SERVICE_USER_DRIVER_NAME		"McrSync_REV_A"
#define REGISTRY_SOURCE_NAME			(_T("McrSyncConfiguration"))

// MCR products
#define MCR_PRODUCT_REV_A        		0x00000100L
#define MCR_PRODUCT_UNDEFINED			0xffffffffL

// time code products

//#define TC_PRODUCT_TCR200				0x00000001L
//#define TC_PRODUCT_TCR200I			0x00000002L
//#define TC_PRODUCT_TCR200E			0x00000003L
//#define TC_PRODUCT_TCR500				0x00000004L
//#define TC_PRODUCT_TCR500E			0x00000005L
//#define TC_PRODUCT_TCR500PCI			0x00000006L
//#define TC_PRODUCT_TCR1000			0x00000007L
#define TC_PRODUCT_MCR_REV_A		0x00000008L
#define TC_PRODUCT_MCR_X64_REV_A	0x00000009L
#define TC_PRODUCT_UNDEFINED			0xffffffffL

// time code sources

//#define TC_SOURCE_GPS100				0x00000001L	// GPS-100 master clock
//#define TC_SOURCE_OTHER				0x00000002L	// ...any other
//#define TC_SOURCE_GPS200				0x00000003L	// GPS-200/200A
//#define TC_SOURCE_TCI232				0x00000004L // TCI-232
//#define TC_SOURCE_TCI232A				0x00000005L	// TCI-500/TCI-232A
//#define TC_SOURCE_TCO100				0x00000006L // TCO-100
// The defines above are no longer used. Instead define the
// information included with the time code.
#define TC_SOURCE_DATE_NOT_INCLUDED                   0x00000007L
#define TC_SOURCE_DATE_LEITCH_FORMAT                  0x00000008L // SMPTE, Leitch Date format
//#define TC_SOURCE_DATE_LEITCH_FORMAT_WITH_MC_CONTROL_BITS  0x00000009L // Masterclock control bits not implemented
//#define TC_SOURCE_DATE_IRIG_1344                    0x0000000AL // not implemented
//#define TC_SOURCE_DATE_SMPTE_309M                   0x0000000BL // not implemented
#define TC_SOURCE_DATE_SMPTE_309M_WITH_TIME_ZONE      0x0000000CL // SMPTE 309M
#define TC_SOURCE_DATE_SMPTE_309M_NO_TIME_ZONE        0x0000000DL // SMPTE 309M
#define TC_SOURCE_DATE_DOY_YY                         0x0000000EL // IRIG Day Of Year and Year
#define TC_SOURCE_DATE_DOY                            0x0000000FL // IRIG Day Of year only
#define TC_SOURCE_UNDEFINED				              0xffffffffL // undefined

#define NO_DATE_AVAILABLE                     0
#define DATE_AVAILABLE_MMDDYY                 1
#define DATE_AVAILABLE_MMDDYY_WITH_TIME_ZONE  2
#define DATE_AVAILABLE_MMDDYY_NO_TIME_ZONE    3
#define DAY_OF_YEAR_AVAILABLE                 4


// time code types

// now defined in IntAptDefs.h
#define TC_TYPE_MASK            0x0000000FL
#define TC_TYPE_SMPTE			0x00000001L
#define TC_TYPE_IRIG			0x00000002L	// deprecated
#define TC_TYPE_IRIG_B1			0x00000002L	// originally, IRIG was only B1
#define TC_TYPE_IRIG_B			0x00000003L
#define TC_TYPE_UNDEFINED		0x00008000L

// time code sub-types
// Note: not sure if these get returned from the driver
#define TC_SUBTYPE_MASK			0x00FF0000L
#define TC_SUBTYPE_SMPTE30FPS	0x00010000L
#define TC_SUBTYPE_EBU25FPS		0x00020000L
#define TC_SUBTYPE_FILM24FPS	0x00040000L
#define TC_SUBTYPE_IRIG_B1		0x00080000L
#define TC_SUBTYPE_IRIG_B		0x00100000L
#define TC_SUBTYPE_UNDEFINED	0x80000000L
#pragma pack(push, 4)  // force word structure packing so that all programs using
					   // this header are guaranteed the same size objects

//
// a structure defining a SMPTE time code frame
//

typedef struct _SMPTE
{
	unsigned char           fu;         // frame units
	unsigned char           u1;         // user group 1
	unsigned char           ft;         // frame tens
	unsigned char           u2;         // user group 2
	unsigned char           su;         // second units
	unsigned char           u3;         // user group 3
	unsigned char           st;         // second tens
	unsigned char           u4;         // user group 4
	unsigned char           mu;         // minute units
	unsigned char           u5;         // user group 5
	unsigned char           mt;         // minute tens
	unsigned char           u6;         // user group 6
	unsigned char           hu;         // hour units
	unsigned char           u7;         // user group 7
	unsigned char           ht;         // hour tens
	unsigned char           u8;         // user group 8
} SMPTE, *PSMPTE;


//
// a structure defining an IRIG-B time code frame
//

typedef struct _IRIG
{
	unsigned char           su;         // second units
	unsigned char           st;         // second tens
	unsigned char           mu;         // minute units
	unsigned char           mt;         // minute tens
	unsigned char           hu;         // hour units
	unsigned char           ht;         // hour tens
	unsigned char           du;         // day units
	unsigned char           dt;         // day tens
	unsigned char           dh;         // day hundreds
	unsigned char			cf1;		// control functions 1
	unsigned char			cf2;		// control functions 2
	unsigned char			cf3;		// control functions 3
	unsigned char			cf4;		// control functions 4
	unsigned char			cf5;		// control functions 5
	unsigned char			cf6;		// control functions 6
	unsigned char			cf7;		// control functions 7
} IRIG, *PIRIG;

union uTimeCode
{
	IRIG        irig;
	SMPTE       smpte;
};

//********************************************************************************
//
// Data structures and definitions used by McrSync's named pipe and DDE
// interprocess communications.
//
// The Windows NT/2000 McrSync services utilize the named pipe mechanism.
// The Windows 9x/Me McrSync services utilize the DDE mechanism.
//
//********************************************************************************

#define MCRSYNC_IPC_NAME				"MCRSYNC_PIPE"
#define MCRSYNC_IPC_SYNCHISTMAX			10

// get synchronization service version
#define MCRSYNC_IPCCMD_VERSION			20
#define MCRSYNC_IPCCMD_VERSION_STR		"MCRSYNC_VERSION"

// get synchronization history
#define MCRSYNC_IPCCMD_SYNCHIST			21
#define MCRSYNC_IPCCMD_SYNCHIST_STR		"MCRSYNC_SYNCHIST"

// get current UTC time from service
#define MCRSYNC_IPCCMD_CURTIME			22
#define MCRSYNC_IPCCMD_CURTIME_STR		"MCRSYNC_CURTIME"

// get MCRSYNC operation info
#define MCRSYNC_IPCCMD_INFO				23
#define MCRSYNC_IPCCMD_INFO_STR			"MCRSYNC_INFO"

// causes synchronization service to set 60Hz interrupt frequency
#define MCRSYNC_IPCCMD_SETINTFREQ		24
#define MCRSYNC_IPCCMD_SETINTFREQ_STR	"MCRSYNC_SETINTFREQ"

// tell the TCR card to recalibrate
#define MCRSYNC_IPCCMD_RECALIBRATE		25
#define MCRSYNC_IPCCMD_RECALIBRATE_STR	(_T("MCRSYNC_RECALIBRATE"))

// tell the PCIe card to set the time
#define MCRSYNC_IPCCMD_SETTIME			26
#define MCRSYNC_IPCCMD_SETTIME_STR		(_T("MCRSYNC_SETTIME"))

// parameters for MCRSYNC_IPCCMD_SETINTFREQ

#define INTERRUPT_FREQ_DEFAULT			100	// 1/60Hz interrupt from time code reader
#define INTERRUPT_FREQ_PERSEC			101	// 1Hz interrupt from time code reader


// kpw
// get the AVR's firmware version
#define MCRSYNC_IPCCMD_FIRMWARE_VERSION		27
#define MCRSYNC_IPCCMD_FIRMWARE_VERSION_STR	"MCRSYNC_FIRMWARE_VERSION"

// kpw - command from monitor to send HSO info.
#define MCRSYNC_IPCCMD_LOCK_STATUS		28
#define MCRSYNC_IPCCMD_LOCK_STATUS_STR	"MCRSYNC_LOCK_STATUS"



// time synchronization status flags

#define CURTIME_HEALTHY				0   // locked to a reference, GPS or time code
#define CURTIME_UNHEALTHY			1   // We have had a lock in the past, but it is not locked right now
#define CURTIME_INVALID				2   // We have never been locked to a reference

// time synchronization failure detail status flags

#define REASON_NOREASON			0	// indicates no reason/no condition
#define REASON_TIMELOSS			1	// time is unhealthy due to no reference lock
#define REASON_SERIOUSTIMELOSS	2	// serious time code loss has occurred
#define REASON_GPSLOSS			3	// time code reference is freewheeling
									// GPS-200A/MCR1000/MCR5000 operation statuses when Masterclock, Inc proprietary control
									// bits are included in SMPTE time code, Leitch date encoding)
#define REASON_SERIOUSGPSLOSS	4	// serious satellite acquisition problem
									// GPS-200A/MCR1000/MCR5000 operation statuses when Masterclock, Inc proprietary control
									// bits are included in SMPTE time code, Leitch date encoding)
#define REASON_NEVER_LOCKED		5	// Never been locked to a reference


// Pipe transaction control structure:
// -----------------------------------

typedef struct
{
	DWORD	cmd;						// command
	DWORD	length;						// length of data to be expected (0 if none)
} PIPE_CONTROL, *PPIPE_CONTROL;


// Software version information:
// -----------------------------

typedef struct
{
	DWORD	verMaj;						// McrSync major version number
	DWORD	verMin;						// McrSync minor version number
} MCRSYNC_IPC_VERSION, *PMCRSYNC_IPC_VERSION;



// kpw -  Firmware version structure. copied from above.
// -----------------------------

typedef struct
{
	DWORD	verMaj;
	DWORD	verMin;
	DWORD	verRev;
} MCRSYNC_IPC_FIRMWARE_VERSION, *PMCRSYNC_IPC_FIRMWARE_VERSION;


/*
// Copied from McrSerialProtocol.h in api. Also is in AVR, same file name.
typedef struct _LAST_LOCK_STATUS
{
	UINT8 ucLockStatus;
	UINT8 ucRtcStatus;
	UINT8 ucGpsStatus;
	UINT8 ucTimeCodeStatus;
	UINT8 ucHsoStatus;
    UINT8 ucHsoAchievementLevel;
	UINT8 ucReserved1;
	UINT8 ucReserved2;
	UINT8 ucReserved3;
	UINT8 ucReserved4;
} LAST_LOCK_STATUS;

typedef struct _LAST_ALARM_STATUS
{
	LAST_LOCK_STATUS srLockStatus; // above
	UINT8 ucHwStatus;
	UINT8 verMajor;
	UINT8 verMinor;
	UINT8 verRevision;
	UINT8 ucReserved1;
	UINT8 ucReserved2;
	UINT8 ucReserved3;
	UINT8 ucReserved4;
} LAST_ALARM_STATUS, *LAST_ALARM_STATUS; // kpw - added the pointer.
*/


// kpw - copied from IntApiDefs.h in sync service
typedef struct _API_LAST_LOCK_STATUS
{
	BYTE ucLockStatus;
	BYTE ucRtcStatus;
	BYTE ucGpsStatus;
	BYTE ucTimeCodeStatus;
	BYTE ucHsoStatus;
	BYTE ucHsoAchievementLevel;
	BYTE ucReserved1;
	BYTE ucReserved2;
	BYTE ucReserved3;
	BYTE ucReserved4;
} API_LAST_LOCK_STATUS, *PAPI_LAST_LOCK_STATUS; // kpw - added the pointer.





// Last-sync time information:
// ---------------------------
// The time provided here is a record of the last MCRSYNC_PIPE_SYNCHISTMAX time
// synchronizations that occured on the NT system running McrSync.  The numSyncRecords
// member indicates how many of the record entries are valid (filled in) starting with
// entry 0.  Entry 0 represents the most recent time synchronization, etc.
// (This information is not for time synchronization use!)

typedef struct
{
	DWORD		numSyncRecords;			// number entries in synchronization record
	WCATIME		decodedTime[MCRSYNC_IPC_SYNCHISTMAX];
	WCATIME		utcTime[MCRSYNC_IPC_SYNCHISTMAX];
	SYSTEMTIME	systemTime[MCRSYNC_IPC_SYNCHISTMAX]; // before synchronization
} MCRSYNC_IPC_SYNCHIST, *PMCRSYNC_IPC_SYNCHIST;


// McrSync current time information:
// ---------------------------------
// The current time as derived by McrSync from the time code source.  If the time code source
// does not provide date information date it is derived from the NT system running McrSync
// (but the date members of 'rawTime' will be set to zero).
//
// If 'projectedTime' is non-zero it indicates the number of seconds since the last interrupt
// that provided McrSync with time.  The time contained in the 'rawTime' and 'utcTime' members
// will projected forward by that number of seconds. ('wMilliseconds' member of each of the
// two time objects is also correctly projected.)  Projected time is considered accurate for
// time synchronization over this named pipe mechanism unless timeStatus is flagged other
// than healthy.
//
//  The 'localTime' member of this object is always the current local system time on the
// system running McrSync and is not guaranteed to be synchronized with the
// 'rawTime'/'utcTime' members and should never be used for time synchronization
// purposes.  ONLY THE 'utcTime' MEMBER SHOULD BE USED FOR TIME SYNCHRONIZATION.

typedef struct {
	BYTE			timeStatus;		// status of provided time
	BYTE			extStatus;		// extended explanation of status (if timeStatus not healthy)
	DWORD			projectedTime;	// time given in UTC and local time members is projected
	WCATIME			rawTime;		// time as read from time code reader at last interrupt
	WCATIME			utcTime;		// current UTC time (use this for time sync)
	SYSTEMTIME		localTime;		// current time on host running McrSync

	DWORD			reserved1;
	DWORD			reserved2;
} MCRSYNC_IPC_CURTIME, *PMCRSYNC_IPC_CURTIME;



typedef struct {

	WCATIME			utcTimeToSet;		// current UTC time (use this to set the PCIe time)

	DWORD			reserved1;
	DWORD			reserved2;
} MCRSYNC_IPC_SETTIME, *PMCRSYNC_IPC_SETTIME;



// McrSync operation information:
// ------------------------------
// Operation information and event counts.  All event counts are relative to the last time
// McrSync was started.

typedef struct {
	SYSTEMTIME	startTime;			// local clock time when McrSync started
	DWORD		syncCount;			// number of synchronizations that have been applied
	DWORD		rCompare;			// rolling time comparison failures
	DWORD		seriousTimeLoss;	// serious time losses detected
	DWORD		seriousGpsLoss;		// serious GPS losses detected
	DWORD		TimeCodeType;		// IRIG, SMPTE
	uTimeCode   rawTc;				// raw time code
	DWORD		reserved1;
	DWORD		reserved2;
	DWORD		reserved3;
	DWORD		reserved4;
} MCRSYNC_IPC_INFO, *PMCRSYNC_IPC_INFO;

// in the API and service, these are included in IntApiDefs.h
#define INVALID_DST_ENCODED           0x00000001UL
#define INVALID_FREEWHEEL_DISPLAY     0x00000002UL
#define INVALID_FREEWHEEL_SYNCHRONIZE 0x00000004UL
#define INVALID_TIME_CODE_LOSS_NOTIFY 0x00000008UL
#define INVALID_TIME_CODE_SOURCE      0x00000010UL
#define INVALID_TIME_ZONE_BIAS        0x00000020UL
#define INVALID_STANDARD_BIAS         0x00000040UL
#define INVALID_STANDARD_YEAR         0x00000080UL
#define INVALID_STANDARD_MONTH        0x00000100UL
#define INVALID_STANDARD_DAY_OF_WEEK  0x00000200UL
#define INVALID_STANDARD_DATE         0x00000400UL
#define INVALID_STANDARD_HOUR         0x00000800UL
#define INVALID_STANDARD_MINUTE       0x00001000UL
#define INVALID_DAYLIGHT_BIAS         0x00002000UL
#define INVALID_DAYLIGHT_YEAR         0x00004000UL
#define INVALID_DAYLIGHT_MONTH        0x00008000UL
#define INVALID_DAYLIGHT_DAY_OF_WEEK  0x00010000UL
#define INVALID_DAYLIGHT_DATE         0x00020000UL
#define INVALID_DAYLIGHT_HOUR         0x00040000UL
#define INVALID_DAYLIGHT_MINUTE       0x00080000UL
#define INVALID_NMEA_CONFIGURATION    0x00100000UL
//
//// bitmap for enabled NMEA messages
#define NMEA_MESSAGE_MASK		0x7F
#define NMEA_MESSAGE_BIT_ZDA    0x01
#define NMEA_MESSAGE_BIT_GSV    0x02
#define NMEA_MESSAGE_BIT_GGA    0x04
#define NMEA_MESSAGE_BIT_GLL    0x08
#define NMEA_MESSAGE_BIT_GSA    0x10
#define NMEA_MESSAGE_BIT_RMC    0x20
#define NMEA_MESSAGE_BIT_VTG    0x40
#define NMEA_MESSAGE_BIT_GST    0x80

typedef struct {
	INT32 DstEncoded;				// ==0, time code never offset by daylight time
									// ==1, time code offset by daylight time during daylight period
	INT32 FreewheelDisplay;			// ==0, time code does not contain Masterclock, Inc proprietary freewheeling indicator in SMPTE frame
									// ==1, when freewheeling, time code will indicate such using Masterclock, Inc proprietary lock indicator in SMPTE frame
	INT32 FreewheelSynchronize;		// ==0, when freewheeling (FreewheelDisplay must ==1 to know when freewheeling), do not set system time
									// ==1, when freewheeling (FreewheelDisplay must ==1 to know when freewheeling), set system time to incoming time code
	INT32 TimeCodeLossNotify;		// ==0, do not generate alarm when time code lock is lost
									// ==1, generate alarm when time code lock is lost
	INT32 PciComPort;        		// serial port number to use
	
	// settings for outputting the NMEA messages
	INT32 iNmeaMessagesBitMap;
	INT32 iNmeaBaudrate;
	INT32 iNmeaDataBits;
	INT32 iNmeaParity;
	INT32 iNmeaStopBits;
	INT32 iNmeaComPortNumber;

	INT32 TimeCodeSourceSmpte;		// see TC_SOURCE_DATE_NOT_INCLUDED, etc.
	INT32 TimeCodeSourceIrig;		// see TC_SOURCE_DATE_NOT_INCLUDED, etc.
	INT32 TimeZoneBiasSeconds;		// time zone bias, in seconds
	INT32 StandardBiasSeconds;		// standard bias, in seconds (usually zero)
	INT32 StandardYear;				// zero for relative standard time, else year for absolute standard date (usually zero)
	INT32 StandardMonth;			// month standard time begins
	INT32 StandardDayOfWeek;		// day of week standard time begins, SUN==1, MON==2, TUE==3, etc.
	INT32 StandardDate;             // date of month standard time begins
	INT32 StandardHour;				// hour of day standard time begins
	INT32 StandardMinute;           // minute of hour standard time begins
	INT32 DaylightBiasSeconds;		// daylight bias, in seconds (usually zero)
	INT32 DaylightYear;				// zero for relative daylight time, else year for absolute daylight date (usually zero)
	INT32 DaylightMonth;			// month daylight time begins
	INT32 DaylightDayOfWeek;		// day of week daylight time begins, SUN==1, MON==2, TUE==3, etc.
	INT32 DaylightDate;				// date of month daylight time begins
	INT32 DaylightHour;				// hour of day daylight time begins
	INT32 DaylightMinute;			// minute of hour daylight time begins
} MCR_CONFIGURATION;

#pragma pack(pop)


